#include "ctfapply_gpu.h"
#include "ctfapply_kernel.h"
#include "maskenv_gpu.h"
#include "maskcos_gpu.h"

//  Applies calculated CTF to a particle transform and also masks the 
//  particle in real space with a cosine-bell-edged circle of radius RI
//  Calls rlft3 and MASKENV.
//  Uses function CTF.
//  Used in LMAIN.
//
//   The CUDA/C++ code is based on the Fortran code from v8.06

extern pthread_key_t cuMemKey;

void ctfapply(int *pnsam, CPLX *spec, CPLX *speq, 
	float *pshx, float *pshy, float *pcs, float *pwl, float *pwgh1, float *pwgh2, 
	float *pdfmid1, float *pdfmid2, float *pangast, float *pthetatr, CPLX *ctff, 
	CPLX *ctfs, float *outd, CPLX *outc, CPLX *outq, float *pamagp, 
	float *prih, float *phalfw, float *pri2, float *pri3, float *pri4, 
	float *datd, CPLX *datc, CPLX *datq, float *b3dv, float *pphi, 
	float *ptheta, float *ppsi, float *w, float *pxstd, int *pibuf, 
	float *ptx, float *pty)
{
	int nsam=*pnsam;

	float shx=*pshx;
	float shy=*pshy;
	float cs=*pcs;
	float wl=*pwl;
	float wgh1=*pwgh1;
	float wgh2=*pwgh2;
	
	float dfmid1=*pdfmid1;
	float dfmid2=*pdfmid2;
	float angast=*pangast;
	float thetatr=*pthetatr;
	
	float amagp=*pamagp;
	
	float rih=*prih;
	float halfw=*phalfw;
	float ri2=*pri2;
	float ri3=*pri3;
//	float ri4=*pri4;
	
	float phi=*pphi;
	float theta=*ptheta;
	float psi=*ppsi;
	float xstd=*pxstd;
	int ibuf=*pibuf;
	float tx=*ptx;
	float ty=*pty;
	
	int jc;
	int nsamh,ic;
	float thetatrp;
	float rihp,scal;
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);

//	timeval begin,end;
//	gettimeofday(&begin,NULL);

	thetatrp=thetatr/amagp;
	rihp=rih/amagp;
	scal=nsam*nsam/2.0f;
	//	  THETATRP is the diffraction angle for the first pixel in 
	//	  the transform of a particle sampled at the input pixel size	
	jc=nsam/2+1;
	nsamh=nsam/2;
//	nmid=nsam/2+1;
	ic=nsam*nsam/2;
	dim3 dimBlock(8,16);
	dim3 dimGrid(jc/dimBlock.x+1,nsam/dimBlock.y+1);
	cudaThreadSynchronize();  //because the synchronization was not done in pinsert_s.cu, move it here
	cudaMemcpy((void *)pGM->d_spec,(void *)spec,sizeof(float)*nsam*nsam, cudaMemcpyHostToDevice);
	cudaMemcpy((void *)pGM->d_speq,(void *)speq,sizeof(float)*nsam*2, cudaMemcpyHostToDevice);

	//CORRECT FOR CTF AND APPLY SHIFTS
	CTFAndShift<<<dimGrid,dimBlock>>>((CPLX *)pGM->d_spec, (CPLX *)pGM->d_speq, 
					(CPLX *)pGM->d_datc, (CPLX *)pGM->d_datq, (CPLX *)pGM->d_outc, 
					(CPLX *)pGM->d_outq, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, nsam, jc, ic, 
					shx, shy, cs, wl, wgh1, wgh2, dfmid1, dfmid2, angast, thetatrp, tx, ty);
	cudaThreadSynchronize();

	// do fft
	rlft3_GPU((CPLX *)pGM->d_outc,(CPLX *)pGM->d_outq,nsam,nsam,1,-1);

	//had moved the outd/scal to calculate_sum of maskcos by xueming
	//see the lastest parameters of maskcos below
	
	//cos mask
	maskcos(nsam,pGM->d_outc,ri2,ri3,rih,halfw,amagp,scal);

	if(xstd<0.0)
	{	
		maskenv(nsam,rihp,(float *)pGM->d_outc,b3dv,0.0f,0.0f,phi,theta,psi,w,ibuf,pGM);
	}
	
	//do fft
	rlft3_GPU((CPLX *)pGM->d_outc,(CPLX *)pGM->d_outq,nsam,nsam,1,1);
					
	//SHIFT PARTICLE TO ORIGIN
	ShiftToOrigin<<<dimGrid,dimBlock>>>((CPLX *)pGM->d_datc,(CPLX *)pGM->d_datq, 
							(CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, nsam, jc, nsamh);
	cudaThreadSynchronize();

	//copy the results back to host memory
	cudaMemcpy((void *)datc,(void *)pGM->d_datc,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)datq,(void *)pGM->d_datq,sizeof(float)*nsam*2,cudaMemcpyDeviceToHost);
	
	cudaMemcpy((void *)outc,(void *)pGM->d_outc,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)outq,(void *)pGM->d_outq,sizeof(float)*nsam*2,cudaMemcpyDeviceToHost);

	cudaMemcpy((void *)ctff,(void *)pGM->d_ctff,sizeof(float)*nsam*nsam*2,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)ctfs,(void *)pGM->d_ctfs,sizeof(float)*nsam*4,cudaMemcpyDeviceToHost);	
		
//		gettimeofday(&end,NULL);
//		cout<<" CTFAPPLY time is "<<CalTime(begin,end)<<endl;
}








