#include "cuMem.h"
#include <iostream>
#include "assert.h"
#include "inc.h"
#include "func.h"
using namespace std;

// Class for GPU memory management


//mutul lock for get_rec 
extern pthread_mutex_t mutex_getrec11,mutex_getrec12,mutex_getrec13,mutex_getrec14;
extern pthread_mutex_t mutex_getrec21,mutex_getrec22,mutex_getrec23,mutex_getrec24;
extern pthread_mutex_t mutex_getstat1,mutex_getstat2,mutex_getstat3;


int safecall(cudaError_t call, const char* fname) 
{
	if (call != cudaSuccess)
	{
		printf("Call to '%s' returned '%s'\n", fname, cudaGetErrorString(call));
		return 1;
	}
	
	return 0;
}

void CpyHostTo3DArray(cudaArray **dest, float *src, int width, int height, int depth)
{
	// setup texture dimension
	cudaExtent extent;
	extent.width = width;
	extent.height = height;
	extent.depth = depth;
   cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<CPLX>();
   safecall(cudaMalloc3DArray(dest,&channelDesc,extent),"cudaMalloc3DArray");
   // copy from host memory to device memory
	cudaMemcpy3DParms copyParams = {0};
	copyParams.srcPtr   = make_cudaPitchedPtr((void *)src, extent.width*sizeof(CPLX), extent.width, extent.height);  
	copyParams.dstArray = *dest;
	copyParams.extent   = extent;
	copyParams.kind     = cudaMemcpyHostToDevice;
	safecall(cudaMemcpy3D(&copyParams),"cudaMemcpy3D");
}

void CpyHostTo2DArray(cudaArray **dest, float *src, int width, int height)
{
	// setup texture dimension
   cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<CPLX>();
   safecall(cudaMallocArray(dest,&channelDesc,width,height),"cudaMalloc2DArray");
   // copy from host memory to device memory
	safecall(cudaMemcpyToArray(*dest, 0, 0, src, sizeof(CPLX)*width*height,cudaMemcpyHostToDevice),"cudaMemcpy3D");
}

cuMem::cuMem()
{
	initial();
}

cuMem::~cuMem()
{

}

void cuMem::initial()
{
	d_spec=NULL; // save one particle image 
	d_speq=NULL; 
	
	d_datc=NULL;
	d_datq=NULL;
	
	d_outc=NULL;
	d_outq=NULL;

	d_ctff=NULL;  // save ctf of one particle
	d_ctfs=NULL;
	
	d_fft2d=NULL;
	d_fft2dt=NULL;
	m_fft2dsize=0;
	d_mbuf=NULL;
	d_bf=NULL;
	
	d_sinclut=NULL;
	d_dm=NULL;
	d_wtf=NULL; //used in forward rlft3
	d_wtb=NULL; //used in backward rlft3
	
	d_mccpart=NULL;
	d_msum1=NULL;
	d_msum2=NULL;
	d_misum=NULL;
	
	//computing matrix
	d_a3df=NULL;  //3D reconstructuion 1
	d_a3ds=NULL;
	d_s3df=NULL;  //stat 1
	d_s3ds=NULL;
	
	d_d3df=NULL;  //3D reconstructuion 1
	d_d3ds=NULL;
	d_v3df=NULL;  //stat 1
	d_v3ds=NULL;
	
	d_c3df=NULL;  //reference volum
	d_c3ds=NULL;
	
	d_sumprel=NULL;
	d_nprel=NULL;
	
	d_asum=NULL;
	d_vsum=NULL;
	d_psum=NULL;
	d_ksum=NULL;
	
	d_std=NULL;
	d_vtd=NULL;
	d_nksum=NULL;
	
	d_vn=NULL;
	d_vvn=NULL;
	d_vsn=NULL;
	d_vvsn=NULL;
	d_ic=NULL;
	
	//page-locked memory
	m_plb3dv=NULL;
	
	//constant block size
	rc_x=16;
	rc_y=8;
	
	//private
	m_nrec=-2;

}
void cuMem::setThreadBlock(int idx, int idy)
{
	rc_x=idx;
	rc_y=idy;
}

void cuMem::set(int nsam, int ipad, int impmax, int nnstat,float *sinclut, PageLockedMem *plm)
{
	m_nsam=nsam;
	m_ipad=ipad;
	m_impmax=impmax;
	m_nnstat=nnstat;
	m_sinclut=sinclut;
	m_c3df=plm->pl_c3dv;
	m_c3ds=plm->pl_c3ds;
	//page-locked
	
	//mask volumn
	if(cudaHostGetDevicePointer((void **)&m_plb3dv,(void *)plm->pl_b3dv,0)!=cudaSuccess)
   {
   	cout<<"error: Get CUDA Device Pointer for mapped b3dv failed!"<<endl;
   	exit(0);
   }
   
   //reference volumn
   if(cudaHostGetDevicePointer((void **)&d_c3df,(void *)plm->pl_c3dv,0)!=cudaSuccess)
   {
   	cout<<"error: Get CUDA Device Pointer for mapped c3df failed!"<<endl;
   	exit(0);
   }
   if(cudaHostGetDevicePointer((void **)&d_c3ds,(void *)plm->pl_c3ds,0)!=cudaSuccess)
   {
   	cout<<"error: Get CUDA Device Pointer for mapped c3ds failed!"<<endl;
   	exit(0);
   }
     
}
			   
//allocate memory in GPU
int cuMem::alloc(int nrec)
{

  if(nrec<-1 || nrec>2) return -1;
  if(m_nrec!=-2) return -1;
  m_nrec=nrec;
  	//if nrec==-1: do refinement
	//if nrec==0 : allocate all
	//if nrec==1 : allocate reconstruction 1
	//if nrec==2 : allocate reconstruction 2
  int nerror=0;
	//
  if(cudaMalloc((void **)&d_spec, sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++; 
  if(cudaMalloc((void **)&d_speq, sizeof(float)*m_nsam*2)!=cudaSuccess) nerror++; 
  if(cudaMalloc((void **)&d_datc, sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++; 
  if(cudaMalloc((void **)&d_datq, sizeof(float)*m_nsam*2)!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_outc, sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++; 
  if(cudaMalloc((void **)&d_outq, sizeof(float)*m_nsam*2)!=cudaSuccess) nerror++;
  
  if(cudaMalloc((void **)&d_ctff, sizeof(float)*m_nsam*m_nsam*2)!=cudaSuccess) nerror++; 
  if(cudaMalloc((void **)&d_ctfs, sizeof(float)*m_nsam*4)!=cudaSuccess) nerror++;
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<CPLX>();
  if(cudaMallocArray(&cu_ctff,&channelDesc,m_nsam/2,m_nsam*2)!=cudaSuccess) nerror++;

  
  m_fft2dsize=m_nsam*(m_nsam+2);
  if(cudaMalloc((void **)&d_fft2d,sizeof(float)*m_fft2dsize)!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_fft2dt,sizeof(float)*m_fft2dsize)!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_mbuf, sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++; //size different from mbuf
  if(cudaMalloc((void **)&d_bf, sizeof(float)*m_nsam*m_nsam*20)!=cudaSuccess) nerror++; //wrong in pad.f
  
  if(cudaMalloc((void **)&d_sinclut,sizeof(float)*2000)!=cudaSuccess) nerror++;
  if(cudaMemcpy(d_sinclut,m_sinclut,sizeof(float)*2000,cudaMemcpyHostToDevice)!=cudaSuccess) nerror++;
  
  if(cudaMalloc((void **)&d_dm,sizeof(float)*9)!=cudaSuccess) nerror++;
  
  if(cudaMalloc((void **)&d_mccpart,sizeof(float)*512)!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_msum1,sizeof(float)*1024)!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_msum2,sizeof(float)*512)!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_misum,sizeof(int)*1024)!=cudaSuccess) nerror++;
  
  //*******************************
  //***********for rlft3***********
  if(cudaMalloc((void **)&d_wtf,sizeof(CPLX)*(m_nsam/4+1))!=cudaSuccess) nerror++;
  if(cudaMalloc((void **)&d_wtb,sizeof(CPLX)*(m_nsam/4+1))!=cudaSuccess) nerror++;
  double theta,wi,wpi,wpr,wr,wtemp;
  int isign,i;
  	CPLX wt[512],w;
  	
  //forward
  isign=1;
	theta=6.28318530717959/double(isign*m_nsam);
	wpr=-2.0*pow(sin(0.5*theta),2);
	wpi=sin(theta);
	w.x=0;
	w.y=0;	
	wr=1.0;
	wi=0.0;
	for(i=0;i<=(m_nsam/4);i++)
	{
		wt[i]=w;
		wtemp=wr;
		wr=wr*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
		w.x=wr;
		w.y=wi;
	}
	cudaMemcpy(d_wtf,wt,sizeof(CPLX)*(m_nsam/4+1),cudaMemcpyHostToDevice);
	
	//backward
	isign=-1;
	theta=6.28318530717959/double(isign*m_nsam);
	wpr=-2.0*pow(sin(0.5*theta),2);
	wpi=sin(theta);
	w.x=0;
	w.y=0;	
	wr=1.0;
	wi=0.0;
	for(i=0;i<=(m_nsam/4);i++)
	{
		wt[i]=w;
		wtemp=wr;
		wr=wr*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
		w.x=wr;
		w.y=wi;
	}
	cudaMemcpy(d_wtb,wt,sizeof(CPLX)*(m_nsam/4+1),cudaMemcpyHostToDevice);
  
  //*******************************
  //*******************************

  //
  //construction 1
  if(nrec==0 || nrec==1)
  {
	  //if(cudaMalloc((void **)&d_a3df, sizeof(float)*m_nsam*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_a3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_a3df, sizeof(float)*m_nsam*m_nsam*m_nsam),"Malloc d_a3df");
	  nerror+=safecall(cudaMemset((void *)d_a3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam),"d_a3df");
	  
	  //if(cudaMalloc((void **)&d_a3ds, sizeof(float)*m_nsam*m_nsam*2)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_a3ds,0,sizeof(float)*m_nsam*m_nsam*2)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_a3ds, sizeof(float)*m_nsam*m_nsam*2),"Malloc d_a3ds");
	  nerror+=safecall(cudaMemset((void *)d_a3ds,0,sizeof(float)*m_nsam*m_nsam*2),"d_a3ds");
	  
	  //if(cudaMalloc((void **)&d_s3df, sizeof(float)*m_nsam*m_nsam*m_nsam/2)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_s3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam/2)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_s3df, sizeof(float)*m_nsam*m_nsam*m_nsam/2),"Malloc d_s3df");
	  nerror+=safecall(cudaMemset((void *)d_s3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam/2),"d_s3df");
	  
	  //if(cudaMalloc((void **)&d_s3ds, sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_s3ds,0,sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_s3ds, sizeof(float)*m_nsam*m_nsam),"Malloc d_s3ds");
	  nerror+=safecall(cudaMemset((void *)d_s3ds,0,sizeof(float)*m_nsam*m_nsam),"d_s3ds");
  }
  
  //construction 2
  if(nrec==0 || nrec==2)
  {
	  //if(cudaMalloc((void **)&d_d3df, sizeof(float)*m_nsam*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_d3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_d3df, sizeof(float)*m_nsam*m_nsam*m_nsam),"Malloc d_d3df");
	  nerror+=safecall(cudaMemset((void *)d_d3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam),"d_d3df");
	  
	  //if(cudaMalloc((void **)&d_d3ds, sizeof(float)*m_nsam*m_nsam*2)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_d3ds,0,sizeof(float)*m_nsam*m_nsam*2)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_d3ds, sizeof(float)*m_nsam*m_nsam*2),"Malloc d_d3ds");
	  nerror+=safecall(cudaMemset((void *)d_d3ds,0,sizeof(float)*m_nsam*m_nsam*2),"d_d3ds");
	  
	  //if(cudaMalloc((void **)&d_v3df, sizeof(float)*m_nsam*m_nsam*m_nsam/2)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_v3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam/2)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_v3df, sizeof(float)*m_nsam*m_nsam*m_nsam/2),"Malloc d_v3df");
	  nerror+=safecall(cudaMemset((void *)d_v3df,0,sizeof(float)*m_nsam*m_nsam*m_nsam/2),"d_v3df");
	  
	  //if(cudaMalloc((void **)&d_v3ds, sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  //if(cudaMemset((void *)d_v3ds,0,sizeof(float)*m_nsam*m_nsam)!=cudaSuccess) nerror++;
	  nerror+=safecall(cudaMalloc((void **)&d_v3ds, sizeof(float)*m_nsam*m_nsam),"Malloc d_v3ds");
	  nerror+=safecall(cudaMemset((void *)d_v3ds,0,sizeof(float)*m_nsam*m_nsam),"d_v3ds");
  }

	//other
	if(nrec>=0)
	{
		if(cudaMalloc((void **)&d_sumprel, sizeof(float)*m_impmax*m_nsam/2)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_sumprel,0,sizeof(float)*m_impmax*m_nsam/2)!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_nprel, sizeof(int)*m_impmax*m_nsam/2)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_nprel,0,sizeof(int)*m_impmax*m_nsam/2)!=cudaSuccess) nerror++;
  
    	if(m_nnstat==1)
  		{
			//if(cudaMalloc((void **)&d_asum, sizeof(float)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			//if(cudaMemset((void *)d_asum,0,sizeof(float)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			nerror+=safecall(cudaMalloc((void **)&d_asum, sizeof(float)*m_nsam/2*m_nsam*m_nsam),"Malloc d_asum");
			nerror+=safecall(cudaMemset((void *)d_asum,0,sizeof(float)*m_nsam/2*m_nsam*m_nsam),"d_asum");
	  
			//if(cudaMalloc((void **)&d_vsum, sizeof(float)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			//if(cudaMemset((void *)d_vsum,0,sizeof(float)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			nerror+=safecall(cudaMalloc((void **)&d_vsum, sizeof(float)*m_nsam/2*m_nsam*m_nsam),"Malloc d_vsum");
			nerror+=safecall(cudaMemset((void *)d_vsum,0,sizeof(float)*m_nsam/2*m_nsam*m_nsam),"d_vsum");
	  
			//if(cudaMalloc((void **)&d_psum, sizeof(float)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			//if(cudaMemset((void *)d_psum,0,sizeof(float)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			nerror+=safecall(cudaMalloc((void **)&d_psum, sizeof(float)*m_nsam/2*m_nsam*m_nsam),"Malloc d_psum");
			nerror+=safecall(cudaMemset((void *)d_psum,0,sizeof(float)*m_nsam/2*m_nsam*m_nsam),"d_psum");
	  
			//if(cudaMalloc((void **)&d_ksum, sizeof(int)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			//if(cudaMemset((void *)d_ksum,0,sizeof(int)*m_nsam/2*m_nsam*m_nsam)!=cudaSuccess) nerror++;
			nerror+=safecall(cudaMalloc((void **)&d_ksum, sizeof(int)*m_nsam/2*m_nsam*m_nsam),"Malloc d_ksum");
			nerror+=safecall(cudaMemset((void *)d_ksum,0,sizeof(int)*m_nsam/2*m_nsam*m_nsam),"d_ksum");
  		}
  		
		if(cudaMalloc((void **)&d_std, sizeof(double)*(m_nsam+1)*(m_nsam+1))!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_std,0,sizeof(double)*(m_nsam+1)*(m_nsam+1))!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_vtd, sizeof(double)*(m_nsam+1)*(m_nsam+1))!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_vtd,0,sizeof(double)*(m_nsam+1)*(m_nsam+1))!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_nksum, sizeof(long)*(m_nsam+1)*(m_nsam+1))!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_nksum,0,sizeof(long)*(m_nsam+1)*(m_nsam+1))!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_vn,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_vn,0,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_vvn,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_vvn,0,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_vsn,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_vsn,0,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_vvsn,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_vvsn,0,sizeof(double)*m_nsam)!=cudaSuccess) nerror++;
  
		if(cudaMalloc((void **)&d_ic,sizeof(int)*m_nsam)!=cudaSuccess) nerror++;
		if(cudaMemset((void *)d_ic,0,sizeof(int)*m_nsam)!=cudaSuccess) nerror++;
	}
	
	//for refinement
	if(nrec==-1)
	{
		//for c3df and c3ds used in cc3m
	   	CpyHostTo3DArray(&cu_c3df,m_c3df,m_ipad*m_nsam/2,m_ipad*m_nsam,m_ipad*m_nsam);
	   	CpyHostTo2DArray(&cu_c3ds,m_c3ds,m_ipad*m_nsam,m_ipad*m_nsam);
	}
	
	return nerror;
}

//Add b to a, that is, a=a+b.
//a is page-locked memory; b is device memory.
__global__ void MatAdd(float *a, float *b, int size) 
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	if(i>=size) return;
	a[i]+=b[i];
}
//2D version of MatAdd, used for big matrix
//block dimension is 512*1
//each row has 32 blocks
//so, each row has 512*32=16384 threads
__global__ void Mat2DAdd(float *a, float *b, int size) 
{
	int i=(blockIdx.x*blockDim.x+threadIdx.x)+blockIdx.y*16384;

	if(i>=size) return;
	a[i]+=b[i];
}

//copy back 3D reconstructions
int cuMem::getRec(float *a3df, float *a3ds, float *s3df, float *s3ds,
					   float *d3df, float *d3ds, float *v3df, float *v3ds)
{
	if(m_nrec==-1) return -1;
	if(m_nrec<-1) return m_nrec;
	int size=0;
	float *dst=NULL;
	float *src=NULL;
	float *dsrc=NULL;
	int error=0;
	
	dim3 DimBlock(512,1);
	
	size=m_nsam*m_nsam*m_nsam;
	dim3 DimGridnnn(32,size/16384+1);
	
	size=m_nsam*m_nsam*m_nsam/2;
	dim3 DimGridnn2(32,size/16384+1);
	
	//construction 1
  	if(m_nrec==0 || m_nrec==1)
  	{
  		pthread_mutex_lock(&mutex_getrec11);
 		src=a3df;
 		dsrc=d_a3df;
  		size=m_nsam*m_nsam*m_nsam;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		//else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		else Mat2DAdd<<<DimGridnnn,DimBlock>>>(dst,dsrc,size);
 		error+=safecall(cudaThreadSynchronize(),"a3df");
 		pthread_mutex_unlock(&mutex_getrec11);
 		
 		pthread_mutex_lock(&mutex_getrec12);
 		src=a3ds;
 		dsrc=d_a3ds;
  		size=m_nsam*m_nsam*2;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		cudaThreadSynchronize();
 		pthread_mutex_unlock(&mutex_getrec12);
 		
 		pthread_mutex_lock(&mutex_getrec13);
 		src=s3df;
 		dsrc=d_s3df;
  		size=m_nsam*m_nsam*m_nsam/2;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		//else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		else Mat2DAdd<<<DimGridnn2,DimBlock>>>(dst,dsrc,size);
 		cudaThreadSynchronize();
 		pthread_mutex_unlock(&mutex_getrec13);
 		
 		pthread_mutex_lock(&mutex_getrec14);
 		src=s3ds;
 		dsrc=d_s3ds;
  		size=m_nsam*m_nsam;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		cudaThreadSynchronize();
 		pthread_mutex_unlock(&mutex_getrec14);
  	}
  //cudaMemcpy((void *)a3df,(void *)d_a3df,sizeof(float)*m_nsam*m_nsam*m_nsam, cudaMemcpyDeviceToHost);
  //cudaMemcpy((void *)a3ds,(void *)d_a3ds,sizeof(float)*m_nsam*m_nsam*2, cudaMemcpyDeviceToHost);
  //cudaMemcpy((void *)s3df,(void *)d_s3df,sizeof(float)*m_nsam*m_nsam*m_nsam/2, cudaMemcpyDeviceToHost);
  //cudaMemcpy((void *)s3ds,(void *)d_s3ds,sizeof(float)*m_nsam*m_nsam, cudaMemcpyDeviceToHost);
  
  //construction 2
  	if(m_nrec==0 || m_nrec==2)
  	{
  		pthread_mutex_lock(&mutex_getrec21);
 		src=d3df;
 		dsrc=d_d3df;
  		size=m_nsam*m_nsam*m_nsam;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		//else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		else Mat2DAdd<<<DimGridnnn,DimBlock>>>(dst,dsrc,size);
 		error+=safecall(cudaThreadSynchronize(),"d3df");
 		pthread_mutex_unlock(&mutex_getrec21);
 		
 		pthread_mutex_lock(&mutex_getrec22);
 		src=d3ds;
 		dsrc=d_d3ds;
  		size=m_nsam*m_nsam*2;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		cudaThreadSynchronize();
 		pthread_mutex_unlock(&mutex_getrec22);
 	
 		pthread_mutex_lock(&mutex_getrec23);
 		src=v3df;
 		dsrc=d_v3df;
  		size=m_nsam*m_nsam*m_nsam/2;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		//else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		else Mat2DAdd<<<DimGridnn2,DimBlock>>>(dst,dsrc,size);
 		cudaThreadSynchronize();
 		pthread_mutex_unlock(&mutex_getrec23);
 		
 		pthread_mutex_lock(&mutex_getrec24);
 		src=v3ds;
 		dsrc=d_v3ds;
  		size=m_nsam*m_nsam;
  		if(cudaHostGetDevicePointer((void **)&dst,(void *)src,0)!=cudaSuccess) error++; 
 		else MatAdd<<<size/512+1,512>>>(dst,dsrc,size);
 		cudaThreadSynchronize();
 		pthread_mutex_unlock(&mutex_getrec24);
 	}
  //cudaMemcpy((void *)d3df,(void *)d_d3df,sizeof(float)*m_nsam*m_nsam*m_nsam, cudaMemcpyDeviceToHost);
  //cudaMemcpy((void *)d3ds,(void *)d_d3ds,sizeof(float)*m_nsam*m_nsam*2, cudaMemcpyDeviceToHost);
  //cudaMemcpy((void *)v3df,(void *)d_v3df,sizeof(float)*m_nsam*m_nsam*m_nsam/2, cudaMemcpyDeviceToHost);
  //cudaMemcpy((void *)v3ds,(void *)d_v3ds,sizeof(float)*m_nsam*m_nsam, cudaMemcpyDeviceToHost);

  return error;
}

//copy statstics
int cuMem::getStat(float *sumprel, int * nprel, 
						 float *asum, float * vsum, float *psum, int *ksum,
						 double &std, double &vtd, long &nksum,
						 double *vn, double *vvn, double *vsn, double *vvsn, int *ic)
{
  if(m_nrec==-1) return -1;
  if(m_nrec<-1) return m_nrec;
  
  int i;
  float *h_sumprel=new float[m_impmax*m_nsam/2];
  int *h_nprel=new int[m_impmax*m_nsam/2];
  float *h_asum=NULL;
  float *h_vsum=NULL;
  float *h_psum=NULL;
  int *h_ksum=NULL;
  cudaMemcpy((void *)h_sumprel,(void *)d_sumprel,sizeof(float)*m_impmax*m_nsam/2,cudaMemcpyDeviceToHost);
  cudaMemcpy((void *)h_nprel,(void *)d_nprel,sizeof(int)*m_impmax*m_nsam/2,cudaMemcpyDeviceToHost);
  if(m_nnstat==1)
  {
	  h_asum=new float[m_nsam/2*m_nsam*m_nsam];
	  h_vsum=new float[m_nsam/2*m_nsam*m_nsam];
	  h_psum=new float[m_nsam/2*m_nsam*m_nsam];
	  h_psum=new float[m_nsam/2*m_nsam*m_nsam];
	  cudaMemcpy((void *)h_asum,(void *)d_asum,sizeof(float)*m_nsam/2*m_nsam*m_nsam,cudaMemcpyDeviceToHost);
	  cudaMemcpy((void *)h_vsum,(void *)d_vsum,sizeof(float)*m_nsam/2*m_nsam*m_nsam,cudaMemcpyDeviceToHost);
	  cudaMemcpy((void *)h_psum,(void *)d_psum,sizeof(float)*m_nsam/2*m_nsam*m_nsam,cudaMemcpyDeviceToHost);
	  cudaMemcpy((void *)h_ksum,(void *)d_ksum,sizeof(int)*m_nsam/2*m_nsam*m_nsam,cudaMemcpyDeviceToHost);
  }
  pthread_mutex_lock(&mutex_getstat1); 
  for(i=0;i<m_impmax*m_nsam/2;i++)
  {
  		sumprel[i]+=h_sumprel[i];
  		nprel[i]+=h_nprel[i];
  }
  if(m_nnstat==1)
  {
	  for(i=0;i<m_nsam/2*m_nsam*m_nsam;i++)
	  {
	  		asum[i]+=h_asum[i];
	  		vsum[i]+=h_vsum[i];
	  		psum[i]+=h_psum[i];
	  		ksum[i]+=h_ksum[i];
	  }
  }
  pthread_mutex_unlock(&mutex_getstat1);

  double *h_std=new double[(m_nsam+1)*(m_nsam+1)];
  cudaMemcpy((void *)h_std,(void *)d_std,sizeof(double)*(m_nsam+1)*(m_nsam+1),cudaMemcpyDeviceToHost); 
  double *h_vtd=new double[(m_nsam+1)*(m_nsam+1)];
  cudaMemcpy((void *)h_vtd,(void *)d_vtd,sizeof(double)*(m_nsam+1)*(m_nsam+1),cudaMemcpyDeviceToHost);   
  long *h_nksum=new long[(m_nsam+1)*(m_nsam+1)];
  cudaMemcpy((void *)h_nksum,(void *)d_nksum,sizeof(long)*(m_nsam+1)*(m_nsam+1),cudaMemcpyDeviceToHost);
  
  pthread_mutex_lock(&mutex_getstat2); 
  for(i=0;i<(m_nsam+1)*(m_nsam+1);i++)
  {
  	std+=h_std[i];
  	vtd+=h_vtd[i];
  	nksum+=h_nksum[i];
  }
  pthread_mutex_unlock(&mutex_getstat2);
     
  double *h_vn=new double[m_nsam];
  cudaMemcpy((void *)h_vn,(void *)d_vn,sizeof(double)*m_nsam,cudaMemcpyDeviceToHost);
  double *h_vvn=new double[m_nsam];
  cudaMemcpy((void *)h_vvn,(void *)d_vvn,sizeof(double)*m_nsam,cudaMemcpyDeviceToHost);
    double *h_vsn=new double[m_nsam];
  cudaMemcpy((void *)h_vsn,(void *)d_vsn,sizeof(double)*m_nsam,cudaMemcpyDeviceToHost);
    double *h_vvsn=new double[m_nsam];
  cudaMemcpy((void *)h_vvsn,(void *)d_vvsn,sizeof(double)*m_nsam,cudaMemcpyDeviceToHost);
    int *h_ic=new int[m_nsam];
  cudaMemcpy((void *)h_ic,(void *)d_ic,sizeof(int)*m_nsam,cudaMemcpyDeviceToHost);
  pthread_mutex_lock(&mutex_getstat3);
  for(i=0;i<m_nsam;i++)
  {
  	vn[i]+=h_vn[i];
  	vvn[i]+=h_vvn[i];
  	vsn[i]+=h_vsn[i];
  	vvsn[i]+=h_vvsn[i];
  	ic[i]+=h_ic[i];
  }
  pthread_mutex_unlock(&mutex_getstat3);
  
  	delete [] h_sumprel;
  	delete [] h_nprel;
  	
   if(m_nnstat==1)
   {
	  	delete [] h_asum;
	  	delete [] h_vsum;
	  	delete [] h_psum;
	  	delete [] h_ksum;
   }
  	
  	delete [] h_std;
	delete [] h_vtd;
	delete [] h_nksum;
	
	delete [] h_vn;
	delete [] h_vvn;
	delete [] h_vsn;
	delete [] h_vvsn;
	delete [] h_ic;
	
	return 0;
}					
					
//free memory in GPU
int cuMem::free()
{
	//free GPU memory
	cudaFree(d_spec);
	cudaFree(d_speq);
	cudaFree(d_datc);
	cudaFree(d_datq);
	cudaFree(d_outc);
	cudaFree(d_outq);
	
	cudaFree(d_ctff);
	cudaFree(d_ctfs);
	cudaFreeArray(cu_ctff);
	
	cudaFree(d_fft2d);
	cudaFree(d_fft2dt);
	cudaFree(d_mbuf);
	cudaFree(d_bf);
	cudaFree(d_sinclut);
	cudaFree(d_dm);
	
	cudaFree(d_wtf);
	cudaFree(d_wtb);
	
	cudaFree(d_mccpart);
	cudaFree(d_msum1);
	cudaFree(d_msum2);
	cudaFree(d_misum);

	cudaFree(d_a3df);
	cudaFree(d_a3ds);
	cudaFree(d_s3df);
	cudaFree(d_s3ds);
		
	cudaFree(d_d3df);
	cudaFree(d_d3ds);
	cudaFree(d_v3df);
	cudaFree(d_v3ds);
	
	//temporary
	//cudaFree(d_c3df);
	//cudaFree(d_c3ds);
	cudaFreeArray(cu_c3df);
	cudaFreeArray(cu_c3ds);
	
	cudaFree(d_sumprel);
	cudaFree(d_nprel);
	if(m_nnstat==1)
  	{
		cudaFree(d_asum);
		cudaFree(d_vsum);
		cudaFree(d_psum);
		cudaFree(d_ksum);
	}
	
	cudaFree(d_vn);
	cudaFree(d_vvn);
	cudaFree(d_vsn);
	cudaFree(d_vvsn);
	cudaFree(d_ic);
	

	return 0;
}
