// CUDA/C++ version of Frealign_v8.06

#include "inc.h"
#include <pthread.h>
#include "lmain_parameters.h"
#include "CreateThread.h"
#include "pagelockedmem.h"

//Macro

//        PARAMETER  (NN1=256,NNBIG=512)
static int NN1=512;     //Max size of particles
static int NNBIG=512;   //Max size of padded reference
static int NNPART=500000;  // max number of particles
static int NNSTAT=0;
//        PARAMETER  (NN1=512,NNBIG=512)
//        PARAMETER  (NN1=1000,NNBIG=1000)
//	    NN1 set to 256 means the program needs about 600 MB memory!

//	  PARAMETER  (NNPART=150000,NSCF=50)

static int NSCF=50;    
    
//        PARAMETER  (MAXSET=20,NDOC1=30,NDOC2=NDOC1+MAXSET)
static int MAXSET=20;
static int NDOC1=30;
static int NDOC2=NDOC1+MAXSET;
//NIKO       MAXSET=100 does not work on some machines....  set back to 20

//        PARAMETER  (PI=3.1415926535897)
//static float PI=3.1415926535897;

//        PARAMETER  (HALFW=6.0)
static float HALFW=6.0;

//        PARAMETER  (ITMAXA=10)
static int ITMAXA=10;    //used in card2, to limit itmax

//        PARAMETER  (DAMAX=360.0*PI/180.0)
static float DAMAX=360.0*PI/180.0;

//        PARAMETER  (ISYMAX=100)
static int ISYMAX=100;

//        PARAMETER  (IANG=80000)
static int IANG=80000;

//        PARAMETER  (NPB=370)
static int NPB=370;

//     	  PARAMETER  (IPMAXX=200)
static int IPMAXX=200;   //used in card2, to limit ipmax 
static int IMPMAX=16;


//global variables for CUDA on GPU
vector<planFFT> planlist;

//mutul lock for image read
pthread_mutex_t mutex;
pthread_mutex_t mutex1;
pthread_mutex_t mutex2;

//mutul lock for get_stat 
pthread_mutex_t mutex_getstat1,mutex_getstat2,mutex_getstat3;

//mutul lock for get_rec 
pthread_mutex_t mutex_getrec11,mutex_getrec12,mutex_getrec13,mutex_getrec14;
pthread_mutex_t mutex_getrec21,mutex_getrec22,mutex_getrec23,mutex_getrec24;

//mutul lock for output in refinement
pthread_mutex_t mutex_oppar;

//thread-global variable for GPU memory
pthread_key_t cuMemKey;

//output.par
vector<list<string> > OPPar;

// Subroutine in Fortran codes
extern "C" 
{
	void card1_(char *, char *, int *, logical *, 
	    		logical *, logical *, logical *, int *, logical *, logical *, 
	    		logical *, logical *, int *, int *, 
	    		int *, int *, int *, int *);
	void card2_(float *, float *, float *, float *, float *,
	     		float *, float *, float *, int *, int *, int *, int *);
	void card3_(int *);
	void card4_(int *, int *, int *); 
	void card5_(char *, char *, int *, int *, float *, logical *, 
	 		   int *, int *, int *, int *);
	void card6_(logical *, int *, int *, float *, float *, float *, float *,
				float *, float *, float *, float *, float *);
	void card7_(int *, int *, float *, float *, float *, float *);
	void cards8and9_(int *, int *, char *,	char *, int *, char *, char *, 
				float *, char *, int *, int *, float *, float *, float *, 
				float *, float *, int *, float *);
	void card10_(int *, int *, char *, int *, int *, float *, float *, 
				float *, float *, float *, int *, float *, float *, float *, 
				float *, float *, int *, int *, float *, int *, float *, 
				int *, float *, int * , int *, int *, int *, int *, 
				float *, int *, float *, float *, float *, float *, float *, 
				float *, float *, float *);
	void cards11and12_(int *, int *, int *,
	     		char *, char *, char *, int *, char *, char *, int *, 
	    		logical *, logical *, logical *, logical *, int *, logical *, 
	    		logical *, logical *, logical *, float *, float *, float *, float *, 
	    		float *, float *, float *, char *, int *, float *, int *, 
	    		int *, float *, int *, int *, int *, int *, 
	    		float *, float *, float *, float *, float *, float *, float *, float *, 
	    		float *, float *, char *, char *, char *, int *, float *, float *,
	     		int *);
	void cards13and14_(char *, char *, logical *, 
	    		logical *, logical *, int *, int *, int *, int *, 
	    		int *, int *, int *, int *, float *, float *, float *, 
	    		float *, CPLX *, CPLX *, float *, CPLX *, CPLX *, 
	    		char *, char *, char *, float *, float *, float *, int *, 
	    		int *, logical *);
	void cards15to18_(char *, char *, char *, char *, char *, char *,	int *);
}

extern "C"
{
	void calcsinc_(float *, int *);
   void searchang_(float *, float *, float *, int *, int *, char *, int *, 
   			float *, int *, int *, int *, logical *, int *);
	
	void storeshift_(double *std, double *vtd, int *jc, int *nsam, int *nsamh, 
			int *iredun, int *nn1,	float *qcp, float *asum, int *ksum, 
			float *psum, float *vsum, float *ssnr, CPLX *a3df, CPLX *b3df, 
			CPLX *c3df, CPLX *d3df,	float *s3df, float *v3df, CPLX *ref3df, 
			CPLX *a3ds, CPLX *b3ds, CPLX *c3ds, CPLX *d3ds, float *s3ds, 
			float *v3ds, CPLX *ref3ds, int *ipad, int *nnstat, int *ifsc);
			
	void correct3d_(int *nsam, float *sinclut, float *a3dv,int *irad);
	
	void mask3d_(int *nsam, float *a3dv, float *rio, float *ric, float *halfw);
	
	void opmaps2_(char *f3d1, char *f3d2, int *i3d1, int *i3d2, int *jc, 
			int *nsam, int *nsamh, float *psize, float *a3dv, float *b3dv, 
			char *cform, char *asym, char *vx, float *sbuf);
	
	void sheltest_(int *nsam, int *istp, CPLX *a3df, CPLX *a3ds, CPLX *d3df, 
			CPLX *d3ds, float *asum, float *psum, int *ksum, float *qcp, int *nstp, 
			float *pr, float *fsc, float *acc, float *qf, float *qc, float *pf, 
			int *ns, int *npix, int *npixt, float *sum, CPLX *cfsc, int *iredun, 
			int *nnstat);
			
	void opresstathalf_(int *ndoc1, double *vsn,	double *vn, float *rbins, 
			float *pf, float *pr, int *istp, int *nstp, int *nsam, float *rlim, 
			float *psize, float *fsc, float *qf, float *qc, float *acc, int *npix, 
			int *npixt, char *cdate, char *ctime, char *czone, int *dtval, 
			char *vx, int *nn1, int *ic, double *vvsn, double *vvn, int *iall,
			int *na, float *sumprel, int *nprel, int *iewald);		
			
	void shiftvol_(int *nsam, double *std, double *vtd, int *jc, float *rlim, 
			int *nsamh, int *nn1, int *iredun, CPLX *a3ds, CPLX *d3ds, CPLX *a3df, 
			CPLX *d3df, float *s3ds, float *v3ds, float *s3df, float *v3df, 
			float *asum, float *qcp, int *nnstat, int *ifsc);
			
	void applycref_(int *nsam, CPLX *a3df, CPLX *a3ds, float *fsc);
	
	void beautify_(int *nsam, int *nsym, float *symop,	int *jsym, float *a3dv, 
			float *b3dv, float *ri, float *halfw, int *imp);
			
	void opmaps_(char *fpoi, char *fpha, int *instat, int *ipstat, int *ipoint, 
			int *inpic, float *psum, int *ndoc1, int *ndoc2, int *nset, int *nn1, 
			int *jc, int *nsam, int *nsamh, float *psize, float *a3dv, float *d3dv, 
			float *s3df, float *s3ds, char *cform, char *asym, char *vx, 
			float *sbuf, int *nnstat, int *ifsc);
	
	void opresstatmaps_(int *nstp, int *istp, int *nsam, float *rlim, 
			float *psize, float *pr, float *fsc, float *acc, int *npix, 
			int *npixt, int *nn1);
}

//******************************************************************************
int main()
{
 //        CHARACTER*200 SFILE,FINPAT1(MAXSET),FINPAT2(MAXSET)
  char sfile[200];
  char finpat1[200*MAXSET];  //path of image stack for nset
  char finpat2[200*MAXSET];  //matching projection image stack
//        CHARACTER*200 FINPAR,F3D,FWEIGH,FQF,FAMP,FPHA,FPOI
  char finpar[200];//name of INPUT PARAMETER FILE
  char f3d[200];   //filename of 3D reconstruction file
  char fweigh[200];//filename of 3D weights file 
  char f3d1[200];   //filename of 3D Q-factor file
  char f3d2[200];  //filename of 3D amplitude STD file 
  char fpha[200];  //filename of 3D ave phase residual file
  char fpoi[200]; //filename of 3D point spread function
//        CHARACTER CFORM*1,ASYMTEST*16,ASYM*3,VX*15
  char cform;            //Input/Output image file forma
  char asymtest[]=" CDTOI0123456789";     // DATA  ASYMTEST/' CDTOI0123456789'/
  char asym[3];         //symbol of the symmetry operation
  char vx[]="8.06 - 19.09.08";    //version of frealign
//        CHARACTER CDATE*8,CTIME*10,CZONE*5,TITLE*1600
  char cdate[8];
  char ctime[10];
  char czone[5];
//  char title[1600];  //Frealign vX.XX - XX.XX.XX, output symmetry +VX+ASYM+DATE
//
//        INTEGER NPIX(NN1/2+1),NPIXT(NN1/2+1),NS(NN1),IC(NN1)
  int npix[NN1/2+1];
  int npixt[NN1/2+1];
  int ns[NN1];
  int ic[NN1];
//        INTEGER KSUM(NN1/2*NN1*NN1),ILST(NPB),IRAD,IRADA
  int *ksum=NULL; //[NN1/2*NN1*NN1];
  int ilst[NPB];
  int irad;     // Padding reference volume, give value 1 below card5
  int irada;    //interpolation radius, get value below cards8and9
//        INTEGER FILM(NNPART),ILIST(NNPART),IREDUN,ISYM(ISYMAX)
  int iredun;        //?? used in card5
  int isym[ISYMAX];  //?? used in card5   Had been moved into thread
//        INTEGER NSANG,NNSET(MAXSET),JSYM(ISYMAX),TNSET(MAXSET)
  int nsang;
  int nnset[MAXSET];
  int jsym[ISYMAX];   //??  used in card5   Had been moved into thread
  int tnset[MAXSET];  //particl number in one data set
//        INTEGER LUNARA(100),MASK(5),DTVAL(8),IPAD,ISTAT
//  int lunara[100];  ////lunara[ifile]=  ?? word length??
  int mask[5];   //used in card3
  int dtval[8];
  int ipad;
  int istat;
//        INTEGER NPREL(NN1/2),IEWALD        F7       !MW for Ewald
//        INTEGER NPREL(IMPMAX*NN1/2),IEWALD         F8      !MW for Ewald
  int nprel[IMPMAX*NN1/2];
  int iewald;
//
//        LOGICAL FMAG,FDEF,FFLIP,FMATCH,ASYMSTORE,FHIST,FALL
  logical fmag;    //magnification refine
  logical fdef;    //defocus refine
  logical fflip;    //Handedness tes
  logical fmatch;    //Write out matching projections after the refinement
  logical asymstore;   //has or hasn't sace the symmetry operation??? card5
  logical fhist;      //Write out history of ITMAX randomisation trials
  logical fall=true;    //if relmag<-99 then fall=false, card6

//        LOGICAL FASTIG,FBEAUT,CI(MAXSET),FCREF
  logical fastig;            //astigmatism refinemen
  logical fbeaut;            //Apply extra real space symmetry averaging and masking
  logical ci[MAXSET];
  logical fcref;             //Apply FOM filter to final reconstruction
//
//        REAL RBIN(NN1*2),TX(MAXSET),TY(MAXSET)
  float rbin[NN1*2];
  float tx[MAXSET];  //Beam tilt [mrad] in X, Y direction
  float ty[MAXSET];  //Beam tilt [mrad] in X, Y direction
//        REAL A3DV(NN1*NN1*NN1),B3DV(NN1*NN1*NN1)
//  float a3dv[NN1*NN1*NN1];
//  float b3dv[NN1*NN1*NN1];
//        REAL C3DV(NNBIG*NNBIG*NNBIG),D3DV(NN1*NN1*NN1)
//  float c3dv[NNBIG*NNBIG*NNBIG];
//  float d3dv[NN1*NN1*NN1];
//        REAL MBUF(NPB*NN1*NN1),PBUF(NPB*NN1*NN1)
  float *mbuf=NULL; //[NPB*NN1*NN1];
  float *pbuf=NULL; //[NPB*NN1*NN1];  //save all images in a film, used in lmain line 236-272
//        REAL S3DF(NN1/2*NN1*NN1),T3DF(NN1/2*NN1*NN1)
  float *s3df=NULL;
  float *bf=NULL;//[5*NN1*NN1];  //New in Frealign 8  Wrong size in pad.f(line 18)
//  float t3df[NN1/2*NN1*NN1];
//        REAL U3DF(NN1/2*NN1*NN1),V3DF(NN1/2*NN1*NN1)
//  float u3df[NN1/2*NN1*NN1];
  float *v3df=NULL;
//        REAL SSNR(NN1/2*NN1*NN1),QCP(NN1/2*NN1*NN1)
  float *ssnr=NULL; //[NN1/2*NN1*NN1];
  float *qcp=NULL; //[NN1/2*NN1*NN1];
//        REAL ASUM(NN1/2*NN1*NN1),VSUM(NN1/2*NN1*NN1),PSUM(NN1/2*NN1*NN1)
  float *asum=NULL; //[NN1/2*NN1*NN1];
  float *vsum=NULL; //[NN1/2*NN1*NN1];
  float *psum=NULL; //[NN1/2*NN1*NN1];
//        REAL S3DS(NN1*NN1),T3DS(NN1*NN1),TESTPAR(6,IPMAXX+1)
  float *s3ds=NULL;
//  float t3ds[NN1*NN1];
  float *testpar=NULL; //[6*(IPMAXX+1)];
//        REAL U3DS(NN1*NN1),V3DS(NN1*NN1)
//  float u3ds[NN1*NN1];
  float *v3ds=NULL;
//        REAL DATA(NN1*NN1),OUTD(NN1*NN1)
  float *data=NULL; //[NN1*NN1];
  float *outd=NULL; //[NN1*NN1];
//        REAL DATD(NN1*NN1),CCD(NN1*NN1),NOIS(NN1*NN1)
  float *datd=NULL; //datd[NN1*NN1];
  float *ccd=NULL; //[NN1*NN1];
  float *nois=NULL; //[NN1*NN1];
//        REAL RBINS(NN1*2),RBINN(NN1*2),RAWBINS(NN1*2)
  float rbins[NN1*2];
  float rbinn[NN1*2];
  float rawbins[NN1*2];
//        REAL PR(NN1/2+1),FSC(NN1/2+1),ACC(NN1/2+1)
  float pr[NN1/2+1];
  float fsc[NN1/2+1];
  float acc[NN1/2+1];
//        REAL QF(NN1/2+1),PF(NN1/2+1),QC(NN1/2+1),RBUF(NN1/2+1)
  float qf[NN1/2+1];
  float pf[NN1/2+1];
  float qc[NN1/2+1];
  float rbuf[NN1/2+1];
//        REAL PSI(NNPART),THETA(NNPART),PHI(NNPART)
 /* float psi[NNPART];
  float theta[NNPART];
  float phi[NNPART];

  int film[NNPART];   //film number of each particle
  int ilist[NNPART];  //particle numbers list  in output_***.par

//        REAL AMAGP(NNPART),PRESA(NNPART),ANGAST(NNPART)
  float amagp[NNPART];  ////relmag[nset-1]*dstep[nset-1]*10000.0/(abs(absmagp)*psize)

  float presa[NNPART];
  float angast[NNPART];

//        REAL DFMID1(NNPART),DFMID2(NNPART),SHX(NNPART),SHY(NNPART)
  float dfmid1[NNPART];
  float dfmid2[NNPART];
  float shx[NNPART];
  float shy[NNPART];
*/
  float *psi=new float[NNPART];
  float *theta=new float[NNPART];
  float *phi=new float[NNPART];

  int *film=new int[NNPART];   //film number of each particle
  int *ilist=new int[NNPART];  //particle numbers list  in output_***.par

//        REAL AMAGP(NNPART),PRESA(NNPART),ANGAST(NNPART)
  float *amagp=new float[NNPART];  ////relmag[nset-1]*dstep[nset-1]*10000.0/(abs(absmagp)*psize)

  float *presa=new float[NNPART];
  float *angast=new float[NNPART];

//        REAL DFMID1(NNPART),DFMID2(NNPART),SHX(NNPART),SHY(NNPART)
  float *dfmid1=new float[NNPART];
  float *dfmid2=new float[NNPART];
  float *shx=new float[NNPART];
  float *shy=new float[NNPART];
//        REAL RELMAG(MAXSET),CS(MAXSET),WL(MAXSET),DSTEP(MAXSET)
  float relmag[MAXSET];  //Relative magnification
  float cs[MAXSET];    //Cs
  float wl[MAXSET];    //wave length
  float dstep[MAXSET];  //Densitometer step size in microns
//        REAL RREC(MAXSET),RMAX1(MAXSET),RMAX2(MAXSET),RBFAC(MAXSET)
  float rrec[MAXSET];   //rrec[nset-1]=psize/rmax1[nset-1];
  float rmax1[MAXSET];  //RMAX1,RMAX2 - Resol. in refinement in Angstroms
  float rmax2[MAXSET];  //same above
  float rbfac[MAXSET];
//        REAL TARGET(MAXSET),THRESH(MAXSET),FSCT(NN1/2)
  float target[MAXSET];  //Target phase residual
  float thresh[MAXSET];  //Phase residual cut-off
  float fsct[NN1/2];  //initial to 0.0 from element 1 to nsam/2  below cards8and9
//        REAL SANG(3,IANG),SYMOP(3,3,ISYMAX),SINCLUT(2000)
  float sang[3*IANG];       //#define IANG 80000
  float symop[3*3*ISYMAX];       //matrix of symmetry operators,#define ISYMAX 100    Had been moved into thread
  float sinclut[2000];      //Table of sin(x)/x. Calculated with calcsinc
  
//        REAL CMIN(MAXSET),CMAX(MAXSET),XM(MAXSET),YM(MAXSET)
  float cmin[MAXSET];
  float cmax[MAXSET];
  float xm[MAXSET];
  float ym[MAXSET];
//        REAL SX(MAXSET),SY(MAXSET)
  float sx[MAXSET];
  float sy[MAXSET];

//        REAL*8 STD,VTD,REF3DV(NN1*NN1*NN1)
  double std;
  double vtd;
//  double ref3dv[NN1*NN1*NN1];
//        REAL SUMPREL(NN1/2)                 F7     !MW for Ewald diff.vector
//        REAL SUMPREL(IMPMAX*NN1/2)             F8         !MW for Ewald diff.vector
  float sumprel[IMPMAX*NN1/2];
//        DOUBLEPRECISION DMMM(10,4),PFREE,VSN(NN1),VN(NN1)
//  double dmmm[10*4];  //dmmm[ifile,n] n=:1-DMIN 2-DMAX 3-DMEAN*N1*N2*N3 4-N1*N2*N3
//  double pfree;
  double vsn[NN1];
  double vn[NN1];
//        DOUBLEPRECISION VVSN(NN1),VVN(NN1)
  double vvsn[NN1];
  double vvn[NN1];
  
  /* Common Block Declarations */
	//#define NNN 256*256*256   //=NN1*NN1*NN1
	//#define NBNBNB 512*512*512  //=NNBIG*NNBIG*NNBIG
  	float *b3dv=NULL; //[NNN];
	float *c3dv=NULL; //[NBNBNB];
//	float s3dv;
//	float v3dv;
//
//        COMPLEX A3DF(NN1/2*NN1*NN1),A3DS(NN1*NN1)
  float *a3dv=NULL;
  #define a3df ((CPLX *)a3dv)
  CPLX *a3ds=NULL;
//        COMPLEX B3DF(NN1/2*NN1*NN1),B3DS(NN1*NN1)
  #define b3df ((CPLX *)b3dv)  //B3D: binary mask for masking input particles (if XSTD < 0)
  CPLX *b3ds=NULL; //[NN1*NN1];
//        COMPLEX C3DF(NNBIG/2*NNBIG*NNBIG),C3DS(NNBIG*NNBIG)
  #define c3df ((CPLX *)c3dv)  //C3D:  reference volume for refinement
  CPLX *c3ds=NULL; //[NNBIG*NNBIG];
//        COMPLEX D3DF(NN1/2*NN1*NN1),D3DS(NN1*NN1)
  float *d3dv=NULL;
  #define d3df ((CPLX *)d3dv)
  CPLX *d3ds=NULL;
//        COMPLEX REF3DF(NN1/2*NN1*NN1),REF3DS(NN1*NN1)
  CPLX *ref3df=NULL; //[NN1/2*NN1*NN1]
  CPLX *ref3ds=NULL; //[NN1*NN1];
//        COMPLEX SPEC(NN1/2*NN1),DATC(NN1/2*NN1)
  #define spec ((CPLX *)data)
  #define datc ((CPLX *)datd)
//        COMPLEX OUTC(NN1/2*NN1),NOIC(NN1/2*NN1),QBUQ(NPB*NN1)
  #define outc ((CPLX *)outd)
  #define noic ((CPLX *)nois)
  CPLX *qbuq=NULL; //[NPB*NN1];
//        COMPLEX CBUF(NN1/2+1),CCC(NN1/2*NN1),QBUC(NPB*NN1/2*NN1)
  CPLX cbuf[NN1/2+1];
  #define ccc ((CPLX *)ccd)
  CPLX *qbuc=NULL; //[NPB*NN1/2*NN1];
//        COMPLEX CCBUF(NN1/2*NN1),CCS(NN1),PBUQ(NPB*NN1)
  CPLX *ccbuf=NULL; //[NN1/2*NN1];
  CPLX *ccs=NULL; //[NN1];
  CPLX *pbuq=NULL; //[NPB*NN1];
//        COMPLEX DATQ(NN1),OUTQ(NN1),NOIQ(NN1),SPEQ(NN1)
  CPLX *datq=NULL; //[NN1];
  CPLX *outq=NULL; //[NN1];
  CPLX noiq[NN1];
  CPLX *speq=NULL; //[NN1];
//        COMPLEX CTFF(NN1*NN1),CTFS(2*NN1)
  CPLX *ctff=NULL; //[NN1*NN1];
  CPLX *ctfs=NULL; //[2*NN1];
//
//        EQUIVALENCE (DATA,SPEC)
//      	EQUIVALENCE (DATD,DATC)
//      	EQUIVALENCE (OUTD,OUTC)
//      	EQUIVALENCE (NOIS,NOIC)
//      	EQUIVALENCE (CCD,CCC)
//      	EQUIVALENCE (A3DV,A3DF)
//      	EQUIVALENCE (B3DV,B3DF)
//      	EQUIVALENCE (C3DV,C3DF)
//      	EQUIVALENCE (D3DV,D3DF)
//      	EQUIVALENCE (REF3DV,REF3DF)
//
//        COMMON /AA/A3DV
//        COMMON /BB/B3DV
//        COMMON /CC/C3DV
//        COMMON /DD/D3DV
//        COMMON /SS/S3DV
//        COMMON /VV/V3DV


//	DATA  IPOINT/3/
//	DATA  I3D1/1/
//	DATA  I3D2/2/
//	DATA  IPSTAT/10/
//	DATA  INSTAT/9/
//	DATA  INPIC/8/
//	DATA  INPROJ/4/
//	DATA  IOPROJ/7/
	
  int ipoint=3;
  int i3d1=1;
  int i3d2=2;
  int ipstat=10;
  int instat=9;
  int inpic=8;
  int inproj=4;
  int ioproj=7;
//      	DATA  FALL/.TRUE./
//  logical fall=true;
//      	DATA  ASYMTEST/' CDTOI0123456789'/
//      	DATA  DATA  VX/'8.06 - 19.09.08'/
//       15 chars 'X.XX - XX.XX.XX' <--<--<--<--<--<--<--<--
//------------------------------------------------------------------------------
	int i,j,k;
	
	int GPUNum;
	vector<int> GPUList;
	vector<int> GPUListMonitor;
   cudaGetDeviceCount(&GPUNum);

	int version;

   //for(i=0;i<GPUNum;i++)
   for(i=GPUNum-1;i>=0;i--)
   {
		cudaDeviceProp prop;
		if(cudaGetDeviceProperties(&prop, i) == cudaSuccess) 
		{
			version= prop.major*10+prop.minor;
			if(version>=13) 
			{
				if(prop.kernelExecTimeoutEnabled==0) GPUList.push_back(i);
				else GPUListMonitor.push_back(i);
				
			}
		}

   }
   
   GPUNum=GPUList.size()+GPUListMonitor.size();
   if(GPUNum>=2)
   {
   	cout<<"You have "<<GPUNum<<" available GPUs!"<<endl;
   	cout<<"And "<<GPUListMonitor.size()<<" of them connect monitor."<<endl;
   }
   else
   {
   	cout<<"Wrong: you must havt at least two GPUs."<<endl;
   	exit(0);
   }
   
	if(cudaSetDevice(GPUList[0])!=cudaSuccess)
   {
   	cout<<"Set CUDA Device #"<<GPUList[0]<<" for main thread failed!"<<endl;
   	exit(0);
   }

  if(cudaSetDeviceFlags(cudaDeviceMapHost)!=cudaSuccess)
   {
   	cout<<"Set CUDA Device Flag failed!"<<endl;
   	exit(0);
   }
  //Varibles not defined in Fortran code
  
  //card1
  int iflag;
  int ifsc;

  //card2
  float ri;   //Radius of reconstruction in Angstroms from centre of particle
  float ric;
  float psize;  //Pixel size in Angstroms required for output map
  float wgh;  //Amplitude contrast for CTF correction (0...1): 0.07 - for ice
  float xstd;  //number of standard deviations above mean for masking 3D model
  float pbc;  //Phase residual / pseudo-B-factor conversion Constant
  float boff;  //average phase residual
  float dangin; //angular step size for the angular search used in modes IFLAG=3,4
  int ipmax;  //number of potential matches in a search
	      //in lsfreconstruction.com always ipmax=1
               //in lsfrefine.com ipmax is set in PARAMETER file
  int itmax;  //number of cycles of randomised search/refinement

  //card4
  int ifirst;  //first particle number
  int ilast;   //last particle number

  //card5
  int nsym;
  int nasym;  //???how many symmetry operation
  //int isym;   //how many fold symmetry (rotation)
  
  //card6
  float akv;

  //cards8and9
//  int inproj;  //ifile, file ID
  int nsam;  //Particle dimensions, ==n1
  int icmp; //=3.0*ri/64 or nsam/64 or 1   ???
  
  //card10


  //cards11and12

  //cards14and15
  float utd;
  float stdd;

	char* ncpus=NULL;getenv("NCPUS");
		
	int imp=0;
	if(ncpus!=NULL)
	{
		sscanf(ncpus,"%d",&imp);
	}
	if(imp<=0) imp=1;

  //Prepare parameters
   int tempint=2000;
	calcsinc_(sinclut,&tempint);

  //CFORM,IFLAG,FMAG,FDEF,FASTIG,FFLIP,IEWALD,FMATCH,FHIST,FBEAUT,FCREF
  card1_(vx,     //version of Frealign
	&cform,      //Input/Output image file format
	&iflag,       //refine or recons option 1,2,3,4,-3,-4
	&fmag,       //Magnification refinement
	&fdef,       // Defocus refinemen
	&fastig,    //astigmatism refinement
	&fflip,      //Handedness test
	&iewald,      //Ewald correction
	&fmatch,     //Write out matching projections after the refinement
	&fhist,      //Write out history of ITMAX randomisation trials
	&fbeaut,     //Apply extra real space symmetry averaging and masking to
                     // beautify final map just prior to output.3Dvol_000.mrc
	&fcref,    //Apply FOM filter to final reconstruction using function 
                     //   SQRT(2.0*FSC/(1.0+FSC))
   &NN1,
   &NNBIG,
   &NNPART,
   &NNSTAT,
   &ifsc,
   &imp);

  card2_(&ri,   //Radius of reconstruction in Angstroms from centre of particle
  	&ric,
	&psize,  //Pixel size in Angstroms required for output map
	&wgh,  //Amplitude contrast for CTF correction (0...1): 0.07 - for ice
	&xstd,  //number of standard deviations above mean for masking 3D model
	&pbc,  //Phase residual / pseudo-B-factor conversion Constant
	&boff,  //average phase residual
	&dangin, //angular step size for the angular search used in modes IFLAG=3,4
	&ipmax,  //number of potential matches in a search
	        //in lsfreconstruction.com always ipmax=1
                //in lsfrefine.com ipmax is set in PARAMETER file
	&itmax,  //number of cycles of randomised search/refinement
	&ITMAXA,  //#define  ITMAXA  10
	&IPMAXX);  //#define  IPMAXX  200 

  card3_(mask);  //MASK - [0/1] parameters to include in refinement [1 1 1 1 1]

  card4_(&ifirst,   //first particle number
	&ilast,    //last particle number
	&NNPART);   //#define NNPART 150000


  card5_(asymtest,  //DATA  ASYMTEST/' CDTOI0123456789'/
	asym,   //symmetry required CnDn,T,O,I,I1,I2 or N (can be zero)
	jsym,  //list of isym(see below)
	&iredun,
	symop,
	&asymstore,   // has or hasn't save???
	&nsym,
	&nasym,  //how many folds symmetry
	isym,    //how many fold symmetry (rotation)
	&ISYMAX);
	
  //-------------------------------------------------------------
  float halfwc=HALFW;
  float wgh1=sqrt(1.0-double(wgh)*double(wgh));  //wgh is Amplitude contrast for CTF correction, card2
  float wgh2=wgh;
  float wc2=0.1;
  irad=1;
  float ameanphas=0.0;
  int nphas=0;
  float rlim=0.0;  //??resolution limit??
  int nang=0;   // number of angles (projections)
  int nset=1;   //current data set number.
  int nsetst=0;
  int nss=1;
  float rmm=0.0;  //?? rmax2
  
  int iquadmax;
  //--------------------------------------------------------------


  while(1)
    {
      card6_(&fall,  //if relmag<-99 then fall=false. 
			&nset,   //dataset number, corresponding to MAXSET
			&MAXSET,   //#define MAXSET 20
			relmag, //Relative magnification to apply to data set NSET
		   dstep,  //Densitometer step size in microns in set NSET
			target,  //Target phase residual in set NSET
			thresh,  //Phase residual cut-off  in set NSET
			cs,  //Cs of the microscope [mm] in set NSET
			&akv,  //Voltage of the microscope [kV], after this becomes [V]
			wl,  // wave length in set NSET
			tx,  //Beam tilt [mrad] in X, Y direction in set NSET
			ty);
	
      if(relmag[nset-1]==0.0 || !fall) break;

      card7_(&nset,
	    &MAXSET,
	    rrec, //Resolution to which the reconstruction is calculated.
	    rmax1, //RMAX1,RMAX2 - Resol. in refinement in Angstroms
	    rmax2,
	    rbfac); //B-factor to apply to particle image projections

      ci[nset-1]=false;
      if(rmax2[nset-1]<10.0) ci[nset-1]=true;

      cards8and9_(&nset,
		 &MAXSET,
		 finpat1,  //path of image stack for nset
		 finpat2,  //matching projection image stack
		 &inproj,  //ifile, file ID
		 &cform,
		 asym,
		 &psize,
		 vx,
		 &nsam,    //Particle dimensions, ==n1
		 &icmp,   //=3.0*ri/64 or nsam/64 or 1   ???
		 &rlim,   //??resolution limit??
		 &rmm,    //?? rmax2
		 rmax1,  //RMAX1,RMAX2 - Resol. in refinement in Angstroms
		 rmax2,
		 rrec,   // rrec[nset-1]=psize/rmax1[nset-1];
		 &NN1,
		 &ri); //Radius of reconstruction in Angstroms from centre of particle
 
 
      
 //     if(irtflg==-1) break;
      ipad=int(NNBIG/nsam);
      if(ipad>4) ipad=4;
      //if(ipad==3) ipad=3;
      //C       determine the need for interpolation with radius IRADA
      //C       based on particle radius RI. To be safe, pretend that 3*RI
      //C       is particle diameter. If volume dimension IPAD*NSAM is 4 times
      //C       that size, IRADA=0 (one point interpolation), if IPAD*NSAM is
      //C       2 times the particle diameter, IRADA=1 (3 x 3 x 3 points 
      //C       interpolation), if IPAD*NSAM still smaller, than use IRADA=2
      //C       (5 x 5 x 5 points interpolation).
      i=int(ipad*nsam/(3*ri));
      if(i<=1) irada=2;
      if(i>=2) irada=1;
      if(i>=3) irada=0;

      if(ipad != 1)
		{
	  		printf("Padding reference volume, IPAD = %d\n",ipad);
		}
      printf("Interpolation radius for reference IRADA = %d\n",irada);
	
      if(nset==1)
		{
		  for(i=1;i<=nsam/2;i++)
	  	  {
	 	     fsct[i]=0.0;
	 	   }
		}
      
      card10_(&nang,   //number of particles
	     &iflag,
	     finpar,  //name of INPUT PARAMETER FILE
	     &NNPART,    //max particle number
	     ilist,  // particle numbers list  in output_***.par
	     psi,  //angle list
	     theta, //angle list
	     phi,   //angle list
	     shx,   //x shift list
	     shy,   //y shift list
	     film,    //film[NNPART], No. of film
	     dfmid1,  //defocus1 list
	     dfmid2,  //defocus2 list
	     angast,  //
	     presa,   //
	     &ameanphas,  // sum of presa<180
	     &nphas,   //number of presa<180
	     &nsam,    // box size
	     amagp,  //relmag[nset-1]*dstep[nset-1]*10000.0/(abs(absmagp)*psize)
	     &nset,    //data set number
	     &psize,  //pixel number
	     &MAXSET,   //max data set number
	     relmag,  //relative magnification
	     &nss,
	     &ifirst,  
	     &ilast,
	     nnset,
	     &nsetst,
	     dstep,  //# Densitometer step size in microns.
	     tnset,   //particle number in data set nset
	     fsct,
	     thresh,
	     cmin,
	     cmax,
	     xm,
	     ym,
	     sx,
	     sy);

       cards11and12_(&NDOC1,
		    &NDOC2,
		    &nset,
		    cdate,
		    ctime,
		    czone,
		    dtval,
		    vx,
		    &cform,
		    &iflag,
		    &fmag,
		    &fdef,
		    &fastig,
		    &fflip,
		    &iewald,
		    &fmatch,
		    &fhist,
		    &fbeaut,
		    &fcref,
		    &ri,
		    &ric,
		    &psize,
		    &wgh,
		    &xstd,
		    &pbc,
		    &boff,
		    asym,
		    &nsym,
		    symop,
		    &ipmax,
		    &itmax,
		    &dangin,
		    &ifirst,
		    &ilast,
		    &nss,
		    &MAXSET,
		    relmag,
		    dstep,
		    target,
		    thresh,
		    cs,
		    &akv,
		    rrec,
		    rmax1,
		    rmax2,
		    rbfac,
		    finpat1,
		    finpat2,
		    finpar,
		    &ISYMAX,
		    tx,
		    ty,
		    &ifsc);

    }

  //C   End of input data cards - only map files are read in after this
  //c----------------------------------------------------------------------------------
  //C        IF ((FSCT(1).EQ.0.0).AND.(IFLAG.NE.0))
  //C     .                 WRITE(*,*)' *** No FSC table - refinement ',
  //C     .                           ' slightly less accurate'

  iflag=abs(iflag);
  if(nang<ilast) ilast=nang;
  nset=nset-1;
  printf("End of input datasets, number of datasets =%d\n",nset);
  if(boff==0)
    {
      if(nphas!=0) ameanphas=ameanphas/nphas;
      boff=ameanphas;
      printf(" Average phase residual used %8.4f\n",boff);
    }

  //c----------------------------------------------------------------------------------
  if(iflag>=3)
    searchang_(&rmm,
	      &ri,
	      &dangin,  //dang
	      &nsang,   //number of angles which will be searched
	      &iquadmax,  //unknown==4
	      asym,
	      &nasym,  
	      sang,  //save the value of phi, theta and psi 
	      &NDOC1,  // file handle
	      &nset,
	      &IANG,   //#define IANG 80000
	      &asymstore,
	      mask);  //not use
  //end of loop for search angle generation.
  //c----------------------------------------------------------------------------------

  float rii=ri+HALFW/2.0;
  float rih=ri-HALFW/2.0;
  if(rih<0) rih=0;
  float ri2=rih*rih;
  float ri3=rii*rii;
  float ri4=ri*ri;
  int iran=-100;

  int  jc=nsam/2+1;
  int nsamh=nsam/2;
  std=0;
  vtd=0;

  if(fall)
    {
		ssnr=new float[nsam/2*nsam*nsam];
		qcp=new float[nsam/2*nsam*nsam];
		asum=new float[nsam/2*nsam*nsam];
		vsum=new float[nsam/2*nsam*nsam];
		psum=new float[nsam/2*nsam*nsam];
		ksum=new int[nsam/2*nsam*nsam];
		memset(ssnr,0,sizeof(float)*nsam/2*nsam*nsam);
		memset(qcp,0,sizeof(float)*nsam/2*nsam*nsam);
		memset(asum,0,sizeof(float)*nsam/2*nsam*nsam);
		memset(vsum,0,sizeof(float)*nsam/2*nsam*nsam);
		memset(psum,0,sizeof(float)*nsam/2*nsam*nsam);
		memset(ksum,0,sizeof(int)*nsam/2*nsam*nsam);
	  
		for(k=1;k<=nsam/2;k++)
		{
			vsn[k]=0;
			vn[k]=0;
			vvsn[k]=0;
			vvn[k]=0;
			ic[k]=0;
		}
	}
 
  //Allocate page-locked memeroy for some big matrix
  PageLockedMem plm;
  cudaError_t PLMerror;
  int errnum=0;

  if(iflag==0)
  {
  	 //######for reconstruction 1 ######
  	 PLMerror=cudaHostAlloc((void **)&plm.pl_a3dv,sizeof(float)*nsam*nsam*nsam,
  						cudaHostAllocPortable|cudaHostAllocMapped);
 	 if(PLMerror!=cudaSuccess) errnum++;
 	 else
 	 {
 	 	 a3dv=plm.pl_a3dv;
 	 	 memset((void *)a3dv,0,sizeof(float)*nsam*nsam*nsam);
 	 }
 	 
 	 PLMerror=cudaHostAlloc((void **)&plm.pl_a3ds,sizeof(float)*nsam*nsam*2,
 	 					cudaHostAllocPortable|cudaHostAllocMapped);
 	 if(PLMerror!=cudaSuccess) errnum++;
 	 else
 	 {
 	 	 a3ds=(CPLX *)plm.pl_a3ds;
  		 memset((void *)a3ds,0,sizeof(float)*nsam*nsam*2);
 	 }
  	
  	 PLMerror=cudaHostAlloc((void **)&plm.pl_s3df,sizeof(float)*nsam*nsam*nsam/2,
  						cudaHostAllocPortable|cudaHostAllocMapped);
  	 if(PLMerror!=cudaSuccess) errnum++;
  	 else
  	 {
  		s3df=plm.pl_s3df;
  		memset((void *)s3df,0,sizeof(float)*nsam*nsam*nsam/2);
  	 }
  	
  	 PLMerror=cudaHostAlloc((void **)&plm.pl_s3ds,sizeof(float)*nsam*nsam,
  						cudaHostAllocPortable|cudaHostAllocMapped);
  	 if(PLMerror!=cudaSuccess) errnum++;
  	 else
 	 {
 		s3ds=plm.pl_s3ds;
  		memset((void *)s3ds,0,sizeof(float)*nsam*nsam);
 	 }
  	
  	 //######for reconstruction 2 ######
 	 PLMerror=cudaHostAlloc((void **)&plm.pl_d3dv,sizeof(float)*nsam*nsam*nsam,
 	 					cudaHostAllocPortable|cudaHostAllocMapped);
  	 if(PLMerror!=cudaSuccess) errnum++;
 	 else
 	 {
  		d3dv=plm.pl_d3dv;
  		memset((void *)d3dv,0,sizeof(float)*nsam*nsam*nsam);
 	 }
  	
    PLMerror=cudaHostAlloc((void **)&plm.pl_d3ds,sizeof(float)*nsam*nsam*2,
 	 					cudaHostAllocPortable|cudaHostAllocMapped);
 	 if(PLMerror!=cudaSuccess) errnum++;
 	 else
  	 {
  		d3ds=(CPLX *)plm.pl_d3ds;
  		memset((void *)d3ds,0,sizeof(float)*nsam*nsam*2);
 	 }
 	 
 	 PLMerror=cudaHostAlloc((void **)&plm.pl_v3df,sizeof(float)*nsam*nsam*nsam/2,
 						cudaHostAllocPortable|cudaHostAllocMapped);
 	 if(PLMerror!=cudaSuccess) errnum++;
 	 else
 	 {
 	 	v3df=plm.pl_v3df;
 		memset((void *)v3df,0,sizeof(float)*nsam*nsam*nsam/2);
    }
  
  	 PLMerror=cudaHostAlloc((void **)&plm.pl_v3ds,sizeof(float)*nsam*nsam,
  						cudaHostAllocPortable|cudaHostAllocMapped);
    if(PLMerror!=cudaSuccess) errnum++;
  	 else
  	 {
  		v3ds=plm.pl_v3ds;
  		memset((void *)v3ds,0,sizeof(float)*nsam*nsam);
  	 }
  }
  //######for others ######
  PLMerror=cudaHostAlloc((void **)&plm.pl_b3dv,sizeof(float)*nsam*nsam*nsam,
  					cudaHostAllocPortable|cudaHostAllocMapped);
  b3dv=plm.pl_b3dv;
  memset(b3dv,0,sizeof(float)*nsam*nsam*nsam);
  if(PLMerror!=cudaSuccess) errnum++;
  
  PLMerror=cudaHostAlloc((void **)&plm.pl_b3ds,sizeof(float)*nsam*nsam*2,
  					cudaHostAllocPortable|cudaHostAllocMapped);
  b3ds=(CPLX *)plm.pl_b3ds;
  memset(b3ds,0,sizeof(float)*nsam*nsam*2);
  if(PLMerror!=cudaSuccess) errnum++;
    
  PLMerror=cudaHostAlloc((void **)&plm.pl_c3dv,sizeof(float)*ipad*nsam*ipad*nsam*ipad*nsam,
  					cudaHostAllocPortable|cudaHostAllocMapped);
  cudaThreadSynchronize();
  c3dv=plm.pl_c3dv;
  memset(c3dv,0,sizeof(float)*ipad*nsam*ipad*nsam*ipad*nsam);
  if(PLMerror!=cudaSuccess) errnum++;
  
  PLMerror=cudaHostAlloc((void **)&plm.pl_c3ds,sizeof(float)*ipad*nsam*ipad*nsam*2,
  					cudaHostAllocPortable|cudaHostAllocMapped);
  cudaThreadSynchronize();
  c3ds=(CPLX *)plm.pl_c3ds;
  memset(c3ds,0,sizeof(float)*ipad*nsam*ipad*nsam*2);
  if(PLMerror!=cudaSuccess) errnum++;
    					
  if(errnum!=0)
  {
  		cout<<"error: Allocate page-locked memory failed!"<<endl;
  		cout<<"error: "<<errnum<<" error(s) was found!"<<endl;
  		exit(0);;
  }

 
 
  //c----------------------------------------------------------------------------------
  cards13and14_(f3d,  //filename of 3D MAP FILE FOR INPUT and OUTPUT
	       fweigh,  //filename of 3D WEIGHTS FILE FOR INPUT and OUTPUT
	       &fmatch,
	       &fdef,
	       &fmag,
	       &inpic,  //file ID of f3d
	       &instat, //file ID of fweigh
	       &iflag,
	       &istat,
	       &nsam,
	       &nsamh,
	       &NN1,
	       &jc,
	       &psize,
	       &utd,
	       &stdd,
	       b3dv,  //binary mask for masking input particles (if XSTD < 0)
	       b3df,
	       b3ds,
	       c3dv,  // reference volume for refinement
	       c3df,
	       c3ds,
	       &cform,
	       asym,
	       vx,
	       &ri,
	       &HALFW,
	       &xstd,
	       &ipad,
	       &iewald,
	       &fall);

  //c----------------------------------------------------------------------------------
  //########################################
  //Because the padded reference may be bigger than the available memory in griaphic card
  //so, the temperory page-locked memory were used in fft
  //meantime, the fft plan will use many gra 
  for(i=0;i<planlist.size();i++)
  {
		cufftDestroy(planlist[i].plan);
  }
  planlist.clear();
  
  if(errnum!=0)
  {
  		cout<<"error: Allocate page-locked memory for c3d failed!"<<endl;
  		cout<<"error: "<<errnum<<" error(s) was found!"<<endl;
  		exit(0);;
  }
  //########################################
  
  
  long nksum=0;
  int  nstart=ifirst;
  int iswitch=0;
  if(ifsc==2) iswitch=1;
  int fflag=0;
  float psis=0;
  float thetas=0;
  float phis=0;
  float dshxs=0;
  float dshys=0;
  float dfmid1s=0;
  float dfmid2s=0;
  float angasts=0;
  float absmags=0;
  float press=0;
  nset=nsetst;
  int iexcl=0;
  int   ibuf=0;
  int iall=0;
  float pall=0;
  float ccave=0;
  int icount=0;
  int iccount=0;
  for(i=0;i<nsam*2;i++)
  {
     rbins[i]=0;
     rbinn[i]=0;
     rawbins[i]=0;
   }
    
  //c--------------------------------------------------------------------------
  cards15to18_(f3d,     //filename of 3D reconstruction file 
	      fweigh, //filename of 3D weights file 
	      f3d1,     //filename of 3D Q-factor file
	      f3d2,    //filename of 3D amplitude STD file 
	      fpha,    //filename of 3D ave phase residual file
	      fpoi,    //filename of 3D point spread function
	      &NDOC1);
  //c--------------------------------------------------------------------------

  
  //C	MAIN LOOP OVER PARTICLES.....
  int na=0;
  
  //Create MultiThread on CPU
  //thread-global variable for GPU memory
  pthread_key_create(&cuMemKey,NULL);
  
  	//avoid to use the GPU attached a monitor
	if(iflag==0) 
	{
		for(i=0;i<GPUListMonitor.size();i++) GPUList.push_back(GPUListMonitor[i]);
	}
	GPUNum=GPUList.size();
  //Prepare parameters for main loop 
  if(iflag==0 && GPUNum%2!=0) GPUNum--;
  int jobstyle,loopstep,kstart,kend;
  int pnumperjob=int(float(ilast-nstart+1)/float(GPUNum)+GPUNum+1)/2*2;
  
  if(iflag!=0)
  {
		OPPar.resize(GPUNum);
		for(i=0;i<GPUNum;i++) 
		{
			OPPar[i].clear();
		}
  }

  lmain_parameters *jobs=new lmain_parameters[GPUNum];
  
  for(i=0;i<GPUNum;i++)
  {	 
  		jobstyle=(iflag==0)?(i%2+1):-1;
  		loopstep=(iflag==0)?GPUNum:1;
  		kstart=(iflag==0)?(nstart+i):(nstart+i*pnumperjob);
  		kend=(iflag==0)?ilast:(nstart+(i+1)*pnumperjob-1);
  		if(kend>ilast) kend=ilast;

  		jobs[i].initial();
  		jobs[i].setReconstructionNum(jobstyle);  //-1-refine; 0-rec1&2; 1-rec1; 2-rec2 
  		jobs[i].setDevice(GPUList[i]);
  		jobs[i].setThreadNum(i, GPUNum);
  		jobs[i].setStep(loopstep);  //step to loop particles for each thread
  		jobs[i].setBlockSize(16,16);
  		jobs[i].set(kstart,kend, &plm, 
  		 		IMPMAX, fdef, fmag, inproj, cform, data, rrec, nsam, spec, shx,
		 		shy, cs, rawbins, wl, wgh1, wgh2, dfmid1, dfmid2, angast, outc, 
				outd, amagp, rih, HALFW, halfwc, ri2, ri3, ri4,datd, datc, 
				b3dv, phi, theta, psi, xstd, iflag, irada, c3df, c3ds, rbins, 
			 	rbinn, nset, noic, nois, presa, sang, nsang,ccd, ccc, iquadmax, 
				NDOC1, NDOC2, mask, rbfac, rmax1, rmax2, iran, psize, dstep, film, 
		 		NN1, NNPART, MAXSET,	fhist, ilist, NPB, itmax, DAMAX, asymstore, target, 
		 		iccount, iswitch, ipad, pall, iall, irad, ri, a3df, a3ds, s3df, 
		 		s3ds, std, d3df, d3ds, ilst, v3df, v3ds, vtd, pbc, boff, 
		 		fflip, asum, vsum, psum, ksum, thresh, nsamh, jc, fflag, ccave, 
		 		icmp, ifirst, ioproj, iewald,NSCF, sfile, nang, wgh, psis, thetas, 
		 		phis, dshxs, dshys,press, iexcl, ibuf, icount, fmatch, dfmid1s, dfmid2s, 
		 		angasts, absmags, nstart, finpat1, finpat2, asym,vx, ilast, nnset, IANG, 
		 		cdate, ctime, czone, dtval, fastig,sinclut, mbuf, pbuf, ctff, ctfs, 
		 		rbin, ccbuf, ccs, datq, outq, noiq, speq, pbuq, qbuc, qbuq, 
		 		istat, tnset, fall, ic, vsn, vn, vvsn, vvn, ci, bf, 
		 		ns, na, fsct, sumprel, nprel, tx, ty, testpar, ipmax, nsym, 
		 		isym, jsym, symop, NNSTAT, nksum, xm, ym, sx, sy, ifsc, 
		 		imp,ISYMAX);
		jobs[i].set_New2DImage();

	}
	cout<<flush;
	//step 3: Run threads
   //some locks for atomic operation
   pthread_mutex_init(&mutex,NULL); 
    
   pthread_mutex_init(&mutex_getstat1,NULL);
   pthread_mutex_init(&mutex_getstat2,NULL);
   pthread_mutex_init(&mutex_getstat3,NULL);
   
   pthread_mutex_init(&mutex_getrec11,NULL);
   pthread_mutex_init(&mutex_getrec12,NULL);
   pthread_mutex_init(&mutex_getrec13,NULL);
   pthread_mutex_init(&mutex_getrec14,NULL);
   pthread_mutex_init(&mutex_getrec21,NULL);
   pthread_mutex_init(&mutex_getrec22,NULL);
   pthread_mutex_init(&mutex_getrec23,NULL);
   pthread_mutex_init(&mutex_getrec24,NULL);

   pthread_t tid;
   vector<pthread_t> tids;
   tids.reserve(GPUNum);
  	int terror;
  	for(i=0;i<GPUNum;i++)
  	{
  		terror=pthread_create(&tid,NULL,CreateThread,(void *)&(jobs[i]));
  		tids.push_back(tid);
  		if(terror!=0)
  		{
			cout<<"error: Failed to create pthread #"<<i<<"! Exit!"<<endl;
      	exit(0);
   	}
   	cout<<"pthread #"<<i<<" (tid="<<tids[i]<<") is created for GPU device #"<<GPUList[i]<<". "<<endl;
  		//if(i==0) sleep(1);
  		sleep(2);
  	}

 /*timeval begin,end;

 for(k=nstart;k<=ilast;k++)
    { iswitch=0;
    cout<<data<<endl;
    cout<<spec<<endl;
    cout<<speq<<endl;
   gettimeofday(&begin,NULL);
      lmain_(&k, &fdef, &fmag, &inproj, &cform, data, rrec, &nsam, spec, shx,
		 shy, cs, rawbins, wl, &wgh1, &wgh2, dfmid1, dfmid2, angast, outc, 
		 outd, amagp, &rih, &HALFW, &halfwc, &ri2, &ri3, &ri4,datd, datc, 
		 b3dv, phi, theta, psi, &xstd, &iflag, &irada, c3df, c3ds, rbins, 
		 rbinn, &nset, noic, nois, presa, sang, &nsang,ccd, ccc, &iquadmax, 
		 &NDOC1, &NDOC2, mask, rbfac, rmax1, rmax2, &iran, &psize, dstep, film, 
		 &NN1, &NNPART, &MAXSET,	&fhist, ilist, &NPB, &itmax, &DAMAX, &asymstore, target, 
		 &iccount, &iswitch, &ipad, &pall, &iall, &irad, &ri, a3df, a3ds, s3df, 
		 s3ds, &std, d3df, d3ds, ilst, v3df, v3ds, &vtd, &pbc, &boff, 
		 &fflip, asum, vsum, psum, ksum, thresh, &nsamh,	&jc, &fflag, &ccave, 
		 &icmp, &ifirst, &ioproj, &iewald,&NSCF, sfile, &nang, &wgh, &psis, &thetas, 
		 &phis, &dshxs, &dshys,&press, &iexcl, &ibuf, &icount, &fmatch, &dfmid1s, &dfmid2s, 
		 &angasts, &absmags, &nstart, finpat1, finpat2, asym,vx, &ilast, nnset, &IANG, 
		 cdate, ctime, czone, dtval, &fastig,sinclut, mbuf, pbuf, ctff, ctfs, 
		 rbin, ccbuf, ccs, datq, outq, noiq, speq, pbuq, qbuc, qbuq, 
		 &istat, tnset, &fall,ic, vsn, vn, vvsn, vvn, ci, bf, 
		 ns, &na, fsct, sumprel,	nprel, tx, ty, testpar, &ipmax, &nsym, 
		 isym, jsym, symop, &NNSTAT, &nksum, xm, ym, sx, sy, &ifsc, 
		 &imp,&GM1);
gettimeofday(&end,NULL);

//cout<<" time is "<<CalTime(begin,end)<<endl;
cout<<k<<" : "<<iswitch<<endl;
    }*/
    
     
   //Join all threads
   void *TReturn;
	for(i=0;i<GPUNum;i++)
	{
   	terror=pthread_join(tids[i],&TReturn);
   	if(terror!=0)
   	{
      	cout<<"error:Thread #"<<i<<" doesn't exit ... "<<endl;
   	}
   	cout<<"Thread #"<<i<<" exit code is "<<(int *)TReturn<<endl;
   }
   //###############################################################
   std=0;
   vtd=0;
   nksum=0;
   for(i=0;i<GPUNum;i++)
   {
		std+=jobs[i].m_std;
		vtd+=jobs[i].m_vtd;
		nksum+=jobs[i].m_nksum;
		
		psis+=jobs[i].m_psis;
  		thetas+=jobs[i].m_thetas;
		phis+=jobs[i].m_phis;
		dshxs+=jobs[i].m_dshxs;
		dshys+=jobs[i].m_dshys;
		dfmid1s+=jobs[i].m_dfmid1s;
		dfmid2s+=jobs[i].m_dfmid2s;
		angasts+=jobs[i].m_angasts;
		absmags+=jobs[i].m_absmags;
		press+=jobs[i].m_press;
		iexcl+=jobs[i].m_iexcl;
		ibuf+=jobs[i].m_ibuf;
		iall+=jobs[i].m_iall;
		pall+=jobs[i].m_pall;
		ccave+=jobs[i].m_ccave;
		icount+=jobs[i].m_icount;
		iccount+=jobs[i].m_iccount;
		for(j=0;j<nsam*2;j++)
		{
			//rbins[j]+=jobs[i].m_rbins[j];
			//rbinn[j]+=jobs[i].m_rbinn[j];
			rawbins[j]+=jobs[i].m_rawbins[j];
		}
		
		jobs[i].free();
	}
	delete [] jobs;
	//###############################################################
	//delete thread-glolbal variable for cuMem
	pthread_key_delete(cuMemKey);
	
	//output refinement results to output_*.par
	int strlen;
	list<string>::iterator it;
	if(iflag!=0)
	{
		for(i=0;i<GPUNum;i++) 
			for(it=OPPar[i].begin();it!=OPPar[i].end();it++)
			{
				tempint=NDOC1+nset;
				strlen=int((*it).length());
				strtofile_(&tempint,(char *)(*it).c_str(),&strlen);
			}
 				fflush_(&tempint);
 	}
 	//iclose_(&inproj);  //had been moved into CreateThread.cu
 	//if(fmatch) iclose_(&ioproj); //had been moved into CreateThread.cu
 	char tempstr[256];
 
	printf(" PARTICLES ABOVE THRESHOLD (NOT INCLUDED) IN SET %3d: %10d\n",nset,iexcl);
	
	if(icount!=0)
	{
		sprintf(tempstr,"C      %8.3f%8.3f%8.3f%8.3f%8.3f%7.0f.      %9.1f%9.1f%8.2f%7.2f",
				sqrt(psis/icount),sqrt(thetas/icount),sqrt(phis/icount),
				sqrt(dshxs/icount),sqrt(dshys/icount),sqrt(absmags/icount),
				sqrt(dfmid1s/icount),sqrt(dfmid2s/icount),sqrt(angasts/icount));
		tempint=NDOC2+nset;
		strlen=94;
		strtofile_(&tempint,tempstr,&strlen);
		fflush_(&tempint);
	}

// 
	if(iall != 0) pall=pall/iall/PI*180;
	printf("%6d PARTICLES FOR OVERALL PHASE RESIDUAL CALCULATION\n       PHASE RESIDUAL (between resolution limits) =%11.6f\n",
	 		iall,pall);
	tempint=6;
	fflush_(&tempint);
	sprintf(tempstr,"C  Total particles, overall phase residual,%6d%11.6f",iall,pall);
	tempint=NDOC1+1;
	strlen=60;
	strtofile_(&tempint,tempstr,&strlen);
 

 
//
	if(iccount != 0)
	{
	 	ccave=ccave/iccount;
	 	printf("\n%6d PARTICLES REFINED, AVERAGE CC %12.8f\n",iccount,ccave);
	 	
	 	tempint=6;
	 	fflush_(&tempint);
	 	
	 	sprintf(tempstr,"C No. of particles refined, ave. CC%6d%12.8f",iccount,ccave);
	 	tempint=NDOC1+1;
	 	strlen=53;
	 	strtofile_(&tempint,tempstr,&strlen);
	}

	if(!fall)
	{
	 	cout<<" Normal termination, no output data-files"<<endl;
	 	//Destroy FFT Plan
		for(i=0;i<planlist.size();i++)
		{
			cufftDestroy(planlist[i].plan);
		}
	 	return 1;
	}
 


//
//			STD=STD/NSAM/NSAM/NSAM		! invalid for low resolution
	 std=std/nksum;
	 std=std*wc2;
	//      	VTD=VTD/NSAM/NSAM/NSAM		! invalid for low resolution
	 vtd=vtd/nksum;
	 vtd=vtd*wc2;
	//
	// S3DF,S3DS : stats file 1
	// V3DF,V3DS : stats file 2
	// A3DF,A3DS : reconstruction 1
	// D3DF,D3DS : reconstruction 2
	// ASUM,---- : sum of amplitudes
	// VSUM,---- : sum of amplitudes**2
	// PSUM,---- : sum of phase residuals
	// KSUM,---- : number of terms contributing to voxel
	//
	//
	ref3df=new CPLX[nsam/2*nsam*nsam];
	ref3ds=new CPLX[nsam*nsam];
	//memset(ref3df,0,sizeof(CPLX)*nsam/2*nsam*nsam));
	//memset(ref3ds,0,sizeof(CPLX)*nsam*nsam));
	 storeshift_(&std, &vtd, &jc, &nsam, &nsamh, &iredun, &NN1, qcp, asum, ksum,
	 			psum, vsum, ssnr, a3df, b3df, c3df, d3df, s3df, v3df, ref3df, a3ds, 
	 			b3ds, c3ds, d3ds, s3ds, v3ds, ref3ds, &ipad, &NNSTAT, &ifsc);
 

	 //added by xueming
	 int istp;
	 int nstp;
 
	 if (ifsc == 0) 
	 {
		// $OMP PARALLEL SECTIONS
		// $OMP SECTION
		tempint=-1;
		rlft3o_((CPLX *)b3dv,b3ds,&nsam,&nsam,&nsam,&tempint);
		correct3d_(&nsam, sinclut, b3dv, &irad);
		mask3d_(&nsam, b3dv, &ri, &ric, &HALFW);
		// $OMP SECTION 
		tempint=-1;
		rlft3o_((CPLX *)c3dv, c3ds, &nsam, &nsam, &nsam, &tempint);
		correct3d_(&nsam, sinclut, c3dv, &irad);
		mask3d_(&nsam, c3dv, &ri, &ric, &HALFW);

		// $OMP END PARALLEL SECTIONS 
		opmaps2_(f3d1, f3d2, &i3d1, &i3d2, &jc, &nsam, &nsamh, &psize, 
				b3dv, c3dv, &cform, asym, vx, rbuf);

		// $OMP PARALLEL SECTIONS 
		// $OMP SECTION 
		tempint=1;
		rlft3o_((CPLX *)b3dv, b3ds, &nsam, &nsam, &nsam, &tempint);
		// $OMP SECTION 
		tempint=1;
		rlft3o_((CPLX *)c3dv, c3ds, &nsam, &nsam, &nsam, &tempint);
		// $OMP END PARALLEL SECTIONS 
		// 	TEST RESOLUTION - ISTP is step size in pixels for resolution statistics
		istp = 1;
		// 	PSUM contains average phase errors - not as informative as SSNR 
		// 	-> PSUM replaced by SSNR. Change back if PSUM needed. Output of 
		// 	shell-averaged values is still PF. 
		sheltest_(&nsam, &istp, b3df, b3ds, c3df, c3ds, asum, ssnr, ksum, qcp,
		 		&nstp, pr, fsc, acc, qf, qc, pf, ns, npix, npixt, rbuf, cbuf,
		 		&iredun, &NNSTAT);
							
		opresstathalf_(&NDOC1, vsn, vn, rawbins, pf, pr, &istp, &nstp, &nsam, 
				&rlim, &psize, fsc, qf, qc, acc, npix, npixt, cdate, ctime, 
				czone, dtval, vx, &NN1, ic, vvsn, vvn, &iall, &na, sumprel,
				nprel, &iewald);
					
    } 
    else 
    {
		printf("\n *** IFSC set to%2d to save memory: FSC statistics not done ***",ifsc);
		tempint=6;
		fflush_(&tempint);
	 }
	 
	 shiftvol_(&nsam, &std, &vtd, &jc, &rlim, &nsamh, &NN1, &iredun, a3ds, 
	    		d3ds, a3df, d3df, s3ds, v3ds, s3df, v3df, asum, qcp, &NNSTAT, &ifsc);
	    		
    if (fcref && ifsc == 0) 
    {
		// $OMP PARALLEL SECTIONS 
		// $OMP SECTION
		applycref_(&nsam, a3df, a3ds, fsc);
		// $OMP SECTION
		applycref_(&nsam, d3df, d3ds, fsc);
		// $OMP END PARALLEL SECTIONS 
    }
	 // $OMP PARALLEL SECTIONS
	 // $OMP SECTION 
	 tempint=-1;
    rlft3o_((CPLX *)a3dv, a3ds, &nsam, &nsam, &nsam, &tempint);
    correct3d_(&nsam, sinclut, a3dv, &irad);
	 
	 //$OMP SECTION 
    if (ifsc == 0) 
    {
    	tempint=-1;
		rlft3o_((CPLX *)d3dv, d3ds, &nsam, &nsam, &nsam, &tempint);
		correct3d_(&nsam, sinclut, d3dv, &irad);
    }
	 // $OMP END PARALLEL SECTIONS
	 
    if (fbeaut) 
    {
		beautify_(&nsam, &nsym, symop, jsym, a3dv, b3dv, &ri, &HALFW, &imp);
    }
    mask3d_(&nsam, a3dv, &ri, &ric, &HALFW);
    
    opmaps_(fpoi, fpha, &instat, &ipstat, &ipoint, &inpic, psum, &NDOC1, 
    		&NDOC2, &nset, &NN1, &jc, &nsam, &nsamh, &psize, a3dv, d3dv, 
    		s3df, s3ds, &cform, asym, vx, rbuf, &NNSTAT, &ifsc);
    		
    if (NNSTAT != 0) 
    {
	 	//	NOW CALCULATE BRIEF STATISTICS BETWEEN STORED REFERENCE TRANSFORM 
	 	// REF3D AND THE COMPLETE MERGED NEW DATA A3DF VERSUS RESOLUTION 
	 	tempint=1;
		rlft3o_((CPLX *)a3dv, a3ds, &nsam, &nsam, &nsam, &tempint);
		sheltest_(&nsam, &istp, a3df, a3ds, ref3df, ref3ds, asum, ssnr, ksum, 
				qcp, &nstp, pr, fsc, acc, qf, qc, pf, ns, npix, npixt, rbuf, 
				cbuf, &iredun,&NNSTAT); //missing a para NNSTAT in Frealign!!!
		opresstatmaps_(&nstp, &istp, &nsam, &rlim, &psize, pr, fsc, acc, npix,
		 		npixt, &NN1);
    } 
    else 
    {
    	puts("\n *** NNSTAT set to 0 to save memory: some statistics not done ***");
    	tempint=6;
		fflush_(&tempint);
    }
    
    date_and_time__(cdate, ctime, czone, dtval);
    printf(" Final date/time %c%c-%c%c-%c%c%c%c/%c%c:%c%c:%c%c%c%c\n",
   	 cdate[6],cdate[7],cdate[4],cdate[5],cdate[0],cdate[1],cdate[2],cdate[3],
    	 ctime[0],ctime[1],ctime[2],ctime[3],ctime[4],ctime[5],ctime[6],ctime[7]);
    tempint=6;
	 fflush_(&tempint);

  delete [] psi;
  delete [] theta;
  delete [] phi;
  delete [] film;
  delete [] ilist;
  delete [] amagp;
  delete [] presa;
  delete [] angast;
  delete [] dfmid1;
  delete [] dfmid2;
  delete [] shx;
  delete [] shy;
	
	//Destroy FFT Plan
	for(i=0;i<planlist.size();i++)
	{
		cufftDestroy(planlist[i].plan);
	}

	delete [] ref3df;
	delete [] ref3ds;
	
	if(fall)
	{
		delete [] ssnr;
		delete [] qcp;
		delete [] asum;
		delete [] vsum;
		delete [] psum;
		delete [] ksum;
	}
	
	puts(" NORMAL TERMINATION OF FREALIGN");
/*
*/
  
  return 0;
}
