#include "maskcos_gpu.h"
extern pthread_key_t cuMemKey;

//the block size must be 512 and image size be less than 512
__global__ void calculate2_sum(float *outd, float *dsum, int *disum,
								int nsam, float ri2p, float ri3p)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int j=blockIdx.y*blockDim.y+threadIdx.y+1;
	int ii,jj,id;
	int sid=threadIdx.x*blockDim.y+threadIdx.y;
	float crad2;
	
	__shared__ float sum[256];
	__shared__ int isum[256];
	
	sum[sid]=0.0f;
	isum[sid]=0;
	
	if(i<=nsam && j<=nsam)
	{	
		ii=i;
		if(ii>(nsam/2)) ii-=nsam;	
		jj=j;
		if(jj>(nsam/2)) jj-=nsam;	
		id=j+nsam*(i-1); 
		crad2=pow(float(jj-1),2)+pow(float(ii-1),2);
	
		if(crad2>ri2p && crad2<=ri3p)
		{
			sum[sid]=outd[id-1];
			isum[sid]=1;
		}
	}
	__syncthreads();
	
	//tree-type sum
//	if(sid<256)
//	{
//		sum[sid]+=sum[sid+256];
//		isum[sid]+=isum[sid+256];
//	}
//	__syncthreads();
	if(sid<128)
	{
		sum[sid]+=sum[sid+128];
		isum[sid]+=isum[sid+128];
	}
	__syncthreads();
	if(sid<64)
	{
		sum[sid]+=sum[sid+64];
		isum[sid]+=isum[sid+64];
	}
	__syncthreads();
	if(sid<32)
	{
		sum[sid]+=sum[sid+32];
		isum[sid]+=isum[sid+32];
	}
	__syncthreads();
	if(sid<16)
	{
		sum[sid]+=sum[sid+16];
		isum[sid]+=isum[sid+16];
	}
	__syncthreads();
	if(sid<8)
	{
		sum[sid]+=sum[sid+8];
		isum[sid]+=isum[sid+8];
	}
	__syncthreads();
	if(sid<4)
	{
		sum[sid]+=sum[sid+4];
		isum[sid]+=isum[sid+4];
	}
	__syncthreads();
	if(sid<2)
	{
		sum[sid]+=sum[sid+2];
		isum[sid]+=isum[sid+2];
	}
	__syncthreads();
	if(sid<1)
	{
		sum[sid]+=sum[sid+1];
		isum[sid]+=isum[sid+1];
		
		sid=blockIdx.x*gridDim.y+blockIdx.y;
		dsum[sid]=sum[0];
		disum[sid]=isum[0];
	}
	__syncthreads();
}

__global__ void calsumOnGPU2(float *dsum, int *disum)
{
	int sid=threadIdx.x;
	int tid=blockIdx.x*blockDim.x+threadIdx.x;
	__shared__ float sum[256];
	__shared__ int isum[256];
	sum[sid]=dsum[sid];
	isum[sid]=disum[sid];
	__syncthreads();
	
	//tree-type sum
//	if(sid<256)
//	{
//		sum[sid]+=sum[sid+256];
//		isum[sid]+=isum[sid+256];
//	}
//	__syncthreads();
	if(sid<128)
	{
		sum[sid]+=sum[sid+128];
		isum[sid]+=isum[sid+128];
	}
	__syncthreads();
	if(sid<64)
	{
		sum[sid]+=sum[sid+64];
		isum[sid]+=isum[sid+64];
	}
	__syncthreads();
	if(sid<32)
	{
		sum[sid]+=sum[sid+32];
		isum[sid]+=isum[sid+32];
	}
	__syncthreads();
	if(sid<16)
	{
		sum[sid]+=sum[sid+16];
		isum[sid]+=isum[sid+16];
	}
	__syncthreads();
	if(sid<8)
	{
		sum[sid]+=sum[sid+8];
		isum[sid]+=isum[sid+8];
	}
	__syncthreads();
	if(sid<4)
	{
		sum[sid]+=sum[sid+4];
		isum[sid]+=isum[sid+4];
	}
	__syncthreads();
	if(sid<2)
	{
		sum[sid]+=sum[sid+2];
		isum[sid]+=isum[sid+2];
	}
	__syncthreads();
	if(sid<1)
	{
		sum[sid]+=sum[sid+1];
		isum[sid]+=isum[sid+1];	
		
		dsum[tid]=sum[sid];
		disum[tid]=isum[sid];	
	}
	__syncthreads();
	if(tid==0)
	{
		dsum[0]+=dsum[256]+dsum[512]+dsum[768];
		disum[0]+=disum[256]+disum[512]+disum[768];
		
		if(disum[0]!=0)	dsum[0]=dsum[0]/disum[0];
	}
	__syncthreads();
}

__global__ void calculate2_cosedge(float *outd, float *dsum, int nsam, float rihp,
							float ri2p, float ri3p, float halfw)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int j=blockIdx.y*blockDim.y+threadIdx.y+1;
	float sum=dsum[0];
	int ii,jj,id;
	float edge;
	float PI=3.1415926535897;
	
	if(i>nsam || j>nsam) return;
	
	ii=i;
	if(ii>(nsam/2)) ii-=nsam;
	jj=j;
	if(jj>(nsam/2)) jj-=nsam;
	
	id=j+nsam*(i-1); 
	edge=pow(float(ii-1),2)+pow(float(jj-1),2);
	
	if(edge>ri2p && edge<=ri3p)
	{
		edge=sqrt(edge);
		edge=(1.0f+cos(PI*(edge-rihp)/halfw))/2.0f;
		outd[id-1]=outd[id-1]*edge+(1.0-edge)*sum;				
	}
	else if(edge>ri3p)
	{
		outd[id-1]=sum;
	}

}

void maskcos2(int nsam, float *outd, float ri2, float ri3, 
					float rih, float halfw, float amagp)
{
	//Applies cosine-edged mask to 2D image
	//size of outd has been changed to nsam*(nsam+2)
	
	float rihp, ri2p, ri3p;
	
	dim3 dimBlock(16,16);
	dim3 dimGrid(nsam/dimBlock.x+1,nsam/dimBlock.y+1);
	
	if(nsam>512) printf("error:  maskcos2 will be wrong.");
	
	rihp=rih/amagp;
	ri2p=ri2/amagp/amagp;
	ri3p=ri3/amagp/amagp;

	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
	float *dsum=NULL;
	int *disum=NULL;	
	dsum=pGM->d_msum1;
	cudaMemset(dsum,0,sizeof(float)*1024);   //boxsize can't be bigger than 512*512
	disum=pGM->d_misum;
	cudaMemset(disum,0,sizeof(int)*1024);   //boxsize can't be bigger than 512*512
	
	cudaThreadSynchronize();
	
	calculate2_sum<<<dimGrid,dimBlock>>>(outd, dsum, disum, nsam, ri2p, ri3p);
	cudaThreadSynchronize();
	
	calsumOnGPU2<<<4,256>>>(dsum, disum);
	cudaThreadSynchronize();
	
	//MASK PARTICLE, COSINE EDGE
	calculate2_cosedge<<<dimGrid,dimBlock>>>(outd, dsum, nsam, rihp, ri2p, ri3p, halfw);
	cudaThreadSynchronize();
		
}
