#include "pinsert_gpu.h"
#include "pinsert_s_gpu.h"

void matmul_t(float *a, float *b, float *c)  // c=a*b
{
	int i,j;
	float t[9];
	
	for(i=0;i<3;i++)
		for(j=0;j<3;j++)
		{
			t[i*3+j]=a[i*3]*b[j*3]+a[i*3+1]*b[j*3+1]+a[i*3+2]*b[j*3+2];
		}

	for(i=0;i<9;i++)
	{
		c[i]=t[i];
	}
	
}

void pinsert(int &nsam, int &irad, float &ri, float &rrec, 
		float &amag, CPLX *spec, CPLX *speq, CPLX *ctff, 
		CPLX *ctfs, CPLX *a3df, CPLX *a3ds, float *s3df, float *s3ds, 
		CPLX *b3df, CPLX *b3ds, float *t3df, float *t3ds, double *ttd,
		 float &phi, float &theta, float &psi, float &shx, float &shy, float &presa,
		 float &pbc, float &boff, float *asum, float *vsum, float *psum, 
		 CPLX *c3df, int *ksum, int isign, float *sinclut, int &irada, 
		int &ipad, float thet, int &iewald, CPLX *c3ds, float *sumprel, 
		int *nprel, int &nsym, int *isym, int *jsym, 
		float *symop, int &nnstat, long *nksum, int &imp, cuMem *pGM)
{
	int i,j,ic,jc,irad2;
	int l,nsamh,id=0,is,irrec,irrec2;
	float dm[9],tm[9],st;
	float cphi,sphi;
	float cthe,sthe,cpsi,spsi;
	float rii,pref,tmp;

	//CPLX samp,bsamp;
	//float tlim=2.0,ctfv2,ctfv2s2,phase,pshft,wgt;
	
	pref=pbc/180*PI;
	jc=nsam/2+1;
	ic=nsam*nsam/2;
	nsamh=nsam/2;
	rii=2*PI*ri/nsam;
	irrec=int(float(nsam)*rrec);

	if (irrec > nsamh) 
	{
		irrec=nsamh;
   }

    irrec2=irrec*irrec;
    cphi = cos(phi);
    sphi = sin(phi);
    cthe = cos(theta);
    sthe = sin(theta);
    cpsi = cos(psi);
    spsi = sin(psi);
    
    irad2 = irad*irad;
    tmp = -(presa/pref-boff/pbc);
    if (tmp > 2.f) tmp = 2.f;
    tmp/=(float(nsamh)*float(nsamh));

    tm[0] = (cphi * cthe * cpsi - sphi * spsi) / abs(amag);
    tm[1] = (sphi * cthe * cpsi + cphi * spsi) / abs(amag);
    tm[2] = -sthe * cpsi / abs(amag);
    tm[3] = (-cphi * cthe * spsi - sphi * cpsi) / abs(amag);
    tm[4] = (-sphi * cthe * spsi + cphi * cpsi) / abs(amag);
    tm[5] = sthe * spsi / abs(amag);
//    MW uncommented for 3rd dimension
    tm[6] = sthe * cphi / abs(amag);
    tm[7] = sthe * sphi / abs(amag);
    tm[8] = cthe / abs(amag);

    if (isign < 0) 
    {
		for(l=0;l<nsam;l++) 
	    for (j=0;j<nsam;j++) 
	    {
			is = j + nsam * i;
			b3ds[is] = a3ds[is];
			for (i=0;i<nsam/2;i++) 
				{
		    		id = i+nsamh*(j+nsam*l);
		    		b3df[id] = a3df[id];
				}
	    }
    }


    for (i=0;i<nsym;i++) 
    {
		isym[i] = 1;
    }

//L900:
	 while(1)
	 {
	    for (i=0; i<9; i++) 
 	    {
			dm[i]=tm[i];
 	    }
    
 	    for (i=0;i<nsym;i++) 
 	    {
			if (isym[i] != jsym[i]) 
			{
		    	for (j = 0; j< isym[i]; j++) 
		    	{ 
					matmul_t(dm, &symop[i*9],dm);
		    	}
		 	}
 	    }

  	  	 st=float(2*jc-1)/float(imp);
		 // $OMP PARALLEL DO,REDUCTION(+:NKSUM,TTD)

  	  	 for (i=1;i<=imp;i++) 
  	    { 
			 	pinsert_s(nsam, irad, rii, irrec2, spec, speq, ctff, 
					ctfs, b3df, b3ds, t3df, t3ds, ttd, shx, 
					shy, asum, vsum, psum, c3df, ksum, isign, sinclut,
					irada, ipad, thet, iewald, c3ds, &sumprel[id], 
					&nprel[id], nnstat, nksum, jc, dm, tmp, nsamh, ic, 
					irad2, st, i, pGM);
 	    }

  	    for (i=1; i<imp;i++) 
  	    {
			  id=i*nsam/2;
			  for(j=0;j<nsam/2;j++) 
			  {
		      	sumprel[j]+=sumprel[id+j];
		      	nprel[j]+=nprel[id+j];
			  }
       }

 	    i=0;
       
 	    while(1) //L990:
 		 {		
  		  	  isym[i]++;
  		  	  if(isym[i]>jsym[i]) 
   		  {
					isym[i] = 1;
					i++;
	
					if (i<nsym) continue; //goto L990;
					return;
  			  } 
  			  else
  			  {
  			  		i=0;
  			  		break;
  			  }
  			  
  		  }
  		  
//   	 else 
 //  	 {
//			i=1;    //useless
//			goto L900;
 //  	 }
    }
}
