#include "presb_gpu.h"
#include "maskcos2_gpu.h"
#include "cuMem.h"
extern pthread_key_t cuMemKey;

/*__device__ inline void atomicFloatAdd(float *address, float val)
{
      int i_val = __float_as_int(val);
      int tmp0 = 0;
      int tmp1;

      while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0)
      {
              tmp0 = tmp1;
              i_val = __float_as_int(val + __int_as_float(tmp1));
      }
}*/
//******************
static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

static __device__ float cabs(CPLX f)
{
	return sqrt(f.x*f.x+f.y*f.y);
}

static __device__  CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}

static __device__ float boxft_lut(float *arg,float *sinclut)
{
	int i,j;
	float sinc,prod=1.0f;
	
	for(i=0;i<3;i++)
	{
		arg[i]=abs(arg[i]);
		if(arg[i]<0.02f)
		{
			sinc=1.0f;
		}
		else
		{
			j=int(180.0f*arg[i]+0.5f);
			sinc=sinclut[j-1];
		}
		prod=prod*sinc;
	}
	return prod;

}


static __device__ CPLX ainterpo3dbig(int nsama, int ipad, CPLX *a3df, CPLX *a3ds, 
                                      float &x, float &y, float &z)
{
	int nsam,jc;
	int l,m,n,ll,mm,nn,id,nsamh;
	CPLX samp;
	
	nsam=ipad*nsama;
	samp.x=0.0f;
	samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	l=(x>=0.0f)?int(ipad*x+0.5f):int(ipad*x-0.5f);
	m=(y>=0.0f)?int(ipad*y+0.5f):int(ipad*y-0.5f);
	n=(z>=0.0f)?int(ipad*z+0.5f):int(ipad*z-0.5f);
	
	ll=abs(l)+1;
	mm=abs(m)+1;
	if(mm<1) mm=mm+nsam;
	nn=abs(n)+1;
	if(nn<1) nn=nn+nsam;
	
	if(ll!=jc)
	{
		id=ll+nsamh*((mm-1)+nsam*(nn-1));
		samp=cPc(samp,(l>=0)?a3df[id-1]:conjg(a3df[id-1]));
	}
	else
	{
		id=mm+nsam*(nn-1);
		samp=cPc(samp,(l>=0)?a3ds[id-1]:conjg(a3ds[id-1]));
	}

	return samp;
}    

/*if(l>=0)
	{
		ll=l+1;
		mm=m+1;
		if(mm<1) mm=mm+nsam;
		nn=n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,a3df[id-1]);
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,a3ds[is-1]);
		}
	}
	else
	{
		ll=-l+1;
		mm=-m+1;
		if(mm<1) mm=mm+nsam;
		nn=-n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,conjg(a3df[id-1]));
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,conjg(a3ds[is-1]));
		}
	}*/       

static __device__ CPLX ainterpo3ds(int nsam, int irad, CPLX *a3df, CPLX *a3ds, 
                        float &xx, float &yy, float &zz, float *sinclut, int &ipad)
{
	int ls,lt,ns,nt,ms,mt,jc;
	int l,m,n,ll,mm,nn,is,id,nsamh;
	float x,y,z,arg[3],boxftv;
	CPLX samp;
	
	samp.x=samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	x=ipad*xx;
	y=ipad*yy;
	z=ipad*zz;
	
	ls=int(x)-irad+1;
	if(x<0.0f) ls=ls-1;
	if(ls<(-jc)) ls=-jc+1;
	lt=int(x)+irad;
	if(x<0.0f) lt=lt-1;
	if(lt>=jc) lt=jc-1;
	
	ms=int(y)-irad+1;
	if(y<0.0f) ms=ms-1;
	if(ms<=(-jc)) ms=-jc+1;
	mt=int(y)+irad;
	if(y<0.0f) mt=mt-1;
	if(mt>=jc) mt=jc-1;
	
	ns=int(z)-irad+1;
	if(z<0.0f) ns=ns-1;
	if(ns<=(-jc)) ns=-jc+1;
	nt=int(z)+irad;
	if(z<0.0f) nt=nt-1;
	if(nt>=jc) nt=jc-1;
	
	for(l=ls;l<=lt;l++)
	{
		arg[0]=x-l;
		for(m=ms;m<=mt;m++)
		{
			arg[1]=y-m;
			for(n=ns;n<=nt;n++)
			{
				arg[2]=z-n;
				boxftv=boxft_lut(arg,sinclut);
				if(l>=0)
				{
					ll=l+1;
					mm=m+1;
					if(mm<1) mm=mm+nsam;
					nn=n+1;
					if(nn<1) nn=nn+nsam;
					if(ll!=jc)
					{
						id=ll+nsamh*((mm-1)+nsam*(nn-1));
						samp=cPc(samp,cXr(a3df[id-1],boxftv));
					}
					else
					{
						is=mm+nsam*(nn-1);
						samp=cPc(samp,cXr(a3ds[is-1],boxftv));
					}
				}
				else
				{
					ll=-l+1;
					mm=-m+1;
					if(mm<1) mm=mm+nsam;
					nn=-n+1;
					if(nn<1) nn=nn+nsam;
					if(ll!=jc)
					{
						id=ll+nsamh*((mm-1)+nsam*(nn-1));
						samp=cPc(samp,conjg(cXr(a3df[id-1],boxftv)));
					}
					else
					{
						is=mm+nsam*(nn-1);
						samp=cPc(samp,conjg(cXr(a3ds[is-1],boxftv)));
					}
				}
				
			}
		}
	}
	
	return samp;
}           

static __device__ CPLX ewaldex(int &nsam, int &irad, CPLX *a3df, CPLX *a3ds, 
             float *sinclut, int &ipad,
             int &i, int &j,float *dm, float &thet, CPLX &ctfr, CPLX &ctfl)
{
	float x3,y3,z3;
	float thetah,gpix,x,y,z;
	CPLX ainpo1,ainpo2;
	
	gpix=sqrt(float(i*i+j*j));  //length of resol.vector g in pixel
	thetah=gpix*thet/2;       //THET=(WL/(PSIZE*NSAM))/AMAGP
	x=i*cos(thetah);          //THETAH=scattering angle/2.
	y=j*cos(thetah);
	z=gpix*sin(thetah);
	x3=dm[0]*x+dm[3]*y+dm[6]*z;
	y3=dm[1]*x+dm[4]*y+dm[7]*z;
	z3=dm[2]*x+dm[5]*y+dm[8]*z;
	
	if(irad==0) //AINPO1 is first beam
	{
		ainpo1=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
	}
	else
	{
		ainpo1=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
	}
	
	x3=dm[0]*x+dm[3]*y-dm[6]*z;  //AINPO2 is second beam = conjg(x,y,-z)
	y3=dm[1]*x+dm[4]*y-dm[7]*z;
	z3=dm[2]*x+dm[5]*y-dm[8]*z;
	if(irad==0)
	{
		ainpo2=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
	}
	else
	{
		ainpo2=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
	}
	return cPc(cXc(ainpo1,ctfr),cXc(ainpo2,conjg(ctfl)));  //combine beams
}

//******************

__global__ void CalPresb_1(float *buftd, CPLX *ctff, CPLX *ctfs, CPLX *c3df, CPLX *c3ds, 
						float *sinclut, int nsam, int jc, int ic, 
						int iewald, int irad, int ipad, float scal,float *dm, float thetatr)
{
	//for(i=0;i<=(jc-1);++i)
	//for(j=(-jc+1);j<=(jc-1);++j)
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;
	int ii,jj,ib,id,itemp;
	float x3,y3,z3;
	CPLX ctfr,ctfl,ainpo;
	
	if(i>(jc-1) || j>(jc-1)) return;
	
	ii=i+1;
	itemp=i*i+j*j;
	if(!(itemp>=1 && itemp<ic/2)) return;

	jj=j+1;
	if(jj<1) jj+=nsam;
	if(ii!=jc)
	{
		id=ii+nsam/2*(jj-1);
		ctfr=ctff[id-1];
		ctfl=ctff[id+ic-1];
	}
	else
	{
		ctfr=ctfs[jj-1];
		ctfl=ctfs[jj+nsam-1];
	}
	//MW call ewald sphere corrected extraction
	if(iewald!=0)
	{
		if(iewald<0)
		{
			ctfr=conjg(ctfr);
			ctfl=conjg(ctfl);
		}
		ainpo=ewaldex(nsam,irad,c3df,c3ds,sinclut,ipad,
							i,j,dm,thetatr,ctfr,ctfl);
		ainpo=cXc(ainpo,cPc(ctfr,conjg(ctfl)));
	}	
	else
	{
		x3=dm[0]*i+dm[3]*j;
		y3=dm[1]*i+dm[4]*j;
		z3=dm[2]*i+dm[5]*j;
		if(irad==0)
		{
			ainpo=ainterpo3dbig(nsam,ipad,c3df,c3ds,x3,y3,z3);
		}
		else
		{
			ainpo=ainterpo3ds(ipad*nsam,irad,c3df,c3ds
							,x3,y3,z3,sinclut,ipad);
		}	
		ainpo=cXr(ainpo,pow(cabs(cPc(ctfr,conjg(ctfl))),2)); 	
	}
				
	if(ii!=jc)
	{
		ib=2*id;
		buftd[ib-2]=ainpo.x/scal;
		buftd[ib-1]=ainpo.y/scal;
	}
	else
	{
		ib=nsam*nsam+2*jj;
		buftd[ib-2]=ainpo.x/scal;
		buftd[ib-1]=ainpo.y/scal;
	}

}


__global__ void CalPresb_2(float *buftd, float *tempbuf, CPLX *outc, CPLX *outq, 
					float shx, float shy, int nsam, int jc, int maxr)
{
	//for(i=0;i<=(jc-1);++i)
	//for(j=(-jc+1);j<=(jc-1);++j)
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;	
	int ii,jj,itemp;
	float phase;
	int id,ib,ibin;

	if(i>(jc-1) || j>(jc-1)) return;
			
	itemp=i*i+j*j;
	
	if(!(itemp>=1 && itemp<maxr)) return;

	ii=i+1;
	jj=j+1;
	if(jj<1) jj+=nsam;

	phase=shx*i+shy*j;
	CPLX pshft,samp,ainpo;
	pshft.x=cos(phase);
	pshft.y=sin(phase);
				
	if(ii!=jc)
	{
		id=ii+nsam/2*(jj-1);
		samp=cXc(outc[id-1],pshft);

		ib=2*id;
		ainpo.x=buftd[ib-2];
		ainpo.y=buftd[ib-1];
	}
	else
	{
		samp=cXc(outq[jj-1],pshft);
		
		ib=nsam*nsam+2*jj;
		ainpo.x=buftd[ib-2];
		ainpo.y=buftd[ib-1];
	}

	ibin=int(sqrt(float(itemp))+0.5f);
	//bufab[ibin]+=(cXc(samp,conjg(ainpo))).x;
	//bufab[ibin]+=pow(cabs(samp),2);
	//bufab[ibin]+=pow(cabs(ainpo),2);
	//bufab[ibin]+=1.0f;
	
	ib=(nsam+1)*i+(j+nsam/2);
	ii=(nsam/2+1)*(nsam+1);
	*((int *)&tempbuf[ib])=ibin;
	tempbuf[ii*1+ib]=(cXc(samp,conjg(ainpo))).x;
	tempbuf[ii*2+ib]=pow(cabs(samp),2);
	tempbuf[ii*3+ib]=pow(cabs(ainpo),2);
	tempbuf[ii*4+ib]=1.0f;
	
}


void presb(int nsam, int irad, float amag, CPLX *spec, 
	CPLX *speq, CPLX *c3df, CPLX *c3ds, 
	float phi, float theta, float psi, float shx, float shy, 
	float *rbin,float *sinclut, int ipad, float* rbuf, float thet , int iewald,
	CPLX *ctff, CPLX *ctfs, float ri2, float ri3, float rih, float halfw)
{
	int ic, jc;
	float dm[9];
	float cphi, cthe, scal;
	float cpsi, spsi, sphi, sthe;
	
	int tmpsize;
	int i;
	int ibin;

//   timeval begin0,end0,begin1,end1,begin2,end2;
//   	gettimeofday(&begin0,NULL);
    
	jc=nsam/2+1;
	ic=nsam*nsam/2;

		
	cphi=cos(phi);
	sphi=sin(phi);
	cthe=cos(theta);
	sthe=sin(theta);
	cpsi=cos(psi);
	spsi=sin(psi);

	dm[0]=(cphi*cthe*cpsi-sphi*spsi)/abs(amag);
	dm[1]=(sphi*cthe*cpsi+cphi*spsi)/abs(amag);
	dm[2]=-sthe*cpsi/abs(amag);
	dm[3]=(-cphi*cthe*spsi-sphi*cpsi)/abs(amag);
	dm[4]=(-sphi*cthe*spsi+cphi*cpsi)/abs(amag);
	dm[5]=sthe*spsi/abs(amag);
	//MW next lines uncommented for Ewald sphere correction
	dm[6]=sthe*cphi/abs(amag);
	dm[7]=sthe*sphi/abs(amag);
	dm[8]=cthe/abs(amag);
	
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);

	float *drbuf=pGM->d_bf;
	
	float *buftd=drbuf;
	
	float *tempbuf=buftd+nsam*nsam+2*nsam;
	tmpsize=(nsam/2+1)*(nsam+1);
	
	cudaMemset(buftd,0,sizeof(float)*(nsam*nsam+2*nsam+5*tmpsize));
	
	cudaMemcpy((void *)pGM->d_dm,(void *)dm,sizeof(float)*9, cudaMemcpyHostToDevice);
	
	scal=float(nsam*nsam/2);

	//###########################################################
	dim3 dimBlock(8,16);
	dim3 dimGrid(jc/dimBlock.x+1,(nsam+1)/dimBlock.y+1);
	CalPresb_1<<<dimGrid,dimBlock>>>(buftd, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, 
								(CPLX *)pGM->d_c3df, (CPLX *)pGM->d_c3ds, pGM->d_sinclut, 
								nsam, jc, ic,iewald, irad, ipad,scal,
								pGM->d_dm, thet);
	cudaThreadSynchronize();
	//###########################################################

	//###########################################################				
	// do fft
	rlft3_GPU((CPLX *)(buftd),(CPLX *)(buftd+nsam*nsam),nsam,nsam,1,-1);
	
	maskcos2(nsam,(buftd),ri2,ri3,rih,halfw,amag);

	//do fft
	rlft3_GPU((CPLX *)(buftd),(CPLX *)(buftd+nsam*nsam),nsam,nsam,1,1);	

	//###########################################################

	//###########################################################
	CalPresb_2<<<dimGrid,dimBlock>>>(buftd, tempbuf, (CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, 
												shx, shy,nsam, jc, nsam*nsam/4);
	cudaThreadSynchronize();
	//###########################################################

	//gettimeofday(&begin1,NULL);
	cudaMemcpy((void *)(rbuf+nsam*4),(void *)tempbuf,
						sizeof(float)*5*tmpsize, cudaMemcpyDeviceToHost);
	memset(rbuf,0,sizeof(float)*4*nsam);	
	//gettimeofday(&end1,NULL);
	int *ns=(int *)rbuf;
	float *bufa=rbuf+nsam;
	float *bufb=rbuf+2*nsam;
	float *bufab=rbuf+3*nsam;
	
	int *t0=(int *)&rbuf[nsam*4+tmpsize*0];
	float *t1=&rbuf[nsam*4+tmpsize*1];
	float *t2=&rbuf[nsam*4+tmpsize*2];
	float *t3=&rbuf[nsam*4+tmpsize*3];
	float *t4=&rbuf[nsam*4+tmpsize*4];
	//gettimeofday(&begin2,NULL);
	for(i=0;i<tmpsize;++i)
	{
		if(ibin=t0[i])
		{
			bufab[ibin]+=t1[i];
			bufa[ibin]+=t2[i];
			bufb[ibin]+=t3[i];
			ns[ibin]+=int(t4[i]);
		}
	}
	//gettimeofday(&end2,NULL);			 

	for(i=1;i<=nsam/2;++i)
	{
		if(ns[i-1]!=0)
		{
			rbin[2*i-2]=bufab[i-1]/sqrt(bufa[i-1]*bufb[i-1]);
			rbin[2*i-1]=1.0;
		}
	}


//		gettimeofday(&end0,NULL);
//   cout<<"@"<<CalTime(begin0,end0)<<endl;
 //  cout<<" **************** 1 is "<<CalTime(begin1,end1)<<endl;
 //  cout<<" **************** 2 is "<<CalTime(begin2,end2)<<endl;

}





