#include "variance_gpu.h"

extern pthread_key_t cuMemKey;
static __device__ float cabs(CPLX f)
{
	return sqrt(f.x*f.x+f.y*f.y);
}

static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

__global__ void var_1(int nsam, int jc, float shx, float shy, CPLX *spec, CPLX *speq,
							CPLX *datc, CPLX *datq)
{
	int l=blockIdx.x*blockDim.x+threadIdx.x+1;
	int m=blockIdx.y*blockDim.y+threadIdx.y+1;
	if(l>jc || m>nsam) return;
	int id;
	int ll=l-1;
	int mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	float phase=shx*ll+shy*mm;
	CPLX pshft;
	pshft.x=cos(phase);
	pshft.y=sin(phase);
	
	if(l!=jc)
	{
		id=l+nsam/2*(m-1);
		datc[id-1]=cXc(spec[id-1],pshft);
	}
	else
	{
		datq[m-1].x=0.0f;
		datq[m-1].y=0.0f;
	}
}

__global__ void var_2(int nsam, float rip, float *datd, float *outd, float *bdsumna)
{
	//bdsumna[5]={bsum, bsum2, dsum, dsum2, int(na)};
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int ii=blockIdx.y*blockDim.y+threadIdx.y+1;
	
	__shared__ float bsum[512];
	__shared__ float bsum2[512];
	__shared__ float dsum[512];
	__shared__ float dsum2[512];
	__shared__ int na[512];
	
	int tid=threadIdx.x*blockDim.y+threadIdx.y;
	bsum[tid]=0.0f;
	bsum2[tid]=0.0f;
	dsum[tid]=0.0f;
	dsum2[tid]=0.0f;;
	na[tid]=0;
	__syncthreads();
	
	if(i>nsam || ii>nsam) return;
	
	int nmid=nsam/2+1;
	int id=ii+nsam*(i-1);
	float crad2=pow(float(ii-nmid),2)+pow(float(i-nmid),2);
	
	datd[id-1]=datd[id-1]/nsam/nsam;
	outd[id-1]=datd[id-1];
	if(crad2>rip)
	{
		bsum[tid]+=outd[id-1];
		bsum2[tid]+=pow(outd[id-1],2);
	}
	else
	{
		na[tid]++;
		dsum[tid]+=datd[id-1];
		dsum2[tid]+=pow(datd[id-1],2);
	}
	
	__syncthreads();	
	if(tid<256)
	{
		bsum[tid]+=bsum[tid+256];
		bsum2[tid]+=bsum2[tid+256];
		dsum[tid]+=dsum[tid+256];
		dsum2[tid]+=dsum2[tid+256];
		na[tid]+=na[tid+256];
	}
	__syncthreads();	
	if(tid<128)
	{
		bsum[tid]+=bsum[tid+128];
		bsum2[tid]+=bsum2[tid+128];
		dsum[tid]+=dsum[tid+128];
		dsum2[tid]+=dsum2[tid+128];
		na[tid]+=na[tid+128];
	}
	__syncthreads();	
	if(tid<64)
	{
		bsum[tid]+=bsum[tid+64];
		bsum2[tid]+=bsum2[tid+64];
		dsum[tid]+=dsum[tid+64];
		dsum2[tid]+=dsum2[tid+64];
		na[tid]+=na[tid+64];
	}
	__syncthreads();	
	if(tid<32)
	{
		bsum[tid]+=bsum[tid+32];
		bsum2[tid]+=bsum2[tid+32];
		dsum[tid]+=dsum[tid+32];
		dsum2[tid]+=dsum2[tid+32];
		na[tid]+=na[tid+32];
	}
	__syncthreads();	
	if(tid<16)
	{
		bsum[tid]+=bsum[tid+16];
		bsum2[tid]+=bsum2[tid+16];
		dsum[tid]+=dsum[tid+16];
		dsum2[tid]+=dsum2[tid+16];
		na[tid]+=na[tid+16];
	}
	__syncthreads();	
	if(tid<8)
	{
		bsum[tid]+=bsum[tid+8];
		bsum2[tid]+=bsum2[tid+8];
		dsum[tid]+=dsum[tid+8];
		dsum2[tid]+=dsum2[tid+8];
		na[tid]+=na[tid+8];
	}
	__syncthreads();	
	if(tid<4)
	{
		bsum[tid]+=bsum[tid+4];
		bsum2[tid]+=bsum2[tid+4];
		dsum[tid]+=dsum[tid+4];
		dsum2[tid]+=dsum2[tid+4];
		na[tid]+=na[tid+4];
	}
	__syncthreads();	
	if(tid<2)
	{
		bsum[tid]+=bsum[tid+2];
		bsum2[tid]+=bsum2[tid+2];
		dsum[tid]+=dsum[tid+2];
		dsum2[tid]+=dsum2[tid+2];
		na[tid]+=na[tid+2];
	}
	__syncthreads();	
	if(tid<1)
	{
		bsum[tid]+=bsum[tid+1];
		bsum2[tid]+=bsum2[tid+1];
		dsum[tid]+=dsum[tid+1];
		dsum2[tid]+=dsum2[tid+1];
		na[tid]+=na[tid+1];
		
		bdsumna[0]+=bsum[0];
		bdsumna[1]+=bsum2[0];
		bdsumna[2]+=dsum[0];
		bdsumna[3]+=dsum2[0];
		bdsumna[4]+=float(na[0]);
	}
	
}

__global__ void var_3(int nsam, float rip, float bsum, float bsum2,
						float *outd, float *datd)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int ii=blockIdx.y*blockDim.y+threadIdx.y+1;
	if(i>nsam || ii>nsam) return;
	
	int id=ii+nsam*(i-1);
	int nmid=nsam/2+1;
	float crad2=pow(float(ii-nmid),2)+pow(float(i-nmid),2);
	if(crad2>rip)
	{
		datd[id-1]=0.0f;
		outd[id-1]=(outd[id-1]-bsum)/bsum2;
	}
	else
	{
		datd[id-1]=(datd[id-1]-bsum)/bsum2;
		outd[id-1]=0.0f;
	}
}

__global__ void var_4(int nsam, int jc, float fraco, float fracd, 
						CPLX *outc, CPLX *datc, int *ic, float *buf)
{
	int l=blockIdx.x*blockDim.x+threadIdx.x+1;
	int m=blockIdx.y*blockDim.y+threadIdx.y+1;
	if(l>nsam/2 || m>nsam) return;
	
	int ll=l-1;
	int mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	int id=l+nsam/2*(m-1);
	int irad=int(sqrt(float(ll*ll+mm*mm))+0.5)+1;
	if(fraco!=0.0f) buf[irad-1]+=pow(cabs(outc[id-1]),2)/fraco;
	if(fracd!=0.0f) buf[nsam+irad-1]+=pow(cabs(datc[id-1]),2)/fracd;
	ic[irad-1]++;
}

__global__ void var_5(int nsam, float *buf, int *ic,
				double *vn, double *vvn, double *vsn, double *vvsn)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	if(i>nsam) return;
	
	if(ic[i-1]!=0)
	{
		buf[i-1]=buf[i-1]/ic[i-1];
		buf[nsam+i-1]=buf[nsam+i-1]/ic[i-1];
	}
	vn[i-1]+=buf[i-1];
	vvn[i-1]+=pow(buf[i-1],2);
	vsn[i-1]+=buf[nsam+i-1];
	vvsn[i-1]+=pow(buf[nsam+i-1],2);
}

__global__ void var_6(int nsam, int *ic)
{
	int i=blockIdx.x*gridDim.x+threadIdx.x;
	if(i>nsam) return;
	ic[i-1]=int(3.1415926535897f*(i-1));
}

extern "C" void variance_(int *pnsam, CPLX *spec, CPLX *speq, float *pshx, float *pshy,
				float *pri, float *outd, CPLX *outc, CPLX *outq, 
				float *datd, CPLX *datc, CPLX *datq, int *ic, double *vsn, double *vn,
				double *vvsn, double *vvn, float *pamagp, float *buf, int *pna)
{
//	timeval begin,end;
//	gettimeofday(&begin,NULL);

	int nsam=*pnsam;
	float shx=*pshx;
	float shy=*pshy;
	float ri=*pri;
	int &na=*pna;
	
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
	spec=(CPLX *)pGM->d_spec;
	speq=(CPLX *)pGM->d_speq;
	
	outd=pGM->d_bf;
	outc=(CPLX *)pGM->d_bf;
	outq=(CPLX *)&(pGM->d_bf[nsam*nsam]);
	
	datd=&(pGM->d_bf[2*nsam*nsam]);
	datc=(CPLX *)&(pGM->d_bf[2*nsam*nsam]);
	datq=(CPLX *)&(pGM->d_bf[3*nsam*nsam]);
	
	buf=&(pGM->d_bf[4*nsam*nsam]);
	
	vn=pGM->d_vn;
	vvn=pGM->d_vvn;
	vsn=pGM->d_vsn;
	vvsn=pGM->d_vvsn;
	ic=pGM->d_ic;
	
	int jc,nsamh;
	float fracd,fraco;
	float rip;
	double dsum,dsum2,bsum,bsum2;
	
	rip=ri*ri;
	jc=nsam/2+1;
	nsamh=nsam/2;
	
	dim3 BlockSize(16,32);
	dim3 var1Grid(jc/BlockSize.x+1,nsam/BlockSize.y+1);
	dim3 var2Grid(nsam/BlockSize.x+1,nsam/BlockSize.y+1);
	dim3 var3Grid(nsam/BlockSize.x+1,nsam/BlockSize.y+1);
	dim3 var4Grid(nsam/2/BlockSize.x+1,nsam/BlockSize.y+1);
	
	dim3 fftBlock(16,32);
	dim3 fftGrid(nsam/fftBlock.x+1,nsamh/fftBlock.y+1);
	
	if(rip<nsamh*nsamh)
	{		
		//#######################
		/*for(l=1;l<=jc;l++)
			for(m=1;m<=nsam;m++)
			{
				ll=l-1;
				mm=m-1;
				if(mm>=jc) mm=mm-nsam;
				phase=shx*ll+shy*mm;
				pshft.x=cos(phase);
				pshft.y=sin(phase);
				if(l!=jc)
				{
					id=l+nsamh*(m-1);
					datc[id-1]=cXc(spec[id-1],pshft);
				}
				else
				{
					datq[m-1].x=0.0f;
					datq[m-1].y=0.0f;
				}
			}*/
		var_1<<<var1Grid,BlockSize>>>(nsam, jc, shx, shy, spec, speq, datc, datq);
		cudaThreadSynchronize();
		//#######################

		//rlft3_(datd,(float *)datq,&nsam,&nsam,&l,&m);			
		rlft3_GPU((CPLX *)datd,datq,nsam,nsam,1,-1);
		
		//MASK PARTICLE, SHARP EDGE
		/*dsum=0.0;
		dsum2=0.0;
		bsum=0.0;
		bsum2=0.0;
		na=0;*/
		
		/*for(i=1;i<=2*nsam;i++)
		{
			buf[i-1]=0.0f;
		}
		for(i=1;i<=nsam;i++)
		{
			ic[i-1]=0;
		}*/
		cudaMemset(buf,0,sizeof(float)*nsam*2);
		cudaMemset(ic,0,sizeof(int)*nsam);
		
		//######################
		/*for(i=1;i<=nsam;i++)
			for(ii=1;ii<=nsam;ii++)
			{
				fi1=pow(float(i-nmid),2);
				id=ii+nsam*(i-1);
				crad2=pow(float(ii-nmid),2)+fi1;
				datd[id-1]=datd[id-1]/nsam/nsam;
				outd[id-1]=datd[id-1];
				if(crad2>rip)
				{
					bsum+=outd[id-1];
					bsum2+=pow(outd[id-1],2);
				}
				else
				{
					na++;
					dsum+=datd[id-1];
					dsum2+=pow(datd[id-1],2);
				}
			}*/
		float h_bdsumna[5];
		float *d_bdsumna=NULL;
		cudaMalloc((void **)&d_bdsumna,sizeof(float)*5);
		cudaMemset((void *)d_bdsumna,0,sizeof(float)*5);
		var_2<<<var2Grid,BlockSize>>>(nsam, rip, datd, outd, d_bdsumna);
		cudaMemcpy(h_bdsumna,d_bdsumna,sizeof(float)*5,cudaMemcpyDeviceToHost);
		cudaFree(d_bdsumna);
		
		bsum=h_bdsumna[0];
		bsum2=h_bdsumna[1];
		dsum=h_bdsumna[2];
		dsum2=h_bdsumna[3];
		na=int(h_bdsumna[4]);
		//############################
			
		dsum=dsum/na;
		dsum2=dsum2/na;
		dsum2=sqrt(dsum2-dsum*dsum);
		bsum=bsum/(nsam*nsam-na);
		bsum2=bsum2/(nsam*nsam-na);
		bsum2=sqrt(bsum2-bsum*bsum);
		
		//##########################
		/*for(i=1;i<=nsam;i++)
			for(ii=1;ii<nsam;ii++)
			{
				fi1=pow(float(i-nmid),2);
				id=ii+nsam*(i-1);
				crad2=pow(float(ii-nmid),2)+fi1;
				if(crad2>rip)
				{
					datd[id-1]=0.0f;
					outd[id-1]=(outd[id-1]-bsum)/bsum2;
				}
				else
				{
					datd[id-1]=(datd[id-1]-bsum)/bsum2;
					outd[id-1]=0.0f;
				}
			}*/
		var_3<<<var3Grid,BlockSize>>>(nsam, rip, bsum, bsum2, outd, datd);
		cudaThreadSynchronize();
		//############################
		
		//rlft3_(datd,(float *)datq,&nsam,&nsam,&l,&m);				
		rlft3_GPU((CPLX *)datd,datq,nsam,nsam,1,1);
		
		//rlft3_(outd,(float *)outq,&nsam,&nsam,&l,&m);			
		rlft3_GPU((CPLX *)outd,outq,nsam,nsam,1,1);

		
		//FRAC will scale the variances to a value which would have
		//been obtained from an unmasked image
		fracd=float(na)/nsam/nsam;
		fraco=float(nsam*nsam-na)/nsam/nsam;
		
		//######################
		/*for(l=1;l<=nsam/2;l++)
			for(m=1;m<=nsam;m++)
			{
				ll=l-1;
				mm=m-1;
				if(mm>=jc) mm=mm-nsam;
				id=l+nsamh*(m-1);
				irad=int(sqrt(float(ll*ll+mm*mm))+0.5)+1;
				if(fraco!=0) buf[irad-1]+=pow(cabs(outc[id-1]),2)/fraco;
				if(fracd!=0) buf[nsam+irad-1]+=pow(cabs(datc[id-1]),2)/fracd;
				ic[irad-1]++;
			}*/
		var_4<<<var4Grid,BlockSize>>>(nsam, jc, fraco, fracd, outc, datc, ic, buf);
		cudaThreadSynchronize();			
		//#######################	
		
		//#######################	
		/*for(i=1;i<=nsam;i++)
		{
			if(ic[i-1]!=0)
			{
				buf[i-1]=buf[i-1]/ic[i-1];
				buf[nsam+i-1]=buf[nsam+i-1]/ic[i-1];
			}
			vn[i-1]+=buf[i-1];
			vvn[i-1]+=pow(buf[i-1],2);
			vsn[i-1]+=buf[nsam+i-1];
			vvsn[i-1]+=pow(buf[nsam+i-1],2);
		}*/
		var_5<<<nsam/512+1,512>>>(nsam, buf, ic, vn, vvn, vsn, vvsn);
		//#######################	
	}
	else
	{
		//####################
		/*for(i=1;i<=nsam;i++)
		{
			ic[i-1]=int(PI*(i-1));
		}*/
		var_6<<<nsam/512+1,512>>>(nsam, ic);
		//####################
	}
	cudaThreadSynchronize();
	
//	gettimeofday(&end,NULL);
//	cout<<" variance time is "<<CalTime(begin,end)<<endl;

}














