#include "CreateThread.h"
#include "lock.h"
#include "lmain_parameters.h"
#include "pagelockedmem.h"

extern vector<planFFT> planlist;

//use GM as thread-global varialbe to manage the memory in device
//Important: GM only is limited to only use in main function of a function or subroutine
//           other places use its pointer pGM to avoid error.
extern pthread_key_t cuMemKey;

void *CreateThread(void *ARG)
{
	//timeval begin,end;
	//gettimeofday(&begin,NULL);
	lock2_();

	lmain_parameters *arg=(lmain_parameters *)ARG;
	
	//initional CUDA device
	 int count;
    cudaGetDeviceCount(&count);
    if(count == 0)
    {
        cout<<"error: There is no device."<<endl;
        exit(0);
    }

    int i;

    if(arg->m_DeviceNum>=count || arg->m_DeviceNum<0)
    {
    	cout<<"error: Wrong Device Number or Device not exist for #"<<arg->m_DeviceNum<<endl;
    	exit(0);
    }

   if(cudaSetDevice(arg->m_DeviceNum)!=cudaSuccess)
   {
   	cout<<"error: Set CUDA Device #"<<arg->m_DeviceNum<<" failed!"<<endl;
   	exit(0);
   }
   
   if(cudaSetDeviceFlags(cudaDeviceMapHost)!=cudaSuccess)
   {
   	cout<<"error: Set CUDA Device #"<<arg->m_DeviceNum<<" Flag failed!"<<endl;
   	exit(0);
   }
    
   //Prepare GPU memory
   //for GPU #0
   cuMem GM;	
   GM.setThreadBlock(arg->m_xBlockSize,arg->m_yBlockSize);
   GM.set(arg->m_nsam,arg->m_ipad,arg->m_IMPMAX,arg->m_NNSTAT,arg->m_sinclut, arg->m_plm);
   int gmerror;
   if(gmerror=GM.alloc(arg->m_nrec))
   {
  	 	 cout<<"error: GPU memory allocation failed!"<<endl;
  	 	 cout<<"error: error number is "<<gmerror<<endl;
  		 exit(0);
   }
   pthread_setspecific(cuMemKey,&GM); 
   //start main loop
   int k;
   int step=arg->m_step;
   arg->m_nstart=arg->m_kstart;
   arg->m_ilast=arg->m_kend;
   
   arg->m_inproj+=(100+arg->m_DeviceNum); //file handle of image_stack.mrc file 
   MRC imstack;
   if(arg->m_fmatch) arg->m_ioproj+=(150+arg->m_DeviceNum);
   
   //for output shft_*_*
   char shftname[64];
   sprintf(shftname,"shft_%d_%d",arg->m_kstart,arg->m_kend);
   FILE *fp;
   if((fp=fopen(shftname,"w"))==NULL) printf("Creating shft file (%s) failed!\n",shftname);
   
	unlock2_();

	for(k=arg->m_kstart;k<=arg->m_kend;k+=step)
	{
		if(arg->m_nrec==1) arg->set_iswitch(0);
		if(arg->m_nrec==2) arg->set_iswitch(1);
		
		lmain_(k, arg->m_fdef, arg->m_fmag, arg->m_inproj, arg->m_cform, arg->m_data, 
			 arg->m_rrec, arg->m_nsam, arg->m_spec, arg->m_shx,
			 arg->m_shy, arg->m_cs, arg->m_rawbins, arg->m_wl, arg->m_wgh1, arg->m_wgh2, 
			 arg->m_dfmid1, arg->m_dfmid2, arg->m_angast, arg->m_outc, 
			 arg->m_outd, arg->m_amagp, arg->m_rih, arg->m_HALFW, arg->m_halfwc, arg->m_ri2, 
			 arg->m_ri3, arg->m_ri4, arg->m_datd, arg->m_datc, 
			 arg->m_b3dv, arg->m_phi, arg->m_theta, arg->m_psi, arg->m_xstd, arg->m_iflag, 
			 arg->m_irada, arg->m_c3df, arg->m_c3ds, arg->m_rbins, 
			 arg->m_rbinn, arg->m_nset, arg->m_noic, arg->m_nois, arg->m_presa, arg->m_sang, 
			 arg->m_nsang, arg->m_ccd, arg->m_ccc, arg->m_iquadmax, 
			 arg->m_NDOC1, arg->m_NDOC2, arg->m_mask, arg->m_rbfac, arg->m_rmax1, 
			 arg->m_rmax2, arg->m_iran,  arg->m_psize, arg->m_dstep, arg->m_film, 
			 arg->m_NN1, arg->m_NNPART, arg->m_MAXSET,	arg->m_fhist, arg->m_ilist, 
			 arg->m_NPB, arg->m_itmax, arg->m_DAMAX, arg->m_asymstore, arg->m_target, 
			 arg->m_iccount, arg->m_iswitch, arg->m_ipad, arg->m_pall, arg->m_iall, 
			 arg->m_irad, arg->m_ri, arg->m_a3df, arg->m_a3ds, arg->m_s3df, 
			 arg->m_s3ds, arg->m_std, arg->m_d3df, arg->m_d3ds, arg->m_ilst, arg->m_v3df, 
			 arg->m_v3ds, arg->m_vtd, arg->m_pbc, arg->m_boff, 
			 arg->m_fflip, arg->m_asum, arg->m_vsum, arg->m_psum, arg->m_ksum, 
			 arg->m_thresh, arg->m_nsamh, arg->m_jc, arg->m_fflag, arg->m_ccave, 
			 arg->m_icmp, arg->m_ifirst, arg->m_ioproj, arg->m_iewald,arg->m_NSCF, 
			 arg->m_sfile, arg->m_nang, arg->m_wgh, arg->m_psis, arg->m_thetas, 
			 arg->m_phis, arg->m_dshxs, arg->m_dshys,arg->m_press, arg->m_iexcl, 
			 arg->m_ibuf, arg->m_icount, arg->m_fmatch, arg->m_dfmid1s, arg->m_dfmid2s, 
			 arg->m_angasts, arg->m_absmags, arg->m_nstart, arg->m_finpat1, 
			 arg->m_finpat2, arg->m_asym, arg->m_vx, arg->m_ilast, arg->m_nnset, arg->m_IANG, 
			 arg->m_cdate, arg->m_ctime, arg->m_czone, arg->m_dtval, arg->m_fastig, 
			 arg->m_sinclut, arg->m_mbuf, arg->m_pbuf, arg->m_ctff, arg->m_ctfs, 
			 arg->m_rbin, arg->m_ccbuf, arg->m_ccs, arg->m_datq, arg->m_outq, arg->m_noiq, 
			 arg->m_speq, arg->m_pbuq, arg->m_qbuc, arg->m_qbuq, 
			 arg->m_istat, arg->m_tnset, arg->m_fall,arg->m_ic, arg->m_vsn, arg->m_vn, 
			 arg->m_vvsn, arg->m_vvn, arg->m_ci, arg->m_bf, 
			 arg->m_ns, arg->m_na, arg->m_fsct, arg->m_sumprel,	arg->m_nprel, arg->m_tx, 
			 arg->m_ty, arg->m_testpar, arg->m_ipmax, arg->m_nsym, 
			 arg->m_isym, arg->m_jsym, arg->m_symop, arg->m_NNSTAT, arg->m_nksum, 
			 arg->m_xm, arg->m_ym, arg->m_sx, arg->m_sy, arg->m_ifsc, 
			 arg->m_imp,arg->m_nthread,fp, imstack);

	}
	fclose(fp);
	//iclose_(&arg->m_inproj);
	imstack.close();
	if(arg->m_fmatch) iclose_(&arg->m_ioproj);

	/*int nerror;
	nerror=GM.getRec((float *)arg->m_a3df,(float *)arg->m_a3ds,(float *)arg->m_s3df,(float *)arg->m_s3ds,
   			 (float *)arg->m_d3df,(float *)arg->m_d3ds,(float *)arg->m_v3df,(float *)arg->m_v3ds);
	if(nerror>0) cout<<"error: Sum reconstruction result failed in CreateThread.cu!"<<endl;

   GM.getStat(arg->m_sumprel,arg->m_nprel,arg->m_asum,arg->m_vsum,arg->m_psum,arg->m_ksum,
   			 arg->m_std,arg->m_vtd,arg->m_nksum,arg->m_vn,arg->m_vvn,arg->m_vsn,arg->m_vvsn,
   			 arg->m_ic); 	*/	    			 

   GM.free();	

   //Destroy FFT Plan created in this thread
	for(i=0;i<planlist.size();i++)
	{
		if(planlist[i].tid==pthread_self())
		{
			cufftDestroy(planlist[i].plan);
			planlist.erase(planlist.begin()+i);
			i--;
		}
	}
	
   //gettimeofday(&end,NULL);
   //cout<<" time is "<<CalTime(begin,end)<<endl;	 
	return (void *)0;
}






