#include "calcfx_gpu.h"
#include "cc3_gpu.h"
#include "cc3m_gpu.h"
#include "fsh_gpu.h"
#include "ctfapply_phase_only_gpu.h"

//     CALCULATES NEW VALUE FOR F TO INPUT TO SUBROUTIN VA04A
//     There are three modes - in all phase residual is returned as value of RF
//	IF IVFLAG=0, phase residual is calculated with 5 new particle parameters
//	IF IVFLAG=1, phase residual is recalculated with old particle parameters
//		after application of new, but unweighted CTF correction (+/-1.0)
//		for use in defocus refinement
//	IF IVFLAG=2, phase residual is recalculated with old particle parameters
//		but using the new MAG and unweighted CTF correction (i.e. +/-1.0)
//		for use in magnification refinement
void calcfx(int nn1, int nx, float *xparmap, float &rf, 
	float *xpar, int npar, int nsam, int &maxr1, int &maxr2, int *mask, 
	CPLX *a3df, CPLX *a3ds, int irad, CPLX *outq, float*pbuf, CPLX *pbuq, 
	float *shx, float *shy, float cs, float wl, float wgh1, float wgh2, 
	float thetatr, CPLX *ctff, CPLX *ctfs, float *amag, float rih, 
	float halfw, float ri2, float ri3, float ri4, float *phi, 
	float *theta, float *psi, float rmax1, float rmax2, float xstd, 
	float *mbuf, int *ilst, CPLX *datc, CPLX *datq, int ibuf, float *b3dv, 
	float *datd, float *sinclut, int ivflag, float rbfact, float *outd, 
	CPLX *outc, CPLX *qbuc, CPLX *qbuq, int npb, int ipad, float *rbuf, 
	float *fsct, float *dfmid1, float *dfmid2, float *angast, int iewald, 
	float tx, float ty, float xm, float ym, float sx, float sy, 
	float sig2)
{	
	int i, k, id, is;
	float df1, df2;
	int idd;
	float ast, rft;
	int ncalcfx_ctf;
	int ipar, jpar;
	
	//timeval begin,end,begin0,end0;
   //gettimeofday(&begin0,NULL);
	
	ncalcfx_ctf=0;
	if(abs(ivflag)==1)
	{
		//for defocus and astigmatism, parameters map directly
		rft=0.0;
		for(i=1;i<=ibuf;++i)
		{
			k=ilst[i-1];
			idd=1+nsam/2*nsam*(i-1);
			id=1+nsam*nsam*(i-1);
			is=1+nsam*(i-1);
			maxr1=int(nsam*rmax1*abs(amag[k-1]));
			maxr2=int(nsam*rmax2*abs(amag[k-1]));
			if(nx==1) xparmap[1]=xparmap[0];
			df1=dfmid1[k-1]+xparmap[0];
			df2=dfmid2[k-1]+xparmap[1];
			ast=angast[k-1]+xparmap[2];
			
			ctfapply_phase_only(nsam, (CPLX *)&pbuf[id-1],&pbuq[is-1],
						shx[k-1],shy[k-1],cs,wl,wgh1,wgh2,df1,df2,ast,
						thetatr,ctff,ctfs,outd,outc,outq,amag[k-1],rih,
						halfw,ri2,ri3,ri4,datd,datc,datq,b3dv,
						phi[k-1],theta[k-1],psi[k-1],&mbuf[id-1],
						xstd,ivflag,tx,ty);
			rft+=cc3(nsam,irad,amag[k-1],outc,outq,a3df,a3ds,
						maxr1,maxr2,phi[k-1],theta[k-1],psi[k-1],0.0,0.0,ivflag,
						&qbuc[idd-1],&qbuq[is-1],rbfact,sinclut,ipad,rbuf,fsct,
						iewald,thetatr,ctff,ctfs,ri2,ri3,rih,halfw);
		}
		
		rft=rft/float(ibuf);
		ivflag=-1;
		ncalcfx_ctf++;
	}
	else if(ivflag==2)
	{
		//for magnification, parameters map directly
		rft=0.0;
		for(i=1;i<=ibuf;++i)
		{
			k=ilst[i-1];
			idd=1+nsam/2*nsam*(i-1);
			id=1+nsam*nsam*(i-1);
			is=1+nsam*(i-1);
			maxr1=int(nsam*rmax1*abs(xparmap[0]));
			maxr2=int(nsam*rmax2*abs(xparmap[0]));
			//gettimeofday(&begin,NULL);
			ctfapply_phase_only(nsam, (CPLX *)&pbuf[id-1],&pbuq[is-1],
						shx[k-1],shy[k-1],cs,wl,wgh1,wgh2,dfmid1[k-1],dfmid2[k-1],angast[k-1],
						thetatr,ctff,ctfs,outd,outc,outq,xparmap[0],rih,
						halfw,ri2,ri3,ri4,datd,datc,datq,b3dv,
						phi[k-1],theta[k-1],psi[k-1],&mbuf[id-1],
						xstd,ivflag,tx,ty);
			rft+=cc3(nsam,irad,xparmap[0],outc,outq,a3df,a3ds,
						maxr1,maxr2,phi[k-1],theta[k-1],psi[k-1],0.0,0.0,ivflag,
						&qbuc[idd-1],&qbuq[is-1],rbfact,sinclut,ipad,rbuf,fsct,
						iewald,thetatr,ctff,ctfs,ri2,ri3,rih,halfw);
			//gettimeofday(&end,NULL);
		}
		
		rft=rft/float(ibuf);
	}
	else
	{
		ipar=0;
		for(jpar=1;jpar<=5;++jpar)
		{
			if(mask[jpar-1]==1)
			{
				ipar++;
				xpar[jpar-1]=xparmap[ipar-1];
			}
		}
		if(npar>ipar) xpar[5]=xparmap[npar-1];
//gettimeofday(&begin,NULL);
		rft=-cc3m(nsam,irad,xpar[5],outc,outq,a3df,a3ds,
				maxr1,maxr2,xpar[0],xpar[1],xpar[2],xpar[3],xpar[4],
				rbfact,sinclut,ipad,rbuf,fsct,iewald,
				thetatr,ctff,ctfs,ri2,ri3,rih,halfw)
			 -fsh(sig2,xpar[3],xpar[4],xm,ym,sx,sy);
//gettimeofday(&end,NULL);
		if(abs(xpar[0])>PI) rft+=pow(PI-abs(xpar[0]),2);
		if(abs(xpar[1])>PI) rft+=pow(PI-abs(xpar[1]),2);
		if(abs(xpar[2])>PI) rft+=pow(PI-abs(xpar[2]),2);
		if(abs(xpar[3])>1.0) rft+=pow(1.0-abs(xpar[3]),2);
		if(abs(xpar[4])>1.0) rft+=pow(1.0-abs(xpar[4]),2);
	}

	rf=rft;

//gettimeofday(&end0,NULL);
  // cout<<" ********calcfx is "<<CalTime(begin0,end0)<<"  "<<CalTime(begin,end)<<endl;
	//cout<<" **************** lmain is "<<CalTime(begin0,end0)<<"  "<<float(CalTime(begin,end))*100/float(CalTime(begin0,end0))<<endl;
	
	
	return;
}



















