#include "cc3_gpu.h"
#include "maskcos2_gpu.h"
//#include "ainterpo3dbig_gpu.h"
//#include "ainterpo3ds_gpu.h"      
//#include "ewaldex_gpu.h"
#include "cuMem.h"

extern pthread_key_t cuMemKey;

	texture<CPLX, 3, cudaReadModeElementType> texRef_c3df;	
	texture<CPLX, 2, cudaReadModeElementType> texRef_c3ds;
	texture<float, 1, cudaReadModeElementType> texRef_sinclut;
	texture<CPLX, 2, cudaReadModeElementType> texRef_ctff;
	texture<CPLX, 1, cudaReadModeElementType> texRef_ctfs;

/*__device__ inline void atomicFloatAdd(float *address, float val)
{
      int i_val = __float_as_int(val);
      int tmp0 = 0;
      int tmp1;

      while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0)
      {
              tmp0 = tmp1;
              i_val = __float_as_int(val + __int_as_float(tmp1));
      }
}*/
//******************
static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

static __device__ float cabs(CPLX f)
{
	return sqrt(f.x*f.x+f.y*f.y);
}

static __device__  CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}

static __device__ float boxft_lut(float *arg,float *sinclut)
{
	int i,j;
	float sinc,prod=1.0f;
	
	for(i=0;i<3;i++)
	{
		arg[i]=abs(arg[i]);
		if(arg[i]<0.02f)
		{
			sinc=1.0f;
		}
		else
		{
			j=int(180.0f*arg[i]+0.5f);
			//sinc=sinclut[j-1];
			sinc=tex1Dfetch(texRef_sinclut,j-1);
		}
		prod=prod*sinc;
	}
	return prod;

}


static __device__ CPLX ainterpo3dbig(int nsama, int ipad, CPLX *a3df, CPLX *a3ds, 
                                      float &x, float &y, float &z)
{
	int nsam,jc;
	int l,m,n,ll,mm,nn,id,nsamh;
	CPLX samp;
	
	nsam=ipad*nsama;
	samp.x=0.0f;
	samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	l=(x>=0.0f)?int(ipad*x+0.5f):int(ipad*x-0.5f);
	m=(y>=0.0f)?int(ipad*y+0.5f):int(ipad*y-0.5f);
	n=(z>=0.0f)?int(ipad*z+0.5f):int(ipad*z-0.5f);
	
	ll=(l>=0)?(l):(-l);
	mm=(l>=0)?(m):(-m);				
	nn=(l>=0)?(n):(-n);
	if(mm<0) mm=mm+nsam;
	if(nn<0) nn=nn+nsam;
	
	if(ll!=jc)
	{
		//id=ll+nsamh*((mm-1)+nsam*(nn-1));
		//samp=cPc(samp,(l>=0)?a3df[id-1]:conjg(a3df[id-1]));
		samp=cPc(samp,(l>=0)?tex3D(texRef_c3df,ll,mm,nn):conjg(tex3D(texRef_c3df,ll,mm,nn)));	
	}
	else
	{
		//id=mm+nsam*(nn-1);
		//samp=cPc(samp,(l>=0)?a3ds[id-1]:conjg(a3ds[id-1]));
		samp=cPc(samp,(l>=0)?tex2D(texRef_c3ds,mm,nn):conjg(tex2D(texRef_c3ds,mm,nn)));
	}

	return samp;
}  

/*if(l>=0)
	{
		ll=l+1;
		mm=m+1;
		if(mm<1) mm=mm+nsam;
		nn=n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,a3df[id-1]);
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,a3ds[is-1]);
		}
	}
	else
	{
		ll=-l+1;
		mm=-m+1;
		if(mm<1) mm=mm+nsam;
		nn=-n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,conjg(a3df[id-1]));
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,conjg(a3ds[is-1]));
		}
	}*/       

static __device__ CPLX ainterpo3ds(int nsam, int irad, CPLX *a3df, CPLX *a3ds, 
                        float &xx, float &yy, float &zz, float *sinclut, int &ipad)
{
	int ls,lt,ns,nt,ms,mt,jc;
	int l,m,n,ll,mm,nn;
	//int is,id,nsamh;
	float x,y,z,arg[3],boxftv;
	CPLX samp,c3d;
	
	samp.x=samp.y=0.0f;
	jc=nsam/2+1;
	//nsamh=nsam/2;
	x=ipad*xx;
	y=ipad*yy;
	z=ipad*zz;
	
	ls=int(x)-irad+1;
	if(x<0.0f) ls=ls-1;
	if(ls<(-jc)) ls=-jc+1;
	lt=int(x)+irad;
	if(x<0.0f) lt=lt-1;
	if(lt>=jc) lt=jc-1;
	
	ms=int(y)-irad+1;
	if(y<0.0f) ms=ms-1;
	if(ms<=(-jc)) ms=-jc+1;
	mt=int(y)+irad;
	if(y<0.0f) mt=mt-1;
	if(mt>=jc) mt=jc-1;
	
	ns=int(z)-irad+1;
	if(z<0.0f) ns=ns-1;
	if(ns<=(-jc)) ns=-jc+1;
	nt=int(z)+irad;
	if(z<0.0f) nt=nt-1;
	if(nt>=jc) nt=jc-1;
	
	for(l=ls;l<=lt;l++)
	{
		arg[0]=x-l;
		for(m=ms;m<=mt;m++)
		{
			arg[1]=y-m;
			for(n=ns;n<=nt;n++)
			{
				arg[2]=z-n;
				boxftv=boxft_lut(arg,sinclut);
				//if(l>=0)
				//{
					ll=(l>=0)?(l):(-l);
					mm=(l>=0)?(m):(-m);				
					nn=(l>=0)?(n):(-n);
					if(mm<0) mm=mm+nsam;
					if(nn<0) nn=nn+nsam;
					c3d=(ll!=(jc-1)) ? tex3D(texRef_c3df,ll,mm,nn) : tex2D(texRef_c3ds,mm,nn);
					if(l<0) c3d.y=-c3d.y;
					samp=cPc(samp,cXr(c3d,boxftv));
				//}
				/*else
				{
					ll=-l+1;
					mm=-m+1;
					if(mm<1) mm=mm+nsam;
					nn=-n+1;
					if(nn<1) nn=nn+nsam;
					c3d=(ll!=jc) ? tex3D(texRef_c3df,ll-1,mm-1,nn-1) : tex2D(texRef_c3ds,mm-1,nn-1);
					samp=cPc(samp,conjg(cXr(c3d,boxftv)));
				}*/
				
			}
		}
	}
	
	return samp;
}           

static __device__ CPLX ewaldex(int &nsam, int &irad, CPLX *a3df, CPLX *a3ds, 
             float *sinclut, int &ipad,
             int &i, int &j,float *dm, float &thet, CPLX &ctfr, CPLX &ctfl)
{
	float x3,y3,z3;
	float thetah,gpix,x,y,z;
	CPLX ainpo1,ainpo2;
	
	gpix=sqrt(float(i*i+j*j));  //length of resol.vector g in pixel
	thetah=gpix*thet/2;       //THET=(WL/(PSIZE*NSAM))/AMAGP
	x=i*cos(thetah);          //THETAH=scattering angle/2.
	y=j*cos(thetah);
	z=gpix*sin(thetah);
	x3=dm[0]*x+dm[3]*y+dm[6]*z;
	y3=dm[1]*x+dm[4]*y+dm[7]*z;
	z3=dm[2]*x+dm[5]*y+dm[8]*z;
	
	if(irad==0) //AINPO1 is first beam
	{
		ainpo1=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
	}
	else
	{
		ainpo1=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
	}
	
	x3=dm[0]*x+dm[3]*y-dm[6]*z;  //AINPO2 is second beam = conjg(x,y,-z)
	y3=dm[1]*x+dm[4]*y-dm[7]*z;
	z3=dm[2]*x+dm[5]*y-dm[8]*z;
	if(irad==0)
	{
		ainpo2=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
	}
	else
	{
		ainpo2=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
	}
	return cPc(cXc(ainpo1,ctfr),cXc(ainpo2,conjg(ctfl)));  //combine beams
}

//******************

__global__ void CalCc3_1(float *rbuf, CPLX *ctff, CPLX *ctfs, CPLX *c3df, CPLX *c3ds, 
						float *sinclut, int nsam, int jc, int ic, int maxr12, int maxr22, 
						int iewald, int irad, int ipad, float scal, float* dm, float thetatr,
						float dm0, float dm1, float dm2, float dm3, float dm4, float dm5)
{
	//for(i=0;i<=(jc-1);++i)
	//for(j=(-jc+1);j<=(jc-1);++j)
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;
	int ii,jj,ib,id,itemp;
	float x3,y3,z3;
	CPLX ctfr,ctfl,ainpo;
	
	if(i>(jc-1) || j>(jc-1)) return;		
	ii=i+1;
	itemp=i*i+j*j;
	if(!(itemp>=maxr12 && itemp<maxr22)) return;

	jj=j+1;
	if(jj<1) jj+=nsam;
	if(ii!=jc)
	{
		id=ii+nsam/2*(jj-1);
		//ctfr=ctff[id-1];
		//ctfl=ctff[id+ic-1];
		ctfr=tex2D(texRef_ctff,ii-1,jj-1);
		ctfl=tex2D(texRef_ctff,ii-1,jj+nsam-1);
	}
	else
	{
		//ctfr=ctfs[jj-1];
		//ctfl=ctfs[jj+nsam-1];
		ctfr=tex1Dfetch(texRef_ctfs,jj-1);
		ctfr=tex1Dfetch(texRef_ctfs,jj+nsam-1);
	}
	
	//MW call ewald sphere corrected extraction
	if(iewald!=0)
	{
		if(iewald<0)
		{
			ctfr=conjg(ctfr);
			ctfl=conjg(ctfl);
		}
		ainpo=ewaldex(nsam,irad,c3df,c3ds,sinclut,ipad,
							i,j,dm,thetatr,ctfr,ctfl);
		ainpo=cXc(ainpo,cPc(ctfr,conjg(ctfl)));
	}	
	else
	{
		x3=dm0*i+dm3*j;
		y3=dm1*i+dm4*j;
		z3=dm2*i+dm5*j;
		//if(ibuf>=0)   //always =0, removed by Xueming
		//{
		if(irad==0)
		{
			ainpo=ainterpo3dbig(nsam,ipad,c3df,c3ds,x3,y3,z3);
		}
		else
		{
			ainpo=ainterpo3ds(ipad*nsam,irad,c3df,c3ds,x3,y3,z3,sinclut,ipad);
		}	
		ainpo=cXr(ainpo,pow(cabs(cPc(ctfr,conjg(ctfl))),2)); 
			//ainpo was moved to here by Xueming					
		//}
		
	}

	if(ii!=jc)
	{
		//if(ibuf>0) pbuc[id-1]=ainpo; //ibuf always =0, removed by Xueming
		//if(ibuf<0) ainpo=pbuc[id-1];
		ib=2*(nsam+id);
		rbuf[ib-2]=ainpo.x/scal;
		rbuf[ib-1]=ainpo.y/scal;
	}
	else
	{
		//if(ibuf>0) pbuq[jj-1]=ainpo;
		//if(ibuf<0) ainpo=pbuq[jj-1];
		ib=nsam*(nsam+2)+2*jj;
		rbuf[ib-2]=ainpo.x/scal;
		rbuf[ib-1]=ainpo.y/scal;
	}

}

__global__ void CalCc3_1fast_iradnoteq0(float *rbuf,int nsam, int jc, int ic, int maxr12, int maxr22, 
						int irad, int ipad, float scal, float thetatr,
						float dm0, float dm1, float dm2, float dm3, float dm4, float dm5)
{
	//for(i=0;i<=(jc-1);++i)
	//for(j=(-jc+1);j<=(jc-1);++j)
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;
	int ii,jj,ib;
	float x3,y3,z3;
	CPLX ctfr,ctfl,ainpo;
	
	if(i>(jc-1) || j>(jc-1)) return;		
	ii=i+1;
	ib=i*i+j*j;
	if(!(ib>=maxr12 && ib<maxr22)) return;
	
	x3=dm0*i+dm3*j;
	y3=dm1*i+dm4*j;
	z3=dm2*i+dm5*j;
	
	jj=j+1;
	if(jj<1) jj+=nsam;
	if(ii!=jc)
	{
		ctfr=tex2D(texRef_ctff,ii-1,jj-1);
		ctfl=tex2D(texRef_ctff,ii-1,jj+nsam-1);
		
		ainpo=ainterpo3ds(ipad*nsam,irad,NULL,NULL,x3,y3,z3,NULL,ipad);
		ainpo=cXr(ainpo,pow(cabs(cPc(ctfr,conjg(ctfl))),2)); 
		
		ib=2*ii+nsam*(jj+1);;
		rbuf[ib-2]=ainpo.x/scal;
		rbuf[ib-1]=ainpo.y/scal;
		
	}
	else
	{
		ctfr=tex1Dfetch(texRef_ctfs,jj-1);
		ctfr=tex1Dfetch(texRef_ctfs,jj+nsam-1);
		
		ainpo=ainterpo3ds(ipad*nsam,irad,NULL,NULL,x3,y3,z3,NULL,ipad);
		ainpo=cXr(ainpo,pow(cabs(cPc(ctfr,conjg(ctfl))),2)); 
		
		ib=nsam*(nsam+2)+2*jj;
		rbuf[ib-2]=ainpo.x/scal;
		rbuf[ib-1]=ainpo.y/scal;
	}

}

__global__ void CalCc3_2(float *rbuf, CPLX *outc, CPLX *outq, float shx, float shy,
								int nsam, int jc, int maxr12, int maxr22)
{
	//for(i=0;i<=(jc-1);++i)
	//for(j=(-jc+1);j<=(jc-1);++j)
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;	
	int ii,jj,itemp;
	float phase;
	int id,ib,ibin;

	if(i>(jc-1) || j>(jc-1)) return;
			
	itemp=i*i+j*j;
	
	if(!(itemp>=maxr12 && itemp<maxr22)) return;

	ii=i+1;
	jj=j+1;
	if(jj<1) jj+=nsam;

	phase=shx*i+shy*j;
	CPLX pshft,samp,ainpo;
	pshft.x=cos(phase);
	pshft.y=sin(phase);
				
	if(ii!=jc)
	{
		id=ii+nsam/2*(jj-1);
		samp=cXc(outc[id-1],pshft);

		ib=2*(nsam+id);
		ainpo.x=rbuf[ib-2];
		ainpo.y=rbuf[ib-1];
	}
	else
	{
		samp=cXc(outq[jj-1],pshft);
		//if(ibuf>0) pbuq[jj-1]=ainpo;
		//if(ibuf<0) ainpo=pbuq[jj-1];
		ib=nsam*(nsam+2)+2*jj;
		ainpo.x=rbuf[ib-2];
		ainpo.y=rbuf[ib-1];
	}

	ibin=4*(int(sqrt(float(itemp))+0.5f)+1);
	//rbuf[ibin-4]+=(cXc(samp,conjg(ainpo))).x;
	//rbuf[ibin-3]+=pow(cabs(samp),2);
	//rbuf[ibin-2]+=pow(cabs(ainpo),2);
	//rbuf[ibin-1]+=1.0f;
	
	ib=(nsam+1)*i+(j+nsam/2);
	id=nsam*(4+nsam);
	ii=(nsam/2+1)*(nsam+1);
	*((int *)&rbuf[id+ib])=ibin;
	rbuf[id+ii*1+ib]=(cXc(samp,conjg(ainpo))).x;
	rbuf[id+ii*2+ib]=pow(cabs(samp),2);
	rbuf[id+ii*3+ib]=pow(cabs(ainpo),2);
	rbuf[id+ii*4+ib]=1.0f;
	
}


double cc3(int nsam, int irad, float amag, CPLX *spec, 
	CPLX *speq, CPLX *a3df, CPLX *a3ds, int maxr1, int maxr2, 
	float phi, float theta, float psi, float shx, float shy, 
	int ibuf, CPLX *pbuc, CPLX *pbuq, float rbfact, float *sinclut, 
	int ipad, float *rbuf, float *fsct, int iewald, float thetatr, 
	CPLX *ctff, CPLX *ctfs, float ri2, float ri3, float rih, float halfw)
{
	int ic, jc;
	int ib1, ib2;
	float dm[9];
	float cphi, cthe, scal;
	float cpsi, spsi, sphi, sthe;
	double sdpha, snpha;
	int maxr12, maxr22;
	
	int tmpsize;
	int i;
	int ibin;
	float w,r,wgt;
//  timeval begin0,end0,begin1,end1,begin2,end2;
//   	gettimeofday(&begin0,NULL);

	jc=nsam/2+1;
	ic=nsam*nsam/2;
	w=1.0f/float(nsam);
	maxr12=maxr1*maxr1;
	maxr22=maxr2*maxr2;
	
	cphi=cos(phi);
	sphi=sin(phi);
	cthe=cos(theta);
	sthe=sin(theta);
	cpsi=cos(psi);
	spsi=sin(psi);

	dm[0]=(cphi*cthe*cpsi-sphi*spsi)/abs(amag);
	dm[1]=(sphi*cthe*cpsi+cphi*spsi)/abs(amag);
	dm[2]=-sthe*cpsi/abs(amag);
	dm[3]=(-cphi*cthe*spsi-sphi*cpsi)/abs(amag);
	dm[4]=(-sphi*cthe*spsi+cphi*cpsi)/abs(amag);
	dm[5]=sthe*spsi/abs(amag);
	//MW next lines uncommented for Ewald sphere correction
	dm[6]=sthe*cphi/abs(amag);
	dm[7]=sthe*sphi/abs(amag);
	dm[8]=cthe/abs(amag);
	
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
	/*for(i=0;i<nsam*(4+nsam);++i)
	{
		rbuf[i]=0.0f;
	}*/
	float *drbuf=pGM->d_bf;
	tmpsize=(nsam/2+1)*(nsam+1);
	
	cudaMemset(drbuf,0,sizeof(float)*(nsam*(4+nsam)+5*tmpsize));

	ib1=2*nsam;
	ib2=ib1+nsam*nsam;
	scal=float(nsam*nsam/2);

	//###################texture######################
	//bind to texture reference
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<CPLX>();
	cudaBindTextureToArray(&texRef_c3df, pGM->cu_c3df, &channelDesc);
	cudaBindTextureToArray(&texRef_c3ds, pGM->cu_c3ds, &channelDesc);
	
	size_t offset=0;
	cudaBindTextureToArray(&texRef_ctff, pGM->cu_ctff, &channelDesc);
	cudaBindTexture(&offset, &texRef_ctfs, (void *)pGM->d_ctfs, &channelDesc,size_t(sizeof(CPLX)*nsam*2));

	cudaChannelFormatDesc channelfloat = cudaCreateChannelDesc<float>();
	cudaBindTexture(&offset, &texRef_sinclut, (void *)pGM->d_sinclut, &channelfloat,size_t(sizeof(float)*2000));	
   //###############################################
//gettimeofday(&begin1,NULL);
	//###########################################################
	dim3 dimBlock(8,16);
	dim3 dimGrid(jc/dimBlock.x+1,(nsam+1)/dimBlock.y+1);

	if(iewald==0 && irad!=0)
	{
		CalCc3_1fast_iradnoteq0<<<dimGrid,dimBlock>>>(drbuf,nsam, jc, ic, maxr12, maxr22,irad, ipad,scal,
								thetatr,dm[0], dm[1], dm[2], dm[3], dm[4], dm[5]);
	}
	else
	{
		cudaMemcpy((void *)pGM->d_dm,(void *)dm,sizeof(float)*9, cudaMemcpyHostToDevice);
		CalCc3_1<<<dimGrid,dimBlock>>>(drbuf, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, 
								(CPLX *)pGM->d_c3df, (CPLX *)pGM->d_c3ds, pGM->d_sinclut, 
								nsam, jc, ic, maxr12, maxr22,	iewald, irad, ipad,scal,
								pGM->d_dm, thetatr,dm[0], dm[1], dm[2], dm[3], dm[4], dm[5]);
	}
	cudaThreadSynchronize();

	//###########################################################
//gettimeofday(&end1,NULL);
	//###########################################################
	//	itemp1=1;
	//	itemp2=-1;
	//rlft3_(&rbuf[ib1],&rbuf[ib2],&nsam,&nsam,&itemp1,&itemp2);
				
	// do fft
	rlft3_GPU((CPLX *)(drbuf+ib1),(CPLX *)(drbuf+ib2),nsam,nsam,1,-1);

	//maskcos2(nsam,&rbuf[ib1],ri2,ri3,rih,halfw,amag);
	maskcos2(nsam,(drbuf+ib1),ri2,ri3,rih,halfw,amag);

	//	itemp1=1;
	//	itemp2=1;
	//rlft3_(&rbuf[ib1],&rbuf[ib2],&nsam,&nsam,&itemp1,&itemp2);
	//do fft
	rlft3_GPU((CPLX *)(drbuf+ib1),(CPLX *)(drbuf+ib2),nsam,nsam,1,1);	
	//###########################################################

	//###########################################################
	//cudaMemcpy((void *)pGM->d_outc,(void *)outc,sizeof(float)*nsam*nsam, cudaMemcpyHostToDevice);
	//cudaMemcpy((void *)pGM->d_outq,(void *)outq,sizeof(float)*nsam*2, cudaMemcpyHostToDevice);
	CalCc3_2<<<dimGrid,dimBlock>>>(drbuf, (CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, 
												shx, shy,nsam, jc, maxr12, maxr22);
	cudaThreadSynchronize();
	//###########################################################
	//gettimeofday(&begin2,NULL);
	cudaMemcpy((void *)(rbuf+nsam*(nsam+4)),(void *)(drbuf+nsam*(nsam+4)),
						sizeof(float)*5*tmpsize, cudaMemcpyDeviceToHost);
	//gettimeofday(&end2,NULL);					
	memset(rbuf,0,sizeof(float)*2*nsam);	

	int *t0=(int *)&rbuf[nsam*(nsam+4)+tmpsize*0];
	float *t1=&rbuf[nsam*(nsam+4)+tmpsize*1];
	float *t2=&rbuf[nsam*(nsam+4)+tmpsize*2];
	float *t3=&rbuf[nsam*(nsam+4)+tmpsize*3];
	float *t4=&rbuf[nsam*(nsam+4)+tmpsize*4];

	
	for(i=0;i<tmpsize;++i)
	{
		if(ibin=t0[i])
		{ 
			rbuf[ibin-4]+=t1[i];
			rbuf[ibin-3]+=t2[i];
			rbuf[ibin-2]+=t3[i];
			rbuf[ibin-1]+=t4[i];
		}
	}
		 
	
	sdpha=0.0;
	snpha=0.0;
	for(i=2;i<=nsam/2;++i)
	{
		if(rbuf[i*4-1]!=0.0f)
		{
			rbuf[4*i-4]/=sqrt(rbuf[4*i-2]*rbuf[4*i-3]);
			r=rbuf[4*i-4]*abs(rbuf[4*i-4])/(w+abs(rbuf[4*i-4]));
			wgt=exp(-rbfact*(i-1)*(i-1));
			sdpha+=r*r*r*wgt;
			snpha+=wgt;
		}
	}
		
	if(sdpha!=0.0) sdpha=exp(log(abs(sdpha)/snpha)/3.0)*sdpha/abs(sdpha);
	if(sdpha>1.0) sdpha=1.0;
	if(sdpha<-1.0) sdpha=-1.0;

	//	gettimeofday(&end0,NULL);
   //cout<<"@"<<CalTime(begin0,end0)<<endl;
   //cout<<" **************** 1 is "<<CalTime(begin1,end1)<<endl;
  // cout<<" **************** 2 is "<<CalTime(begin2,end2)<<endl;

	return acos(sdpha);
}
