#include "ccp_gpu.h"
#include "rotate2d_gpu.h"
#include "maskcos2_gpu.h"
#include "cuMem.h"

extern pthread_key_t cuMemKey;

extern "C" void maskcos2_(int *nsam, float *outd, float *ri2, float *ri3, 
					float *rih, float *halfw, float *amagp);

	texture<CPLX, 3, cudaReadModeElementType> texRef_c3df;	
	texture<CPLX, 2, cudaReadModeElementType> texRef_c3ds;
	texture<float, 1, cudaReadModeElementType> texRef_sinclut;

//******************
static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

static __device__ float cabs(CPLX f)
{
	return sqrt(f.x*f.x+f.y*f.y);
}

static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}

static __device__ float boxft_lut(float *arg,float *sinclut)
{
	int i,j;
	float sinc,prod=1.0f;
	
	for(i=0;i<3;i++)
	{
		arg[i]=abs(arg[i]);
		if(arg[i]<0.02f)
		{
			sinc=1.0f;
		}
		else
		{
			j=int(180.0f*arg[i]+0.5f);
			//sinc=sinclut[j-1];
			sinc=tex1Dfetch(texRef_sinclut,j-1);
		}
		prod=prod*sinc;
	}
	return prod;

}


static __device__ CPLX ainterpo3dbig(int nsama, int ipad, CPLX *a3df, CPLX *a3ds, 
                                      float &x, float &y, float &z)
{
	int nsam,jc;
	int l,m,n,ll,mm,nn,id,nsamh;
	CPLX samp;
	
	nsam=ipad*nsama;
	samp.x=0.0f;
	samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	l=(x>=0.0f)?int(ipad*x+0.5f):int(ipad*x-0.5f);
	m=(y>=0.0f)?int(ipad*y+0.5f):int(ipad*y-0.5f);
	n=(z>=0.0f)?int(ipad*z+0.5f):int(ipad*z-0.5f);
	
	ll=(l>=0)?(l):(-l);
	mm=(l>=0)?(m):(-m);				
	nn=(l>=0)?(n):(-n);
	if(mm<0) mm=mm+nsam;
	if(nn<0) nn=nn+nsam;
	
	if(ll!=jc)
	{
		//id=ll+nsamh*((mm-1)+nsam*(nn-1));
		//samp=cPc(samp,(l>=0)?a3df[id-1]:conjg(a3df[id-1]));
		samp=cPc(samp,(l>=0)?tex3D(texRef_c3df,ll,mm,nn):conjg(tex3D(texRef_c3df,ll,mm,nn)));	
	}
	else
	{
		//id=mm+nsam*(nn-1);
		//samp=cPc(samp,(l>=0)?a3ds[id-1]:conjg(a3ds[id-1]));
		samp=cPc(samp,(l>=0)?tex2D(texRef_c3ds,mm,nn):conjg(tex2D(texRef_c3ds,mm,nn)));
	}

	return samp;
}     

/*if(l>=0)
	{
		ll=l+1;
		mm=m+1;
		if(mm<1) mm=mm+nsam;
		nn=n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,a3df[id-1]);
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,a3ds[is-1]);
		}
	}
	else
	{
		ll=-l+1;
		mm=-m+1;
		if(mm<1) mm=mm+nsam;
		nn=-n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,conjg(a3df[id-1]));
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,conjg(a3ds[is-1]));
		}
	}*/       

static __device__ CPLX ainterpo3ds(int nsam, int irad, CPLX *a3df, CPLX *a3ds, 
                        float &xx, float &yy, float &zz, float *sinclut, int &ipad)
{
	int ls,lt,ns,nt,ms,mt,jc;
	int l,m,n,ll,mm,nn;
	//int is,id,nsamh;
	float x,y,z,arg[3],boxftv;
	CPLX samp,c3d;
	
	samp.x=samp.y=0.0f;
	jc=nsam/2+1;
	//nsamh=nsam/2;
	x=ipad*xx;
	y=ipad*yy;
	z=ipad*zz;
	
	ls=int(x)-irad+1;
	if(x<0.0f) ls=ls-1;
	if(ls<(-jc)) ls=-jc+1;
	lt=int(x)+irad;
	if(x<0.0f) lt=lt-1;
	if(lt>=jc) lt=jc-1;
	
	ms=int(y)-irad+1;
	if(y<0.0f) ms=ms-1;
	if(ms<=(-jc)) ms=-jc+1;
	mt=int(y)+irad;
	if(y<0.0f) mt=mt-1;
	if(mt>=jc) mt=jc-1;
	
	ns=int(z)-irad+1;
	if(z<0.0f) ns=ns-1;
	if(ns<=(-jc)) ns=-jc+1;
	nt=int(z)+irad;
	if(z<0.0f) nt=nt-1;
	if(nt>=jc) nt=jc-1;
	
	for(l=ls;l<=lt;l++)
	{
		arg[0]=x-l;
		for(m=ms;m<=mt;m++)
		{
			arg[1]=y-m;
			for(n=ns;n<=nt;n++)
			{
				arg[2]=z-n;
				boxftv=boxft_lut(arg,sinclut);
				//if(l>=0)
				//{
					ll=(l>=0)?(l):(-l);
					mm=(l>=0)?(m):(-m);				
					nn=(l>=0)?(n):(-n);
					if(mm<0) mm=mm+nsam;
					if(nn<0) nn=nn+nsam;
					c3d=(ll!=(jc-1)) ? tex3D(texRef_c3df,ll,mm,nn) : tex2D(texRef_c3ds,mm,nn);
					if(l<0) c3d.y=-c3d.y;
					samp=cPc(samp,cXr(c3d,boxftv));
				//}
				/*else
				{
					ll=-l+1;
					mm=-m+1;
					if(mm<1) mm=mm+nsam;
					nn=-n+1;
					if(nn<1) nn=nn+nsam;
					c3d=(ll!=jc) ? tex3D(texRef_c3df,ll-1,mm-1,nn-1) : tex2D(texRef_c3ds,mm-1,nn-1);
					samp=cPc(samp,conjg(cXr(c3d,boxftv)));
				}*/
				
			}
		}
	}
	
	return samp;
}  


/*
	for(i=0;i<=(jc-2);++i)
		for(j=(-jc+1);j<=(jc-1);++j)
		{
			ii=i+1;
			itemp=i*i+j*j;
			jj=j+1;
			if(jj<1) jj+=nsam;
			id=ii+nsamh*(jj-1);
			if(itemp>=maxr12 && itemp<maxr22 && abs(j)!=(jc-1))
			{
				//Only low-resolution search done here so no Ewald sphere
				//correction needed... 
				x3=dm[0]*i+dm[3]*j;
				y3=dm[1]*i+dm[4]*j;
				z3=dm[2]*i+dm[5]*j;
				if(irad==0)
				{
					ccc[id-1]=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
				}
				else
				{
					ccc[id-1]=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
				}
			}

		}
*/
__global__ void ccp_1(int nsam, int jc, int maxr12, int maxr22, CPLX *ccc, CPLX *ccs,
							int ipad, int irad, CPLX *c3df, CPLX *c3ds, float dm0, float dm1,
							float dm2, float dm3, float dm4, float dm5)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;
	if(i>(jc-2) || j>(jc-1)) return;
	int ii,jj,itemp,id;
	float x3,y3,z3;
	
	ii=i+1;
	itemp=i*i+j*j;
	jj=j+1;
	if(jj<1) jj+=nsam;
	id=ii+nsam/2*(jj-1);
	if(itemp>=maxr12 && itemp<maxr22 && abs(j)!=(jc-1))
	{
		//Only low-resolution search done here so no Ewald sphere
		//correction needed... 
		x3=dm0*i+dm3*j;
		y3=dm1*i+dm4*j;
		z3=dm2*i+dm5*j;
		if(irad==0)
		{
			ccc[id-1]=ainterpo3dbig(nsam,ipad,c3df,c3ds,x3,y3,z3);
		}
		else
		{
			ccc[id-1]=ainterpo3ds(ipad*nsam,irad,c3df,c3ds,x3,y3,z3,NULL,ipad);
		}
	}

}
	
/*
			for(i=0;i<nsamrh;++i)
				for(j=-nsamrh;j<=(nsamrh-1);++j)
				{
					ii=i+1;
					itemp=i*i+j*j;
					jj=j+1;
					jj2=jj;
					if(jj<1)
					{
						jj+=nsam;
						jj2+=nsamr;
					}
					id=ii+nsamh*(jj-1);
					id2=ii+nsamrh*(jj2-1);

					if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
					{
						weight=exp(-rbfact*itemp);
						ainpo=cbuf[id-1];
						samp=cXr(spec[id-1],weight);
						ccc[id2-1]=cXc(samp,conjg(ainpo));
						sum+=pow(cabs(ainpo),2);
						sum2+=pow(cabs(samp),2);
					}
					else
					{
						ccc[id2-1].x=0.0f;
						ccc[id2-1].y=0.0f;
					}
				}
*/

//iquad==1
//thread number per block must be 128 (8*16)
__global__ void ccp_iquad_1(int nsam, int nsamrh,float *dsum, int maxr22, int jc, 
							float rbfact, CPLX *cbuf, CPLX *spec, CPLX *ccc)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-nsamrh;
	__shared__ float sum[128];
	__shared__ float sum2[128];
	int tid=threadIdx.x+8*threadIdx.y;
	sum[tid]=0.0f;
	sum2[tid]=0.0f;
	if(i>=nsamrh || j>=nsamrh) return;
	int id,id2,ii,jj,jj2,itemp;
	float weight;
	CPLX ainpo,samp;
	//*****************************
	ii=i+1;
	itemp=i*i+j*j;
	jj=j+1;
	jj2=jj;
	if(jj<1)
	{
		jj+=nsam;
		jj2+=nsamrh*2;
	}
	id=ii+nsam/2*(jj-1);
	id2=ii+nsamrh*(jj2-1);

	if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
	{
		weight=exp(-rbfact*itemp);
		ainpo=cbuf[id-1];
		samp=cXr(spec[id-1],weight);
		ccc[id2-1]=cXc(samp,conjg(ainpo));
		sum[tid]=pow(cabs(ainpo),2);
		sum2[tid]=pow(cabs(samp),2);
	}
	else
	{
		ccc[id2-1].x=0.0f;
		ccc[id2-1].y=0.0f;
	}
	//*****************************
	__syncthreads();
	
	if(tid<64)
	{
		sum[tid]+=sum[tid+64];
		sum2[tid]+=sum2[tid+64];
	}
	__syncthreads();
	
	if(tid<32)
	{
		sum[tid]+=sum[tid+32];
		sum2[tid]+=sum2[tid+32];
	}
	__syncthreads();
	
	if(tid<16)
	{
		sum[tid]+=sum[tid+16];
		sum2[tid]+=sum2[tid+16];
	}
	__syncthreads();
	
	if(tid<8)
	{
		sum[tid]+=sum[tid+8];
		sum2[tid]+=sum2[tid+8];
	}
	__syncthreads();
	
	if(tid<4)
	{
		sum[tid]+=sum[tid+4];
		sum2[tid]+=sum2[tid+4];
	}
	__syncthreads();
	
	if(tid<2)
	{
		sum[tid]+=sum[tid+2];
		sum2[tid]+=sum2[tid+2];
	}
	__syncthreads();
	
	if(tid<1)
	{
		sum[tid]+=sum[tid+1];
		sum2[tid]+=sum2[tid+1];
		
		//dimension of grid is 16*32
		tid=blockIdx.x+16*blockIdx.y;
		dsum[tid]=sum[0];
		dsum[tid+512]=sum2[0];
	}	
	__syncthreads();
}

//iquad==2
//thread number per block must be 128 (8*16)
__global__ void ccp_iquad_2(int nsam, int nsamrh,float *dsum, int maxr22, int jc, 
							float rbfact, CPLX *cbuf, CPLX *spec, CPLX *ccc)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-nsamrh;
	__shared__ float sum[128];
	__shared__ float sum2[128];
	int tid=threadIdx.x+8*threadIdx.y;
	sum[tid]=0.0f;
	sum2[tid]=0.0f;
	if(i>=nsamrh || j>=nsamrh) return;
	int id,id2,ii,jj,jj2,iii,jjj,idd,itemp,icon;
	float weight;
	CPLX ainpo,samp;
	
	//*****************************
	ii=i+1;
	itemp=i*i+j*j;
	jj=j+1;
	jj2=jj;
	if(jj<1)
	{
		jj+=nsam;
		jj2+=nsamrh*2;
	}
	id=ii+nsam/2*(jj-1);
	id2=ii+nsamrh*(jj2-1);

	if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
	{
		weight=exp(-rbfact*itemp);
		if(j<=0)
		{
			iii=-j+1;
			jjj=i+1;
			icon=1;
		}
		else
		{
			iii=j+1;
			jjj=-i+1;
			if(jjj<1) jjj+=nsam;
			icon=-1;
		}									
		idd=iii+nsam/2*(jjj-1);
		ainpo=cbuf[idd-1];
		if(icon==-1) ainpo=conjg(ainpo);

		samp=cXr(spec[id-1],weight);
		ccc[id2-1]=cXc(samp,conjg(ainpo));
		sum[tid]=pow(cabs(ainpo),2);
		sum2[tid]=pow(cabs(samp),2);
	}
	else
	{
		ccc[id2-1].x=0.0f;
		ccc[id2-1].y=0.0f;
	}
	//*****************************
	__syncthreads();
	
	if(tid<64)
	{
		sum[tid]+=sum[tid+64];
		sum2[tid]+=sum2[tid+64];
	}
	__syncthreads();
	
	if(tid<32)
	{
		sum[tid]+=sum[tid+32];
		sum2[tid]+=sum2[tid+32];
	}
	__syncthreads();
	
	if(tid<16)
	{
		sum[tid]+=sum[tid+16];
		sum2[tid]+=sum2[tid+16];
	}
	__syncthreads();
	
	if(tid<8)
	{
		sum[tid]+=sum[tid+8];
		sum2[tid]+=sum2[tid+8];
	}
	__syncthreads();
	
	if(tid<4)
	{
		sum[tid]+=sum[tid+4];
		sum2[tid]+=sum2[tid+4];
	}
	__syncthreads();
	
	if(tid<2)
	{
		sum[tid]+=sum[tid+2];
		sum2[tid]+=sum2[tid+2];
	}
	__syncthreads();
	
	if(tid<1)
	{
		sum[tid]+=sum[tid+1];
		sum2[tid]+=sum2[tid+1];
		
		//dimension of grid is 16*32
		tid=blockIdx.x+16*blockIdx.y;
		dsum[tid]=sum[0];
		dsum[tid+512]=sum2[0];
	}	
	__syncthreads();
}
	
	
//iquad==3
//thread number per block must be 128 (8*16)
__global__ void ccp_iquad_3(int nsam, int nsamrh,float *dsum, int maxr22, int jc, 
							float rbfact, CPLX *cbuf, CPLX *spec, CPLX *ccc)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-nsamrh;
	__shared__ float sum[128];
	__shared__ float sum2[128];
	int tid=threadIdx.x+8*threadIdx.y;
	sum[tid]=0.0f;
	sum2[tid]=0.0f;
	if(i>=nsamrh || j>=nsamrh) return;
	int id,id2,ii,jj,jj2,iii,jjj,idd,itemp,icon;
	float weight;
	CPLX ainpo,samp;
	//*****************************
	ii=i+1;
	itemp=i*i+j*j;
	jj=j+1;
	jj2=jj;
	if(jj<1)
	{
		jj+=nsam;
		jj2+=nsamrh*2;
	}
	id=ii+nsam/2*(jj-1);
	id2=ii+nsamrh*(jj2-1);

	if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
	{
		weight=exp(-rbfact*itemp);
		iii=i+1;
		jjj=j+1;
		if(jjj<1) jjj+=nsam;
		icon=-1;			
		idd=iii+nsam/2*(jjj-1);
		ainpo=cbuf[idd-1];
		if(icon==-1) ainpo=conjg(ainpo);

		samp=cXr(spec[id-1],weight);
		ccc[id2-1]=cXc(samp,conjg(ainpo));
		sum[tid]=pow(cabs(ainpo),2);
		sum2[tid]=pow(cabs(samp),2);
	}
	else
	{
		ccc[id2-1].x=0.0f;
		ccc[id2-1].y=0.0f;
	}
	//*****************************
	__syncthreads();
	
	if(tid<64)
	{
		sum[tid]+=sum[tid+64];
		sum2[tid]+=sum2[tid+64];
	}
	__syncthreads();
	
	if(tid<32)
	{
		sum[tid]+=sum[tid+32];
		sum2[tid]+=sum2[tid+32];
	}
	__syncthreads();
	
	if(tid<16)
	{
		sum[tid]+=sum[tid+16];
		sum2[tid]+=sum2[tid+16];
	}
	__syncthreads();
	
	if(tid<8)
	{
		sum[tid]+=sum[tid+8];
		sum2[tid]+=sum2[tid+8];
	}
	__syncthreads();
	
	if(tid<4)
	{
		sum[tid]+=sum[tid+4];
		sum2[tid]+=sum2[tid+4];
	}
	__syncthreads();
	
	if(tid<2)
	{
		sum[tid]+=sum[tid+2];
		sum2[tid]+=sum2[tid+2];
	}
	__syncthreads();
	
	if(tid<1)
	{
		sum[tid]+=sum[tid+1];
		sum2[tid]+=sum2[tid+1];
		
		//dimension of grid is 16*32
		tid=blockIdx.x+16*blockIdx.y;
		dsum[tid]=sum[0];
		dsum[tid+512]=sum2[0];
	}	
	__syncthreads();
}	


//iquad==4
//thread number per block must be 128 (8*16)
__global__ void ccp_iquad_4(int nsam, int nsamrh,float *dsum, int maxr22, int jc, 
							float rbfact, CPLX *cbuf, CPLX *spec, CPLX *ccc)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-nsamrh;
	__shared__ float sum[128];
	__shared__ float sum2[128];
	int tid=threadIdx.x+8*threadIdx.y;
	sum[tid]=0.0f;
	sum2[tid]=0.0f;
	if(i>=nsamrh || j>=nsamrh) return;
	int id,id2,ii,jj,jj2,iii,jjj,idd,itemp,icon;
	float weight;
	CPLX ainpo,samp;
	//*****************************
	ii=i+1;
	itemp=i*i+j*j;
	jj=j+1;
	jj2=jj;
	if(jj<1)
	{
		jj+=nsam;
		jj2+=nsamrh*2;
	}
	id=ii+nsam/2*(jj-1);
	id2=ii+nsamrh*(jj2-1);

	if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
	{
		weight=exp(-rbfact*itemp);

		if(j>0)
		{
			iii=j+1;
			jjj=-i+1;
			if(jjj<1)jjj+=nsam;
			icon=1;
		}
		else
		{
			iii=-j+1;
			jjj=i+1;
			icon=-1;
		}				
		idd=iii+nsam/2*(jjj-1);
		ainpo=cbuf[idd-1];
		if(icon==-1) ainpo=conjg(ainpo);

		samp=cXr(spec[id-1],weight);
		ccc[id2-1]=cXc(samp,conjg(ainpo));
		sum[tid]=pow(cabs(ainpo),2);
		sum2[tid]=pow(cabs(samp),2);
	}
	else
	{
		ccc[id2-1].x=0.0f;
		ccc[id2-1].y=0.0f;
	}
	//*****************************
	__syncthreads();
	
	if(tid<64)
	{
		sum[tid]+=sum[tid+64];
		sum2[tid]+=sum2[tid+64];
	}
	__syncthreads();
	
	if(tid<32)
	{
		sum[tid]+=sum[tid+32];
		sum2[tid]+=sum2[tid+32];
	}
	__syncthreads();
	
	if(tid<16)
	{
		sum[tid]+=sum[tid+16];
		sum2[tid]+=sum2[tid+16];
	}
	__syncthreads();
	
	if(tid<8)
	{
		sum[tid]+=sum[tid+8];
		sum2[tid]+=sum2[tid+8];
	}
	__syncthreads();
	
	if(tid<4)
	{
		sum[tid]+=sum[tid+4];
		sum2[tid]+=sum2[tid+4];
	}
	__syncthreads();
	
	if(tid<2)
	{
		sum[tid]+=sum[tid+2];
		sum2[tid]+=sum2[tid+2];
	}
	__syncthreads();
	
	if(tid<1)
	{
		sum[tid]+=sum[tid+1];
		sum2[tid]+=sum2[tid+1];
		
		//dimension of grid is 16*32
		tid=blockIdx.x+16*blockIdx.y;
		dsum[tid]=sum[0];
		dsum[tid+512]=sum2[0];
	}	
	__syncthreads();
}
			
void ccp(int nsam, int irad, float amag, CPLX *spec, CPLX *speq, 
	CPLX *a3df, CPLX *a3ds, float *rbuf,int maxr1, int maxr2, 
	float &phi, float &theta, float &psi, float &shx, float &shy, 
	float &ccmax, float *ccd, CPLX *ccc, CPLX *ccs, CPLX *cbuf, 
	int iquad, int *mask, float rbfact, float *sinclut, int ipad, 
	int iewald, float thetatr, /*CPLX *ctff, CPLX *ctfs,*/ 
	float ri2, float ri3, float rih, float halfw, int iquadmax)
{
	int i,j,jc,id,iq,np,ix,iy;
	int ixm,iym;
	int ipsi,nsamh;
	int maxr12,maxr22,nsamr,nsamrh;
	float cc,dm[9],sx,sy;
	float ccm,sum,sum2,cphi,cthe;
	float cpsi,sphi,sthe,spsi,psis,psim;
	int ipad2=4;
	int npsi=18;
	int itemp1,itemp2;
	float hsum[1024];
	
//	timeval begin,end,begin0,end0;
//   gettimeofday(&begin0,NULL);
	

	jc=nsam/2+1;
	nsamh=nsam/2;
	nsamrh=nsamh/2;
	if((nsamrh*2)!=nsamh) nsamrh=(nsamh+1)/2;
	nsamr=2*nsamrh;
	maxr12=maxr1*maxr1;
	maxr22=maxr2*maxr2;
	cphi=cos(phi);
	sphi=sin(phi);
	cthe=cos(theta);
	sthe=sin(theta);
	cpsi=cos(psi);
	spsi=sin(psi);
	
	
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	//###################texture######################
	//bind to texture reference
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<CPLX>();
	cudaBindTextureToArray(&texRef_c3df, pGM->cu_c3df, &channelDesc);
	cudaBindTextureToArray(&texRef_c3ds, pGM->cu_c3ds, &channelDesc);
	
	size_t offset=0;
	cudaChannelFormatDesc channelfloat = cudaCreateChannelDesc<float>();
	cudaBindTexture(&offset, &texRef_sinclut, (void *)pGM->d_sinclut, &channelfloat,size_t(sizeof(float)*2000));	
   //###############################################
   
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* To save time, 90 deg rotated projections are generated by index permutation... */
/*  i.e. the projection structure factors are interpolated only for IQUAD=1 */
/*       then recalled from CBUF for the 3 other rotated projections */

/* PSI=		 0  90 180 270 */
/* D1=CPSI=	 1   0  -1   0 */
/* D2=SPSI=	 0   1   0  -1 */
/* D3=0.0 */
/* D4=-SPSI=	 0  -1   0   1 */
/* D5=CPSI=	 1   0  -1   0 */
/* D6=0.0 */
/* X=		 I  -J  -I   J */
/* Y=		 J   I  -J  -I */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*  Below is the standard Eulerian rotation ZYZ */
/*    The model or its transform is rotated first by PHI around Z, then */
/*    by THETA about the new Y, and thirdly by PSI about the new Z. */

/*    The rotation matrix used is  R=R(psi)*R(theta)*R(phi) as in Spider */

/*                 c  s  0          c  0 -s          c  s  0 */
/*                -s  c  0     *    0  1  0     *   -s  c  0 */
/*                 0  0  1          s  0  c          0  0  1 */

/*                 about Z          about Y          about Z */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	dm[0]=(cphi*cthe*cpsi-sphi*spsi)/abs(amag);
	dm[1]=(sphi*cthe*cpsi+cphi*spsi)/abs(amag);
	dm[2]=-sthe*cpsi/abs(amag);
	dm[3]=(-cphi*cthe*spsi-sphi*cpsi)/abs(amag);
	dm[4]=(-sphi*cthe*spsi+cphi*cpsi)/abs(amag);
	dm[5]=sthe*spsi/abs(amag);
/* MW next lines uncommented for Ewald sphere correction */
	dm[6]=sthe*cphi/abs(amag);
	dm[7]=sthe*sphi/abs(amag);
	dm[8]=cthe/abs(amag);

	//memset((void*)ccc,0,sizeof(CPLX)*nsam*nsam/2);
	//memset((void*)ccs,0,sizeof(CPLX)*nsam);
	cudaMemset((void*)pGM->d_ccc,0,sizeof(CPLX)*nsam*nsam/2);
	cudaMemset((void*)pGM->d_ccs,0,sizeof(CPLX)*nsam);

	dim3 dimBlock(8,16);
	dim3 dimGrid(nsam/2/dimBlock.x+1,(nsam+1)/dimBlock.y+1);
	ccp_1<<<dimGrid,dimBlock>>>(nsam, jc, maxr12, maxr22, (CPLX *)pGM->d_ccc, (CPLX *)pGM->d_ccs,
							ipad, irad, a3df, a3ds, dm[0], dm[1], dm[2], dm[3], dm[4], dm[5]);
	cudaThreadSynchronize();

	//itemp1=1;
	//itemp2=-1;
	//rlft3_(ccc,ccs,&nsam,&nsam,&itemp1,&itemp2);	
	rlft3_GPU((CPLX *)pGM->d_ccc,(CPLX *)pGM->d_ccs,nsam,nsam,1,-1);

	//maskcos2_(&nsam,(float *)ccc,&ri2,&ri3,&rih,&halfw,&amag);
	maskcos2(nsam,pGM->d_ccc,ri2,ri3,rih,halfw,amag);
		
	//itemp1=1;
	//itemp2=1;
	//rlft3_(ccc,ccs,&nsam,&nsam,&itemp1,&itemp2);
	rlft3_GPU((CPLX *)pGM->d_ccc,(CPLX *)pGM->d_ccs,nsam,nsam,1,1);	
	
	ccmax=-1e30f;
	np=npsi;
	iq=iquadmax;
	if(mask[2]==0)
	{
		np=1;
		iq=1;
	}
//gettimeofday(&begin,NULL);
//gettimeofday(&end,NULL);
//printf("rotate2d  %15d\n",CalTime(begin,end));
	dim3 DimBlockiq(8,16);
	dim3 DimGridiq(nsamrh/DimBlockiq.x+1,nsamrh*2/DimBlockiq.y+1);
	cudaMemset(pGM->d_msum1,0,sizeof(float)*1024); //move to here from iquad_loop
	for(ipsi=1;ipsi<=np;++ipsi)
	{
		//rotate2d(nsam,-ipad2,ccc,ccs,cbuf,psis,(CPLX *)rbuf);
		psis=(ipsi-1)*PI/2.0f/npsi;
		if(ipsi==1) 
			rotate2d(nsam,-ipad2,(CPLX *)pGM->d_ccc,(CPLX *)pGM->d_ccs,(CPLX *)pGM->d_ccbuf,psis,(CPLX *)pGM->d_bf);
		else 
			rotate2d(nsam,ipad2,(CPLX *)pGM->d_ccc,(CPLX *)pGM->d_ccs,(CPLX *)pGM->d_ccbuf,psis,(CPLX *)pGM->d_bf);
		
		for(iquad=1;iquad<=iq;++iquad)
		{
			sum=0.0f;
			sum2=0.0f;
	
		
//##########################################################################
			switch(iquad)
			{
				case 1:
					ccp_iquad_1<<<DimGridiq,DimBlockiq>>>(nsam, nsamrh,pGM->d_msum1, maxr22, jc, 
							rbfact, (CPLX *)pGM->d_ccbuf, (CPLX *)pGM->d_outc, (CPLX *)pGM->d_ccc);
					break;
				case 2:
					ccp_iquad_2<<<DimGridiq,DimBlockiq>>>(nsam, nsamrh,pGM->d_msum1, maxr22, jc, 
							rbfact, (CPLX *)pGM->d_ccbuf, (CPLX *)pGM->d_outc, (CPLX *)pGM->d_ccc);
					break;
				case 3:
					ccp_iquad_3<<<DimGridiq,DimBlockiq>>>(nsam, nsamrh,pGM->d_msum1, maxr22, jc, 
							rbfact, (CPLX *)pGM->d_ccbuf, (CPLX *)pGM->d_outc, (CPLX *)pGM->d_ccc);
					break;
				case 4:
					ccp_iquad_4<<<DimGridiq,DimBlockiq>>>(nsam, nsamrh,pGM->d_msum1, maxr22, jc, 
							rbfact, (CPLX *)pGM->d_ccbuf, (CPLX *)pGM->d_outc, (CPLX *)pGM->d_ccc);
					break;
			}
			cudaThreadSynchronize();

			cudaMemcpy(hsum,pGM->d_msum1,sizeof(float)*1024,cudaMemcpyDeviceToHost);
			for(i=0;i<512;++i)
			{
				sum+=hsum[i];
				sum2+=hsum[i+512];
			}
//##########################################################################
			cudaMemset(pGM->d_ccs,0,sizeof(CPLX)*nsamr);
			//itemp1=1;
			//itemp2=-1;
			//rlft3_((CPLX *)ccd,ccs,&nsamr,&nsamr,&itemp1,&itemp2);						
			rlft3_GPU((CPLX *)pGM->d_ccc,(CPLX *)pGM->d_ccs,nsamr,nsamr,1,-1);		
			cudaMemcpy(ccc,pGM->d_ccc,sizeof(float)*nsamr*nsamr,cudaMemcpyDeviceToHost);					
	
			ccm=-1e30f;
	
	//cross-correlation is searched in (x,y) unless MASK is zero
			ixm=0;
			iym=0;
	//    Assume that particle is roughly centered: search only over NSAM/2
	//    in each direction 
			for(ix=-6*mask[3];ix<=6*mask[3];++ix)
				for(iy=-6*mask[4];iy<=6*mask[4];++iy)
				{
					j=ix+1;
					if(j<1) j+=nsamr;
					i=iy+1;
					if(i<1) i+=nsamr;
					id=j+nsamr*(i-1);
					if(ccd[id-1]>ccm)
					{
						ccm=ccd[id-1];
						ixm=ix;
						iym=iy;
					}
				}
			ccm=0.0f;
			for(ix=(ixm-1);ix<=(ixm+1);++ix)
				for(iy=(iym-1);iy<=(iym+1);++iy)
				{			
					j=ix+1;
					if(j<1) j+=nsamr;
					i=iy+1;
					if(i<1)i+=nsamr;
					id=j+nsamr*(i-1);
					ccm+=ccd[id-1];
				}
			cc=ccm/sqrt(sum)/sqrt(sum2)/9;
//			printf("%17.9f  %17.9f\n",cc,ccmax);
			if(cc>ccmax)
			{
				sx=(float)ixm/nsamr*PI*2.0f;
				sy=(float)iym/nsamr*PI*2.0f;
				psim=psi+(iquad-1)*PI/2.0f+psis;
				ccmax=cc;
			}

		}

	}

	shx=sx;
	shy=sy;
	psi=psim;
	
	
//	gettimeofday(&end0,NULL);
//	cout<<"@"<<CalTime(begin0,end0)<<endl;
  // printf("ccp  %15d %15d\n",CalTime(begin,end),CalTime(begin0,end0));

	return;
}





/*

*************  iquad == 1  **********************
if(iquad==1)
			for(i=0;i<nsamrh;++i)
				for(j=-nsamrh;j<=(nsamrh-1);++j)
				{
					ii=i+1;
					itemp=i*i+j*j;
					jj=j+1;
					jj2=jj;
					if(jj<1)
					{
						jj+=nsam;
						jj2+=nsamr;
					}
					id=ii+nsamh*(jj-1);
					id2=ii+nsamrh*(jj2-1);

					if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
					{
						weight=exp(-rbfact*itemp);
						ainpo=cbuf[id-1];
						samp=cXr(spec[id-1],weight);
						ccc[id2-1]=cXc(samp,conjg(ainpo));
						sum+=pow(cabs(ainpo),2);
						sum2+=pow(cabs(samp),2);
					}
					else
					{
						ccc[id2-1].x=0.0f;
						ccc[id2-1].y=0.0f;
					}
				}
*************  iquad == 2  **********************
if(iquad==2)
			for(i=0;i<nsamrh;++i)
				for(j=-nsamrh;j<=(nsamrh-1);++j)
				{
					ii=i+1;
					itemp=i*i+j*j;
					jj=j+1;
					jj2=jj;
					if(jj<1)
					{
						jj+=nsam;
						jj2+=nsamr;
					}
					id=ii+nsamh*(jj-1);
					id2=ii+nsamrh*(jj2-1);

					if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
					{
						weight=exp(-rbfact*itemp);
						if(j<=0)
						{
							iii=-j+1;
							jjj=i+1;
							icon=1;
						}
						else
						{
							iii=j+1;
							jjj=-i+1;
							if(jjj<1) jjj+=nsam;
							icon=-1;
						}									
						idd=iii+nsamh*(jjj-1);
						ainpo=cbuf[idd-1];
						if(icon==-1) ainpo=conjg(ainpo);

						samp=cXr(spec[id-1],weight);
						ccc[id2-1]=cXc(samp,conjg(ainpo));
						sum+=pow(cabs(ainpo),2);
						sum2+=pow(cabs(samp),2);
					}
					else
					{
						ccc[id2-1].x=0.0f;
						ccc[id2-1].y=0.0f;
					}
				}
*************  iquad == 3  **********************
if(iquad==3)
			for(i=0;i<nsamrh;++i)
				for(j=-nsamrh;j<=(nsamrh-1);++j)
				{
					ii=i+1;
					itemp=i*i+j*j;
					jj=j+1;
					jj2=jj;
					if(jj<1)
					{
						jj+=nsam;
						jj2+=nsamr;
					}
					id=ii+nsamh*(jj-1);
					id2=ii+nsamrh*(jj2-1);

					if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
					{
						weight=exp(-rbfact*itemp);
						iii=i+1;
						jjj=j+1;
						if(jjj<1) jjj+=nsam;
						icon=-1;			
						idd=iii+nsamh*(jjj-1);
						ainpo=cbuf[idd-1];
						if(icon==-1) ainpo=conjg(ainpo);

						samp=cXr(spec[id-1],weight);
						ccc[id2-1]=cXc(samp,conjg(ainpo));
						sum+=pow(cabs(ainpo),2);
						sum2+=pow(cabs(samp),2);
					}
					else
					{
						ccc[id2-1].x=0.0f;
						ccc[id2-1].y=0.0f;
					}
				}
*************  iquad == 4  **********************
if(iquad==4)
			for(i=0;i<nsamrh;++i)
				for(j=-nsamrh;j<=(nsamrh-1);++j)
				{
					ii=i+1;
					itemp=i*i+j*j;
					jj=j+1;
					jj2=jj;
					if(jj<1)
					{
						jj+=nsam;
						jj2+=nsamr;
					}
					id=ii+nsamh*(jj-1);
					id2=ii+nsamrh*(jj2-1);

					if(itemp>=1 && itemp<maxr22 && abs(j)!=(jc-1))
					{
						weight=exp(-rbfact*itemp);

						if(j>0)
						{
							iii=j+1;
							jjj=-i+1;
							if(jjj<1)jjj+=nsam;
							icon=1;
						}
						else
						{
							iii=-j+1;
							jjj=i+1;
							icon=-1;
						}				
						idd=iii+nsamh*(jjj-1);
						ainpo=cbuf[idd-1];
						if(icon==-1) ainpo=conjg(ainpo);

						samp=cXr(spec[id-1],weight);
						ccc[id2-1]=cXc(samp,conjg(ainpo));
						sum+=pow(cabs(ainpo),2);
						sum2+=pow(cabs(samp),2);
					}
					else
					{
						ccc[id2-1].x=0.0f;
						ccc[id2-1].y=0.0f;
					}
				}

*/










