#include "ctfapply_gpu.h"
#include "ctfapply_kernel.h"
#include "maskenv_gpu.h"
#include "maskcos_gpu.h"


extern pthread_key_t cuMemKey;
/*
__device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

__device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

__device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

__device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}


__device__ CPLX ctf(float cs, float wl, float wgh1, float wgh2,
					float dfmid1, float dfmid2, float angast, float thetatr,
					int ix, int iy, float tx, float ty)
{
	float twopi=6.2831853071794;
	int rad;
	float angle, angspt, c1, c2, angdif, ccos, df, chi, schi, cchi;
	CPLX rctf;
	
	rad=ix*ix+iy*iy;
	if(rad!=0)
	{
		angle=sqrt(float(rad))*thetatr;
		angspt=atan2(float(iy),float(ix));
		c1=twopi*angle*angle/(2.0f*wl);
		c2=c1*cs*angle*angle/2.0f;
		angdif=angspt-angast;
		ccos=cos(2.0f*angdif);
		//Positive defocus in FREALIGN is underfocus, convention used
      //in this formula is that underfocus should be negative.
      //=> Take negative of defocus in following line
		df=-0.5f*(dfmid1+dfmid2+ccos*(dfmid1-dfmid2));
		chi=c1*df+c2;
		//Beam tilt
		chi=chi+twopi*cs*angle*angle*(ix*tx+iy*ty)*thetatr/wl/1000.0f;
		schi=sin(chi);
		cchi=cos(chi);
		rctf.x=wgh1*schi-wgh2*cchi;
		rctf.y=-wgh1*cchi-wgh2*schi;
	}
	else
	{
      //CTF=CMPLX(-WGH2,-WGH1)
      //The imaginary part, -WGH1, will cancel when the left and right
      //beams are added later on => set to zero
		rctf.x=-wgh2;
		rctf.y=0.0f;       	
	}
	return rctf;
}

//###########################################################
//the block size must be 512 and image size be less than 512
__global__ void calculate_sum(float *outd, float *dsum, int *disum,
								int nsam, float ri2p, float ri3p, int nmid, float scal)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int ii=blockIdx.y*blockDim.y+threadIdx.y+1;
	int id;
	int sid=threadIdx.x*blockDim.y+threadIdx.y;
	float crad2,temp;
	
	__shared__ float sum[512];
	__shared__ int isum[512];
	
	sum[sid]=0.0f;
	isum[sid]=0;
	
	if(i>nsam || ii>nsam) return;
	
	id=ii+(nsam+2)*(i-1);   // change nsam to nsam+2 by xueming
	crad2=pow(float(ii-nmid),2)+pow(float(i-nmid),2);
	
	temp=outd[id-1]/scal;  //move in by xueming
	outd[id-1]=temp;
	
	if(crad2>ri2p && crad2<=ri3p)
	{
		sum[sid]=temp;
		isum[sid]=1;
	}
	else
	{
		sum[sid]=0.0f;
		isum[sid]=0;
	}
	
	__syncthreads();
	
	//tree-type sum
	if(sid<256)
	{
		sum[sid]+=sum[sid+256];
		isum[sid]+=isum[sid+256];
	}
	__syncthreads();
	if(sid<128)
	{
		sum[sid]+=sum[sid+128];
		isum[sid]+=isum[sid+128];
	}
	__syncthreads();
	if(sid<64)
	{
		sum[sid]+=sum[sid+64];
		isum[sid]+=isum[sid+64];
	}
	__syncthreads();
	if(sid<32)
	{
		sum[sid]+=sum[sid+32];
		isum[sid]+=isum[sid+32];
	}
	__syncthreads();
	if(sid<16)
	{
		sum[sid]+=sum[sid+16];
		isum[sid]+=isum[sid+16];
	}
	__syncthreads();
	if(sid<8)
	{
		sum[sid]+=sum[sid+8];
		isum[sid]+=isum[sid+8];
	}
	__syncthreads();
	if(sid<4)
	{
		sum[sid]+=sum[sid+4];
		isum[sid]+=isum[sid+4];
	}
	__syncthreads();
	if(sid<2)
	{
		sum[sid]+=sum[sid+2];
		isum[sid]+=isum[sid+2];
	}
	__syncthreads();
	if(sid<1)
	{
		sum[sid]+=sum[sid+1];
		isum[sid]+=isum[sid+1];
		sid=blockIdx.x*gridDim.y+blockIdx.y;
		dsum[sid]=sum[0];
		disum[sid]=isum[0];
	}
}

__global__ void calsumOnGPU(float *dsum, int *disum)
{
	int sid=threadIdx.x;
	__shared__ float sum[512];
	__shared__ int isum[512];
	sum[sid]=dsum[sid];
	isum[sid]=disum[sid];
	__syncthreads();
	
	//tree-type sum
	if(sid<256)
	{
		sum[sid]+=sum[sid+256];
		isum[sid]+=isum[sid+256];
	}
	__syncthreads();
	if(sid<128)
	{
		sum[sid]+=sum[sid+128];
		isum[sid]+=isum[sid+128];
	}
	__syncthreads();
	if(sid<64)
	{
		sum[sid]+=sum[sid+64];
		isum[sid]+=isum[sid+64];
	}
	__syncthreads();
	if(sid<32)
	{
		sum[sid]+=sum[sid+32];
		isum[sid]+=isum[sid+32];
	}
	__syncthreads();
	if(sid<16)
	{
		sum[sid]+=sum[sid+16];
		isum[sid]+=isum[sid+16];
	}
	__syncthreads();
	if(sid<8)
	{
		sum[sid]+=sum[sid+8];
		isum[sid]+=isum[sid+8];
	}
	__syncthreads();
	if(sid<4)
	{
		sum[sid]+=sum[sid+4];
		isum[sid]+=isum[sid+4];
	}
	__syncthreads();
	if(sid<2)
	{
		sum[sid]+=sum[sid+2];
		isum[sid]+=isum[sid+2];
	}
	__syncthreads();
	if(sid<1)
	{
		sum[sid]+=sum[sid+1];
		isum[sid]+=isum[sid+1];
		if(isum[0]!=0)	dsum[0]=sum[0]/isum[0];
	}

	
}

__global__ void calculate_cosedge(float *outd, float *dsum, int nsam, float rihp,
							float ri2p, float ri3p, float halfw, int nmid)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int ii=blockIdx.y*blockDim.y+threadIdx.y+1;
	float sum=dsum[0];
	int id;
	float edge;
	float PI=3.1415926535897;
	
	if(i>nsam || ii>nsam) return;
	
	id=ii+(nsam+2)*(i-1);  // change nsam to nsam+2 by xueming
	edge=pow(float(ii-nmid),2)+pow(float(i-nmid),2);
	if(edge>ri2p && edge<=ri3p)
	{
		edge=sqrt(edge);
		edge=(1.0f+cos(PI*(edge-rihp)/halfw))/2.0f;
		outd[id-1]=outd[id-1]*edge+(1.0-edge)*sum;				
	}
	else if(edge>ri3p)
	{
		outd[id-1]=sum;
	}

}

void maskcos(int nsam, float *outd, float ri2, float ri3, 
					float rih, float halfw, float amagp, float scal)
{
	//Applies cosine-edged mask to 2D image
	
	int nmid;
	float rihp, ri2p, ri3p;
	
	dim3 dimBlock(16,32);
	dim3 dimGrid(nsam/dimBlock.x+1,nsam/dimBlock.y+1);
	
	rihp=rih/amagp;
	ri2p=ri2/amagp/amagp;
	ri3p=ri3/amagp/amagp;
	nmid=nsam/2+1;
	
	float *dsum;
	int *disum;
	cudaMalloc((void **)&dsum,sizeof(float)*512);	
	cudaMemset(dsum,0,sizeof(float)*512);
	cudaMalloc((void **)&disum,sizeof(int)*512);
	cudaMemset(disum,0,sizeof(int)*512);
	
	cudaThreadSynchronize();
	
	calculate_sum<<<dimGrid,dimBlock>>>(outd, dsum, disum, nsam, ri2p, ri3p, nmid, scal);
	cudaThreadSynchronize();
	
	calsumOnGPU<<<1,512>>>(dsum, disum);
	cudaThreadSynchronize();
	
	//MASK PARTICLE, COSINE EDGE
	calculate_cosedge<<<dimGrid,dimBlock>>>(outd, dsum, nsam, rihp, ri2p, ri3p, 
														halfw, nmid);
	cudaThreadSynchronize();
		
	cudaFree(dsum);
	cudaFree(disum);
}

//###########################################################

//###########################################################
__global__ void outdXw(float *outd, float *w, int nsam)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y;
	if(i>=nsam || j>=nsam) return;
	int idd=i*(nsam+2)+j;
	int ids=i*nsam+j;
	outd[idd]*=w[ids];
}

__global__ void calculateW_1(float *w, float *b3dv, float dmi0, float dmi1, 
						float dmi3, float dmi4, float dmi6, float dmi7,
						int nsam, float xh, float yh, float zh, float ri2,
						float pshx, float pshy)
{
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	int i=id%nsam;
	int j=((id-i)/nsam)%nsam;
	int k=((id-i)/nsam-j)/nsam;	
	i++;
	j++;
	k++;
	float rad2;
	int ixp,iyp,ixpp1,iypp1;
	
	if(i>nsam || j>nsam || k>nsam) return;
	
	rad2=(i-xh)*(i-xh)+(j-yh)*(j-yh)+(k-zh)*(k-zh);
	if(rad2<=ri2)
	{
		//id=i+nsam*((j-1)+nsam*(k-1));
		if(b3dv[id-1]!=0.0f)
		{
			ixp=int(dmi0*(i-xh)+dmi3*(j-yh)+dmi6*(k-zh)+xh+pshx);
			iyp=int(dmi1*(i-xh)+dmi4*(j-yh)+dmi7*(k-zh)+yh+pshy);
			ixpp1=ixp+1;
			iypp1=iyp+1;
			if(ixp>=1 && iyp>=1 && ixpp1<=nsam && iypp1<=nsam)
			{		
				id=ixp+nsam*(iyp-1);
				w[id-1]=1.0f;
				id=ixp+nsam*(iypp1-1);
				w[id-1]=1.0f;
				id=ixpp1+nsam*(iyp-1);
				w[id-1]=1.0f;
				id=ixpp1+nsam*(iypp1-1);
				w[id-1]=1.0f;
							
			}
		}
	}
			
}

__global__ void calculateW_2(float *w, int nsam)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int j=blockIdx.y*blockDim.y+threadIdx.y+1;
	int id,nsam2,m,n;
	float edge,rad2;
	float PI=3.1415926535897;
	
	//int irad=4;  ???
	if(i>nsam || j>nsam) return;
	nsam2=nsam*nsam;
	id=i+nsam*(j-1);
	if(w[id-1]==1.0f)
	{
		#pragma unroll
		for(m=-4;m<=4;m++)
		{
			#pragma unroll
			for(n=-4;n<=4;n++)
			{
				rad2=m*m+n*n;
				rad2=sqrt(rad2);
				edge=(1.0+cos(PI*rad2/4))/2.0f;
				id=i+m+nsam*(j+n-1);
				if(id>=1 && id<=nsam2 && rad2<=4)
				{
					if(w[id-1]<edge) w[id-1]=edge;
				}
			}
		}
	}

}

void maskenv(int nsam, float ri, float *outd, float *b3dv,
				float shx, float shy, float phi,float theta, float psi, float *w,
				int ibuf, cuMem *pGM)
{
	// If XSTD is negative (see CALL), masks the extracted matching projection 
	//  of the 3D structure using the D2MASK binary mask, and a cosine bell 
	//  shaped ring at radius RI.
	// Used in MATCH, CTFAPLLY and CTFAPPLY_PHASE_ONLY.
	

	int nsam2=nsam*nsam;
	float pshx,pshy;
	float cphi,sphi,cthe,sthe,cpsi,spsi,xh,yh,zh,ri2;
	float dmi[9];
	dim3 dimBlock(16,32);
	dim3 dimGrid(nsam/dimBlock.x+1,nsam/dimBlock.y+1);
	
	if(ibuf<0)
	{
		cudaMemcpy((void *)pGM->d_mbuf,(void *)w,sizeof(float)*nsam*nsam,cudaMemcpyHostToDevice);
		outdXw<<<dimGrid,dimBlock>>>(outd,pGM->d_mbuf,nsam2);
		cudaThreadSynchronize();
		return;
	} 

	cudaMemset((void *)pGM->d_mbuf,0,sizeof(float)*nsam*nsam);
		
	cphi=cos(-psi);  // why???? psi or phi??
	sphi=sin(-psi);
	cthe=cos(-theta);
	sthe=sin(-theta);
	cpsi=cos(-phi);  // why???? psi or phi??
	spsi=sin(-phi);
	xh=float(nsam/2)+1.0f;
	yh=float(nsam/2)+1.0f;
	zh=float(nsam/2)+1.0f;
	ri2=ri*ri;
	pshx=-shx*nsam/PI/2.0f;
	pshy=-shy*nsam/PI/2.0f;

	
	dmi[0]=cphi*cthe*cpsi-sphi*spsi;
	dmi[1]=sphi*cthe*cpsi+cphi*spsi;
	
	dmi[3]=-cphi*cthe*spsi-sphi*cpsi;
	dmi[4]=-sphi*cthe*spsi+cphi*cpsi;
	
	dmi[6]=sthe*cphi;
	dmi[7]=sthe*sphi;

	//calculate w(mbuf in main program)
	calculateW_1<<<nsam*nsam*nsam/512+1,512>>>(pGM->d_mbuf, pGM->m_plb3dv, 
							dmi[0], dmi[1], dmi[3], dmi[4], dmi[6], dmi[7], 
							nsam, xh, yh, zh, ri2, pshx, pshy);
	cudaThreadSynchronize();						
		
	calculateW_2<<<dimGrid,dimBlock>>>(pGM->d_mbuf, nsam);
	cudaMemcpy((void *)w,(void *)pGM->d_mbuf,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);
	
	//apply w to outd(which actually is pGM->d_fft2d used in ctfapply)	
	outdXw<<<dimGrid,dimBlock>>>(outd,pGM->d_mbuf,nsam);
	cudaThreadSynchronize();
}

//###########################################################


__global__ void CTFAndShift(CPLX *spec, CPLX *speq, CPLX *datc, CPLX *datq,
					CPLX *outc, CPLX *outq, CPLX *ctff, CPLX *ctfs,int nsam, 
					int jc, int ic,float shx, float shy, float cs, float wl, 
					float wgh1, float wgh2, float dfmid1, float dfmid2,
					float angast, float thetatrp, float tx, float ty)
{
	int l,ll,m,mm,id;
	float phase;
	CPLX pshft,ctfr,ctfl;
	
	l=blockIdx.x*blockDim.x+threadIdx.x+1;
	m=blockIdx.y*blockDim.y+threadIdx.y+1;
	
	if(l>jc || m>nsam) return;
	
	ll=l-1;
	mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	phase=shx*ll+shy*mm;
	pshft.x=cos(phase);
	pshft.y=sin(phase);
	ctfr=ctf(cs,wl,wgh1,wgh2,dfmid1,dfmid2,angast,thetatrp,ll,mm,tx,ty);
	ctfl=ctf(cs,wl,wgh1,wgh2,dfmid1,dfmid2,angast,thetatrp,-ll,-mm,tx,ty);
	if(l!=jc)
	{
		id=l+nsam*(m-1)/2;
		outc[id-1]=cXc(spec[id-1],cXc(cPc(ctfr,conjg(ctfl)),pshft));
		datc[id-1]=cXc(spec[id-1],pshft);
		ctff[id-1]=ctfr;
		ctff[id+ic-1]=ctfl;
	}
	else
	{
		outq[m-1]=cXc(speq[m-1],cXc(cPc(ctfr,conjg(ctfl)),pshft));
		datq[m-1]=cXc(speq[m-1],pshft);
		ctfs[m-1]=ctfr;
		ctfs[m+nsam-1]=ctfl;
	}


}


__global__ void ShiftToOrigin(CPLX *datc, CPLX *datq, CPLX *outc, CPLX *outq, 
					int nsam, int jc, int nsamh)
{
	int l=blockIdx.x*blockDim.x+threadIdx.x+1;
	int m=blockIdx.y*blockDim.y+threadIdx.y+1;
	int ll,mm,isum,id;
	float pshftr;
	if(l>jc || m>nsam) return;
	
	ll=l-1;
	mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	isum=ll+mm;
	pshftr=1.0f;
	if(isum%2!=0) pshftr=-1.0f;
	if(l!=jc)
	{
		id=l+nsamh*(m-1);
		datc[id-1]=cXr(datc[id-1],pshftr);
		outc[id-1]=cXr(outc[id-1],pshftr);
	}
	else
	{
		datq[m-1]=cXr(datq[m-1],pshftr);
		outq[m-1]=cXr(outq[m-1],pshftr);
	}
}
*/

void ctfapply(int *pnsam, CPLX *spec, CPLX *speq, 
	float *pshx, float *pshy, float *pcs, float *pwl, float *pwgh1, float *pwgh2, 
	float *pdfmid1, float *pdfmid2, float *pangast, float *pthetatr, CPLX *ctff, 
	CPLX *ctfs, float *outd, CPLX *outc, CPLX *outq, float *pamagp, 
	float *prih, float *phalfw, float *pri2, float *pri3, float *pri4, 
	float *datd, CPLX *datc, CPLX *datq, float *b3dv, float *pphi, 
	float *ptheta, float *ppsi, float *w, float *pxstd, int *pibuf, 
	float *ptx, float *pty)
{
	int nsam=*pnsam;

	float shx=*pshx;
	float shy=*pshy;
	float cs=*pcs;
	float wl=*pwl;
	float wgh1=*pwgh1;
	float wgh2=*pwgh2;
	
	float dfmid1=*pdfmid1;
	float dfmid2=*pdfmid2;
	float angast=*pangast;
	float thetatr=*pthetatr;
	
	float amagp=*pamagp;
	
	float rih=*prih;
	float halfw=*phalfw;
	float ri2=*pri2;
	float ri3=*pri3;
//	float ri4=*pri4;
	
	float phi=*pphi;
	float theta=*ptheta;
	float psi=*ppsi;
	float xstd=*pxstd;
	int ibuf=*pibuf;
	float tx=*ptx;
	float ty=*pty;
	
	int jc;
	int nsamh,ic;
	float thetatrp;
	float rihp,scal;
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);

//	timeval begin,end;
//	gettimeofday(&begin,NULL);

	thetatrp=thetatr/amagp;
	rihp=rih/amagp;
	scal=nsam*nsam/2.0f;
	//	  THETATRP is the diffraction angle for the first pixel in 
	//	  the transform of a particle sampled at the input pixel size	
	jc=nsam/2+1;
	nsamh=nsam/2;
//	nmid=nsam/2+1;
	ic=nsam*nsam/2;
	dim3 dimBlock(8,16);
	dim3 dimGrid(jc/dimBlock.x+1,nsam/dimBlock.y+1);
	cudaThreadSynchronize();  //because the synchronization was not done in pinsert_s.cu, move it here
	cudaMemcpy((void *)pGM->d_spec,(void *)spec,sizeof(float)*nsam*nsam, cudaMemcpyHostToDevice);
	cudaMemcpy((void *)pGM->d_speq,(void *)speq,sizeof(float)*nsam*2, cudaMemcpyHostToDevice);

	//CORRECT FOR CTF AND APPLY SHIFTS
	CTFAndShift<<<dimGrid,dimBlock>>>((CPLX *)pGM->d_spec, (CPLX *)pGM->d_speq, 
					(CPLX *)pGM->d_datc, (CPLX *)pGM->d_datq, (CPLX *)pGM->d_outc, 
					(CPLX *)pGM->d_outq, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, nsam, jc, ic, 
					shx, shy, cs, wl, wgh1, wgh2, dfmid1, dfmid2, angast, thetatrp, tx, ty);
	cudaThreadSynchronize();

	// do fft
	rlft3_GPU((CPLX *)pGM->d_outc,(CPLX *)pGM->d_outq,nsam,nsam,1,-1);

	//had moved the outd/scal to calculate_sum of maskcos by xueming
	//see the lastest parameters of maskcos below
	
	//cos mask
	maskcos(nsam,pGM->d_outc,ri2,ri3,rih,halfw,amagp,scal);

	if(xstd<0.0)
	{	
		maskenv(nsam,rihp,(float *)pGM->d_outc,b3dv,0.0f,0.0f,phi,theta,psi,w,ibuf,pGM);
	}
	
	//do fft
	rlft3_GPU((CPLX *)pGM->d_outc,(CPLX *)pGM->d_outq,nsam,nsam,1,1);
					
	//SHIFT PARTICLE TO ORIGIN
	ShiftToOrigin<<<dimGrid,dimBlock>>>((CPLX *)pGM->d_datc,(CPLX *)pGM->d_datq, 
							(CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, nsam, jc, nsamh);
	cudaThreadSynchronize();

	//copy the results back to host memory
	cudaMemcpy((void *)datc,(void *)pGM->d_datc,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)datq,(void *)pGM->d_datq,sizeof(float)*nsam*2,cudaMemcpyDeviceToHost);
	
	cudaMemcpy((void *)outc,(void *)pGM->d_outc,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)outq,(void *)pGM->d_outq,sizeof(float)*nsam*2,cudaMemcpyDeviceToHost);

	cudaMemcpy((void *)ctff,(void *)pGM->d_ctff,sizeof(float)*nsam*nsam*2,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)ctfs,(void *)pGM->d_ctfs,sizeof(float)*nsam*4,cudaMemcpyDeviceToHost);	
		
//		gettimeofday(&end,NULL);
//		cout<<" CTFAPPLY time is "<<CalTime(begin,end)<<endl;
}








