#include "ctfapply_kernel.h"


static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

static __device__ float cabs(CPLX f)
{
	return sqrt(f.x*f.x+f.y*f.y);
}

static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}


static __device__ CPLX ctf(float cs, float wl, float wgh1, float wgh2,
					float dfmid1, float dfmid2, float angast, float thetatr,
					int ix, int iy, float tx, float ty)
{
	float twopi=6.2831853071796;
	float rad;
	float angle, angspt, c1, c2, angdif, ccos, df, chi, schi, cchi;
	CPLX rctf;
	
	rad=ix*ix+iy*iy;
	if(rad!=0.0)
	{
		rad=sqrt(float(rad));
		angle=rad*thetatr;
		angspt=atan2(float(iy),float(ix));
		c1=twopi*angle*angle/(2.0*wl);
		c2=c1*cs*angle*angle/2.0;
		angdif=angspt-angast;
		ccos=cos(2.0*angdif);
		//Positive defocus in FREALIGN is underfocus, convention used
      //in this formula is that underfocus should be negative.
      //=> Take negative of defocus in following line
		df=-0.5*(dfmid1+dfmid2+ccos*(dfmid1-dfmid2));
		chi=c1*df+c2;
		//Beam tilt
		chi=chi+twopi*cs*angle*angle*(ix*tx+iy*ty)*thetatr/wl/1000.0;
		schi=sin(chi);
		cchi=cos(chi);
		rctf.x=wgh1*schi-wgh2*cchi;
		rctf.y=-wgh1*cchi-wgh2*schi;
	}
	else
	{
      //CTF=CMPLX(-WGH2,-WGH1)
      //The imaginary part, -WGH1, will cancel when the left and right
      //beams are added later on => set to zero
		rctf.x=-wgh2;
		rctf.y=0.0;       	
	}
	return rctf;
}

//###########################################################


//###########################################################

//###########################################################


//###########################################################

//used in ctfapply_phase_only
__global__ void CTFAndPhaseShift(CPLX *spec, CPLX *speq, CPLX *datc, CPLX *datq,
					CPLX *outc, CPLX *outq, CPLX *ctff, CPLX *ctfs,int nsam, 
					int jc, int ic,float shx, float shy, float cs, float wl, 
					float wgh1, float wgh2, float dfmid1, float dfmid2,
					float angast, float thetatrp, float tx, float ty)
{
	int l,ll,m,mm,id;
	float phase;
	CPLX ctfv,pshft,ctfr,ctfl;
	
	l=blockIdx.x*blockDim.x+threadIdx.x+1;
	m=blockIdx.y*blockDim.y+threadIdx.y+1;
	
	if(l>jc || m>nsam) return;
	
	ll=l-1;
	mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	phase=shx*ll+shy*mm;
	pshft.x=cos(phase);
	pshft.y=sin(phase);
	ctfr=ctf(cs,wl,wgh1,wgh2,dfmid1,dfmid2,angast,thetatrp,ll,mm,tx,ty);
	ctfl=ctf(cs,wl,wgh1,wgh2,dfmid1,dfmid2,angast,thetatrp,-ll,-mm,tx,ty);
	//#######here diffrent from ctfapply#######
	//in this unweighted CTF-correction subroutine, the value of 
	//CTFV can only be either +1.0 or -1.0
	ctfv=cPc(ctfr,conjg(ctfl));
	if(cabs(ctfv)!=0) ctfv=cXr(ctfv,1/cabs(ctfv));
	//#################################
	if(l!=jc)
	{
		id=l+nsam*(m-1)/2;
		outc[id-1]=cXc(spec[id-1],cXc(ctfv,pshft));
		datc[id-1]=cXc(spec[id-1],pshft);
		ctff[id-1]=ctfr;
		ctff[id+ic-1]=ctfl;
	}
	else
	{
		outq[m-1]=cXc(speq[m-1],cXc(ctfv,pshft));
		datq[m-1]=cXc(speq[m-1],pshft);
		ctfs[m-1]=ctfr;
		ctfs[m+nsam-1]=ctfl;
	}
}

//used in ctfapply
__global__ void CTFAndShift(CPLX *spec, CPLX *speq, CPLX *datc, CPLX *datq,
					CPLX *outc, CPLX *outq, CPLX *ctff, CPLX *ctfs,int nsam, 
					int jc, int ic,float shx, float shy, float cs, float wl, 
					float wgh1, float wgh2, float dfmid1, float dfmid2,
					float angast, float thetatrp, float tx, float ty)
{
	int l,ll,m,mm,id;
	double phase;
	CPLX pshft,ctfr,ctfl;
	
	l=blockIdx.x*blockDim.x+threadIdx.x+1;
	m=blockIdx.y*blockDim.y+threadIdx.y+1;
	
	if(l>jc || m>nsam) return;
	
	ll=l-1;
	mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	phase=shx*float(ll)+shy*float(mm);
	pshft.x=cos(phase);
	pshft.y=sin(phase);
	ctfr=ctf(cs,wl,wgh1,wgh2,dfmid1,dfmid2,angast,thetatrp,ll,mm,tx,ty);
	ctfl=ctf(cs,wl,wgh1,wgh2,dfmid1,dfmid2,angast,thetatrp,-ll,-mm,tx,ty);
	if(l!=jc)
	{
		id=l+nsam/2*(m-1);
		outc[id-1]=cXc(spec[id-1],cXc(cPc(ctfr,conjg(ctfl)),pshft));
		datc[id-1]=cXc(spec[id-1],pshft);
		ctff[id-1]=ctfr;
		ctff[id+ic-1]=ctfl;
	}
	else
	{
		outq[m-1]=cXc(speq[m-1],cXc(cPc(ctfr,conjg(ctfl)),pshft));
		datq[m-1]=cXc(speq[m-1],pshft);
		ctfs[m-1]=ctfr;
		ctfs[m+nsam-1]=ctfl;
	}

}

__global__ void ShiftToOrigin(CPLX *datc, CPLX *datq, CPLX *outc, CPLX *outq, 
					int nsam, int jc, int nsamh)
{
	int l=blockIdx.x*blockDim.x+threadIdx.x+1;
	int m=blockIdx.y*blockDim.y+threadIdx.y+1;
	int ll,mm,isum,id;
	float pshftr;
	if(l>jc || m>nsam) return;
	
	ll=l-1;
	mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	isum=ll+mm;
	pshftr=1.0f;
	if(isum%2!=0) pshftr=-1.0f;
	if(l!=jc)
	{
		id=l+nsamh*(m-1);
		datc[id-1]=cXr(datc[id-1],pshftr);
		outc[id-1]=cXr(outc[id-1],pshftr);
	}
	else
	{
		datq[m-1]=cXr(datq[m-1],pshftr);
		outq[m-1]=cXr(outq[m-1],pshftr);
	}
}

