#include "ctfapply_phase_only_gpu.h"

#include "ctfapply_kernel.h"
#include "maskenv_gpu.h"
#include "maskcos_gpu.h"

extern pthread_key_t cuMemKey;


void ctfapply_phase_only(int nsam, CPLX *spec, CPLX *speq, 
	float shx, float shy, float cs, float wl, float wgh1, float wgh2, 
	float dfmid1, float dfmid2, float angast, float thetatr, CPLX *ctff, 
	CPLX *ctfs, float *outd, CPLX *outc, CPLX *outq, float amagp, 
	float rih, float halfw, float ri2, float ri3, float ri4, 
	float *datd, CPLX *datc, CPLX *datq, float *b3dv, float phi, 
	float theta, float psi, float *w, float xstd, int ibuf, 
	float tx, float ty)
{
	int jc;
	int nsamh,ic;
	float thetatrp;
	float rihp,scal;
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
//	timeval begin,end;
//	gettimeofday(&begin,NULL);

	thetatrp=thetatr/amagp;
	rihp=rih/amagp;
	scal=nsam*nsam/2.0f;
	//	  THETATRP is the diffraction angle for the first pixel in 
	//	  the transform of a particle sampled at the input pixel size	
	jc=nsam/2+1;
	nsamh=nsam/2;
//	nmid=nsam/2+1;
	ic=nsam*nsam/2;
	dim3 dimBlock(16,16);
	dim3 dimGrid(jc/dimBlock.x+1,nsam/dimBlock.y+1);
	
	cudaMemcpy((void *)pGM->d_spec,(void *)spec,sizeof(float)*nsam*nsam, cudaMemcpyHostToDevice);
	cudaMemcpy((void *)pGM->d_speq,(void *)speq,sizeof(float)*nsam*2, cudaMemcpyHostToDevice);
	
	//CORRECT FOR CTF AND APPLY SHIFTS
	CTFAndPhaseShift<<<dimGrid,dimBlock>>>((CPLX *)pGM->d_spec, (CPLX *)pGM->d_speq, 
					(CPLX *)pGM->d_datc, (CPLX *)pGM->d_datq, (CPLX *)pGM->d_outc, 
					(CPLX *)pGM->d_outq, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, nsam, jc, ic, 
					shx, shy, cs, wl, wgh1, wgh2, dfmid1, dfmid2, angast, thetatrp, tx, ty);			
	
	// do fft
	rlft3_GPU((CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, nsam,nsam,1,-1);
	
	//had moved the outd/scal to calculate_sum of maskcos by xueming
	//see the lastest parameters of maskcos below
	
	//cos mask
	maskcos(nsam,pGM->d_fft2d,ri2,ri3,rih,halfw,amagp,scal);
	
	if(xstd<0.0)
	{	
		maskenv(nsam,rihp,(float *)pGM->d_fft2d,b3dv,0.0f,0.0f,phi,theta,psi,w,ibuf,pGM);
	}
	
	//do fft
	rlft3_GPU((CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, nsam,nsam,1,1);
	
					
	//SHIFT PARTICLE TO ORIGIN
	ShiftToOrigin<<<dimGrid,dimBlock>>>((CPLX *)pGM->d_datc,(CPLX *)pGM->d_datq, 
							(CPLX *)pGM->d_outc, (CPLX *)pGM->d_outq, nsam, jc, nsamh);
		
	//copy the results back to host memory
	cudaMemcpy((void *)datc,(void *)pGM->d_datc,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)datq,(void *)pGM->d_datq,sizeof(float)*nsam*2,cudaMemcpyDeviceToHost);
	
	cudaMemcpy((void *)outc,(void *)pGM->d_outc,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)outq,(void *)pGM->d_outq,sizeof(float)*nsam*2,cudaMemcpyDeviceToHost);
	
	cudaMemcpy((void *)ctff,(void *)pGM->d_ctff,sizeof(float)*nsam*nsam*2,cudaMemcpyDeviceToHost);	
	cudaMemcpy((void *)ctfs,(void *)pGM->d_ctfs,sizeof(float)*nsam*4,cudaMemcpyDeviceToHost);	
		
//		gettimeofday(&end,NULL);
//		cout<<" CTFAPPLY time is "<<CalTime(begin,end)<<endl;
}








