#include "magrefine_gpu.h"
#include "va04_gpu.h"

/* ************************************************************************** */
/*  Refines MAGNIFICATION using Powell minimiser on just 1 parameter */
/*  Calls VA04. */
/*  Used in LMAIN. */
/* ************************************************************************** */
int magrefine(float *amagp, float *dfmid1, float *dfmid2, float *angast, 
	int ifirstf, int lastf, int np, int nsam, int maxr1, int maxr2, int *mask, 
	int nn1, CPLX *c3df, CPLX *c3ds, int irad, CPLX *outq, int npb, 
	float *pbuf, CPLX *pbuq, float *shx, float *shy, int ipad, float cs, 
	float wl, float wgh1, float wgh2, float thetatr, CPLX *ctff, 
	CPLX *ctfs, float amag, float rih, float halfw, float ri2, float ri3, 
	float ri4, float *phi, float *theta, float *psi, float rmax1, float rmax2, 
	float xstd, float *mbuf, int *ilst, CPLX *datc, CPLX *datq, int ibuf, 
	float *b3dv, float *datd, float *sinclut, int ivflag, float rbfact, 
	float *outd, CPLX *outc, CPLX *qbuc, CPLX *qbuq, float *rbuf, float *fsct, 
	int *film, int ilast,int iewald, float tx, float ty)
{
    float epar[6] = { 0.01f,0.0f,0.0f,0.0f,0.0f,0.0f };
    float escale = 100.0f;
	 int ncycls=50;
    int i;
    float xp[6];
    float xpar[6], presa;

/* ************************************************************************** */
/* above assumes relative magnification is near 1.0, and steps more than 0.1 */
/*   are not allowed with final accuracy +/- 0.001 (i.e. 1 part in 1000) */

    printf("Entering MAGREFINE with range %d %d\n",ifirstf,lastf);
    xpar[0] = amagp[ifirstf];
    xpar[1] = dfmid1[ifirstf];
    xpar[2] = dfmid2[ifirstf];
    xpar[3] = angast[ifirstf];
	
	 int c1=1;
	 int c2=0;
	 float c3=0.0;
    va04a(xpar, epar, c1, presa, escale, c2, c1, ncycls, xp, np, 
	    nsam, maxr1, maxr2, mask, nn1, c3df, c3ds, irad, outq, npb, 
	    pbuf, pbuq, shx, shy, ipad, cs, wl, 
	    wgh1, wgh2, thetatr, ctff, ctfs, &amag, rih, halfw, ri2, 
	    ri3, ri4, phi, theta, psi, rmax1, rmax2, xstd, mbuf, ilst, 
	    datc, datq, ibuf, b3dv, datd, 
	    sinclut, ivflag, rbfact, outd, outc, qbuc, qbuq, 
	    rbuf, fsct, dfmid1, dfmid2, angast, iewald, 
	    tx, ty, c3, c3, c3, c3, c3);

    for (i=ifirstf; i<=ilast; i++) 
    {
		if(film[ifirstf-1] == film[i-1]) amagp[i-1] = xpar[0];
		else break;
    }

    printf("Exit from MAGREFINE, relative magnification %f\n", xpar[0]);
    return 0;
}

