#include "maskcos_gpu.h"
extern pthread_key_t cuMemKey;

//the block size must be 512 and image size be less than 512
__global__ void calculate_sum(float *outd, float *dsum, int *disum,
								int nsam, float ri2p, float ri3p, int nmid, float scal)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int ii=blockIdx.y*blockDim.y+threadIdx.y+1;
	int id;
	int sid=threadIdx.x*blockDim.y+threadIdx.y;
	float crad2,temp;
	
	__shared__ float sum[512];
	__shared__ int isum[512];
	
	sum[sid]=0.0f;
	isum[sid]=0;
	
	if(i<=nsam && ii<=nsam)
	{
		id=ii+nsam*(i-1);  
		crad2=pow(float(ii-nmid),2)+pow(float(i-nmid),2);
	
		temp=outd[id-1]/scal;  //move in by xueming
		outd[id-1]=temp;
	
		if(crad2>ri2p && crad2<=ri3p)
		{
			sum[sid]=temp;
			isum[sid]=1;
		}

	}
	__syncthreads();
	
	//tree-type sum
	if(sid<256)
	{
		sum[sid]+=sum[sid+256];
		isum[sid]+=isum[sid+256];
	}
	__syncthreads();
	if(sid<128)
	{
		sum[sid]+=sum[sid+128];
		isum[sid]+=isum[sid+128];
	}
	__syncthreads();
	if(sid<64)
	{
		sum[sid]+=sum[sid+64];
		isum[sid]+=isum[sid+64];
	}
	__syncthreads();
	if(sid<32)
	{
		sum[sid]+=sum[sid+32];
		isum[sid]+=isum[sid+32];
	}
	__syncthreads();
	if(sid<16)
	{
		sum[sid]+=sum[sid+16];
		isum[sid]+=isum[sid+16];
	}
	__syncthreads();
	if(sid<8)
	{
		sum[sid]+=sum[sid+8];
		isum[sid]+=isum[sid+8];
	}
	__syncthreads();
	if(sid<4)
	{
		sum[sid]+=sum[sid+4];
		isum[sid]+=isum[sid+4];
	}
	__syncthreads();
	if(sid<2)
	{
		sum[sid]+=sum[sid+2];
		isum[sid]+=isum[sid+2];
	}
	__syncthreads();
	if(sid<1)
	{
		sum[sid]+=sum[sid+1];
		isum[sid]+=isum[sid+1];
		
		sid=blockIdx.x*gridDim.y+blockIdx.y;
		dsum[sid]=sum[0];
		disum[sid]=isum[0];
	}
}

__global__ void calsumOnGPU(float *dsum, int *disum)
{
	int sid=threadIdx.x;
	__shared__ float sum[512];
	__shared__ int isum[512];
	sum[sid]=dsum[sid];
	isum[sid]=disum[sid];
	__syncthreads();
	
	//tree-type sum
	if(sid<256)
	{
		sum[sid]+=sum[sid+256];
		isum[sid]+=isum[sid+256];
	}
	__syncthreads();
	if(sid<128)
	{
		sum[sid]+=sum[sid+128];
		isum[sid]+=isum[sid+128];
	}
	__syncthreads();
	if(sid<64)
	{
		sum[sid]+=sum[sid+64];
		isum[sid]+=isum[sid+64];
	}
	__syncthreads();
	if(sid<32)
	{
		sum[sid]+=sum[sid+32];
		isum[sid]+=isum[sid+32];
	}
	__syncthreads();
	if(sid<16)
	{
		sum[sid]+=sum[sid+16];
		isum[sid]+=isum[sid+16];
	}
	__syncthreads();
	if(sid<8)
	{
		sum[sid]+=sum[sid+8];
		isum[sid]+=isum[sid+8];
	}
	__syncthreads();
	if(sid<4)
	{
		sum[sid]+=sum[sid+4];
		isum[sid]+=isum[sid+4];
	}
	__syncthreads();
	if(sid<2)
	{
		sum[sid]+=sum[sid+2];
		isum[sid]+=isum[sid+2];
	}
	__syncthreads();
	if(sid<1)
	{
		sum[sid]+=sum[sid+1];
		isum[sid]+=isum[sid+1];
		if(isum[0]!=0)	dsum[0]=sum[0]/isum[0];
		disum[0]=isum[0];
	}

	
}

__global__ void calculate_cosedge(float *outd, float *dsum, int nsam, float rihp,
							float ri2p, float ri3p, float halfw, int nmid)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int ii=blockIdx.y*blockDim.y+threadIdx.y+1;
	float sum=dsum[0];
	int id;
	float edge;
	float PI=3.1415926535897;
	
	if(i>nsam || ii>nsam) return;
	
	id=ii+nsam*(i-1); 
	edge=pow(float(ii-nmid),2)+pow(float(i-nmid),2);
	if(edge>ri2p && edge<=ri3p)
	{
		edge=sqrt(edge);
		edge=(1.0+cos(PI*(edge-rihp)/halfw))/2.0;
		outd[id-1]=outd[id-1]*edge+(1.0-edge)*sum;				
	}
	else if(edge>ri3p)
	{
		outd[id-1]=sum;
	}

}

void maskcos(int nsam, float *outd, float ri2, float ri3, 
					float rih, float halfw, float amagp, float scal)
{
	//Applies cosine-edged mask to 2D image
	
	int nmid;
	float rihp, ri2p, ri3p;
	
	dim3 dimBlock(16,32);
	dim3 dimGrid(nsam/dimBlock.x+1,nsam/dimBlock.y+1);
	
	rihp=rih/amagp;
	ri2p=ri2/amagp/amagp;
	ri3p=ri3/amagp/amagp;
	nmid=nsam/2+1;
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
	float *dsum=NULL;
	int *disum=NULL;
	//cudaMalloc((void **)&dsum,sizeof(float)*512);	
	dsum=pGM->d_msum1;
	cudaMemset(dsum,0,sizeof(float)*512);
	//cudaMalloc((void **)&disum,sizeof(int)*512);
	disum=pGM->d_misum;
	cudaMemset(disum,0,sizeof(int)*512);
	
	cudaThreadSynchronize();
	
	calculate_sum<<<dimGrid,dimBlock>>>(outd, dsum, disum, nsam, ri2p, ri3p, nmid, scal);
	cudaThreadSynchronize();
	
	calsumOnGPU<<<1,512>>>(dsum, disum);
	cudaThreadSynchronize();
	
	
	//MASK PARTICLE, COSINE EDGE
	calculate_cosedge<<<dimGrid,dimBlock>>>(outd, dsum, nsam, rihp, ri2p, ri3p, 
														halfw, nmid);
	cudaThreadSynchronize();
		
	//cudaFree(dsum);
	//cudaFree(disum);
}
