#include "maskenv_gpu.h"

//if(w[id-1]<edge) w[id-1]=edge;
static __device__ inline void atomicFloatMax(float *address, float val)
{
      int i_val = __float_as_int(val);
      int tmp0 = 0;
      int tmp1;

      while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0)
      {
              tmp0 = tmp1;
              if(__int_as_float(tmp1) >= val) i_val=tmp1;
              //i_val = __float_as_int(val + __int_as_float(tmp1));
      }
}

__global__ void outdXw(float *outd, float *w, int nsam)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y;
	if(i>=nsam || j>=nsam) return;
	int idd=i*nsam+j;
	outd[idd]*=w[idd];
}

/*__global__ void calculateW_1(float *w, float *b3dv, float dmi0, float dmi1, 
						float dmi3, float dmi4, float dmi6, float dmi7,
						int nsam, float xh, float yh, float zh, float ri2,
						float pshx, float pshy)
{
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	int i=id%nsam;
	int j=id/nsam%nsam;
	int k=id/nsam/nsam;	
	i++;
	j++;
	k++;
	float rad2;
	int ixp,iyp,ixpp1,iypp1;
	
	if(i>nsam || j>nsam || k>nsam) return;
	
	rad2=(i-xh)*(i-xh)+(j-yh)*(j-yh)+(k-zh)*(k-zh);
	if(rad2<=ri2)
	{
		//id=i+nsam*((j-1)+nsam*(k-1));
		if(b3dv[id]!=0.0f)
		{
			ixp=int(dmi0*(i-xh)+dmi3*(j-yh)+dmi6*(k-zh)+xh+pshx);
			iyp=int(dmi1*(i-xh)+dmi4*(j-yh)+dmi7*(k-zh)+yh+pshy);
			ixpp1=ixp+1;
			iypp1=iyp+1;
			if(ixp>=1 && iyp>=1 && ixpp1<=nsam && iypp1<=nsam)
			{		
				id=ixp+nsam*(iyp-1);
				w[id-1]=1.0f;
				id=ixp+nsam*(iypp1-1);
				w[id-1]=1.0f;
				id=ixpp1+nsam*(iyp-1);
				w[id-1]=1.0f;
				id=ixpp1+nsam*(iypp1-1);
				w[id-1]=1.0f;
							
			}
		}
	}
			
}*/

//2D version of calculateW_1, used for big matrix
//block dimension is 256*1
//each row has 64 blocks
//so, each row has 256*64=16384 threads
__global__ void calculateW_1(float *w, float *b3dv, float dmi0, float dmi1, 
						float dmi3, float dmi4, float dmi6, float dmi7,
						int nsam, float xh, float yh, float zh, float ri2,
						float pshx, float pshy)
{
	int id=(blockIdx.x*blockDim.x+threadIdx.x)+blockIdx.y*16384;
	int i=id%nsam;
	int j=id/nsam%nsam;
	int k=id/nsam/nsam;	
	i++;
	j++;
	k++;
	float rad2;
	int ixp,iyp,ixpp1,iypp1;
	
	if(i>nsam || j>nsam || k>nsam) return;
	
	rad2=(i-xh)*(i-xh)+(j-yh)*(j-yh)+(k-zh)*(k-zh);
	if(rad2<=ri2)
	{
		//id=i+nsam*((j-1)+nsam*(k-1));
		if(b3dv[id]!=0.0f)
		{
			ixp=int(dmi0*(i-xh)+dmi3*(j-yh)+dmi6*(k-zh)+xh+pshx);
			iyp=int(dmi1*(i-xh)+dmi4*(j-yh)+dmi7*(k-zh)+yh+pshy);
			ixpp1=ixp+1;
			iypp1=iyp+1;
			if(ixp>=1 && iyp>=1 && ixpp1<=nsam && iypp1<=nsam)
			{		
				id=ixp+nsam*(iyp-1);
				w[id-1]=1.0f;
				id=ixp+nsam*(iypp1-1);
				w[id-1]=1.0f;
				id=ixpp1+nsam*(iyp-1);
				w[id-1]=1.0f;
				id=ixpp1+nsam*(iypp1-1);
				w[id-1]=1.0f;
							
			}
		}
	}
			
}

__global__ void calculateW_2(float *w, int nsam)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int j=blockIdx.y*blockDim.y+threadIdx.y+1;
	int id,nsam2,m,n;
	float edge,rad2;
	float PI=3.1415926535897;
	
	//int irad=4;  ???
	if(i>nsam || j>nsam) return;
	nsam2=nsam*nsam;
	id=i+nsam*(j-1);
	if(w[id-1]==1.0f)
	//if(abs(w[id-1]-1.0f)<0.000001)
	{
		for(m=-4;m<=4;m++)
			for(n=-4;n<=4;n++)
			{
				rad2=m*m+n*n;
				rad2=sqrt(rad2);
				edge=(1.0f+cos(PI*rad2/4.0f))/2.0f;
				id=i+m+nsam*(j+n-1);
				if(id>=1 && id<=nsam2 && rad2<=4)
				{
					//if(w[id-1]<edge) w[id-1]=edge;
					//if(w[id-1]<edge) atomicExch(&w[id-1],edge);
					atomicFloatMax(&w[id-1],edge);
				}
			}
	}

}

void maskenv(int nsam, float ri, float *outd, float *b3dv,
				float shx, float shy, float phi,float theta, float psi, float *w,
				int ibuf, cuMem *pGM)
{
	// If XSTD is negative (see CALL), masks the extracted matching projection 
	//  of the 3D structure using the D2MASK binary mask, and a cosine bell 
	//  shaped ring at radius RI.
	// Used in MATCH, CTFAPLLY and CTFAPPLY_PHASE_ONLY.
	

	int nsam2=nsam*nsam;
	float pshx,pshy;
	float cphi,sphi,cthe,sthe,cpsi,spsi,xh,yh,zh,ri2;
	float dmi[9];
	dim3 dimBlock(16,16);
	dim3 dimGrid(nsam/dimBlock.x+1,nsam/dimBlock.y+1);
	
	if(ibuf<0)
	{
		cudaMemcpy((void *)pGM->d_mbuf,(void *)w,sizeof(float)*nsam*nsam,cudaMemcpyHostToDevice);
		outdXw<<<dimGrid,dimBlock>>>(outd,pGM->d_mbuf,nsam2);
		cudaThreadSynchronize();
		return;
	} 

	cudaMemset((void *)pGM->d_mbuf,0,sizeof(float)*nsam*nsam);
		
	cphi=cos(-psi);  // why???? psi or phi??
	sphi=sin(-psi);
	cthe=cos(-theta);
	sthe=sin(-theta);
	cpsi=cos(-phi);  // why???? psi or phi??
	spsi=sin(-phi);
	xh=float(nsam/2)+1.0f;
	yh=float(nsam/2)+1.0f;
	zh=float(nsam/2)+1.0f;
	ri2=ri*ri;
	pshx=-shx*nsam/PI/2.0f;
	pshy=-shy*nsam/PI/2.0f;

	
	dmi[0]=cphi*cthe*cpsi-sphi*spsi;
	dmi[1]=sphi*cthe*cpsi+cphi*spsi;
	
	dmi[3]=-cphi*cthe*spsi-sphi*cpsi;
	dmi[4]=-sphi*cthe*spsi+cphi*cpsi;
	
	dmi[6]=sthe*cphi;
	dmi[7]=sthe*sphi;

	dim3 DimBlockBig(256,1);
	dim3 DimGridnnn(64,nsam*nsam*nsam/16384+1);
	//calculate w(mbuf in main program)
	//calculateW_1<<<nsam*nsam*nsam/256+1,256>>>(pGM->d_mbuf, pGM->m_plb3dv, 
	calculateW_1<<<DimGridnnn,DimBlockBig>>>(pGM->d_mbuf, pGM->m_plb3dv,
							dmi[0], dmi[1], dmi[3], dmi[4], dmi[6], dmi[7], 
							nsam, xh, yh, zh, ri2, pshx, pshy);
	cudaThreadSynchronize();						

	calculateW_2<<<dimGrid,dimBlock>>>(pGM->d_mbuf, nsam);
	cudaThreadSynchronize();
	cudaMemcpy((void *)w,(void *)pGM->d_mbuf,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);

	
	//apply w to outd
	outdXw<<<dimGrid,dimBlock>>>(outd,pGM->d_mbuf,nsam);
	cudaThreadSynchronize();
}
