#include "pad.h"

extern pthread_key_t cuMemKey;

__global__ void pad_1(float *out, float *in, int size, double scal)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	if(i>=size) return;
	out[i]=double(in[i])/scal;
}

__global__ void pad_2(float *out, int size, float circ)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	if(i>=size) return;
	//out[i]+=circ;
	out[i]=circ;
}


/*	for(j=1;j<=nsam;++j)
		for(i=1;i<=nsam;++i)
		{
			jj=j;
			if(jj>nsamh) jj+=nsam4-nsam;
			ii=i;
			if(ii>nsamh) ii+=nsam4-nsam;
			id=i+nsam*(j-1);
			id4=ii+nsam4*(jj-1);
			a=outd[id-1]-circ;
			outd[id-1]=circ;
			outd[id4-1]=a;
		}*/

__global__ void pad_loop(float *toutd, float *outd, int nsam,int nsam4)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x+1;
	int j=blockIdx.y*blockDim.y+threadIdx.y+1;
	int ii,jj,id, id4;
	if(i>nsam || j>nsam) return;
	
	jj=j;
	if(jj>(nsam/2)) jj+=nsam4-nsam;
	ii=i;
	if(ii>(nsam/2)) ii+=nsam4-nsam;
	id=i+nsam*(j-1);
	id4=ii+nsam4*(jj-1);
	
	outd[id4-1]=toutd[id-1];
}

void pad(int nsam, int ipad, float *data, CPLX *speq, float *outd, CPLX *outq)
{
	int i,j,ii,jj,id,id4,nsam4,nsamh;
	float circ,a;
	double scal;
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	float *toutd=pGM->d_fft2d;
	
	nsam4=ipad*nsam;
	nsamh=nsam/2;
	scal=nsam*nsam/2;
	float *tempd=new float[nsam*nsam];
	//cudaMemset(outd,0,sizeof(float)*nsam4*nsam4);
	//cudaMemset(toutd,0,sizeof(float)*nsam*nsam);

	/*for(i=0;i<nsam*nsam;++i)
	{
		outd[i]=data[i]/scal;
	}
	for(i=0;i<nsam;++i)
	{
		outq[i].x=speq[i].x/scal;
		outq[i].y=speq[i].y/scal;
	}*/

	pad_1<<<nsam*nsam/256+1,256>>>(toutd,data,nsam*nsam,scal);
	pad_1<<<nsam*2/128+1,128>>>((float *)outq,(float *)speq,nsam*2,scal);
	cudaThreadSynchronize();

	//i=1;
	//j=-1;
	//rlft3_((CPLX *)outd,outq,&nsam,&nsam,&i,&j);
	rlft3_GPU((CPLX *)toutd,outq,nsam,nsam,1,-1);

	cudaMemcpy(tempd,toutd,sizeof(float)*nsam*nsam,cudaMemcpyDeviceToHost);	
	circ=0.0f;
	for(i=1;i<=nsam;++i)
		for(j=1;j<=nsam;++j)
		{
			if(i==(nsamh+1) || j==(nsamh+1) || i==nsamh || j==nsamh)
			{
				id=i+nsam*(j-1);
				circ+=tempd[id-1];
			}
		}
	circ/=(4*nsam-4);

	/*for(i=0;i<nsam4*nsam4;++i)
	{
		outd[i]+=circ;
	}*/
	pad_2<<<nsam4*nsam4/256+1,256>>>(outd,nsam4*nsam4,circ);
	cudaThreadSynchronize();

	/*for(j=nsam;j>=1;--j)
		for(i=nsam;i>=1;--i)
		{
			jj=j;
			if(jj>nsamh) jj+=nsam4-nsam;
			ii=i;
			if(ii>nsamh) ii+=nsam4-nsam;
			id=i+nsam*(j-1);
			id4=ii+nsam4*(jj-1);
			a=outd[id-1]-circ;
			outd[id-1]=circ;
			outd[id4-1]=a;
		}*/

	dim3 DimBlock(16,16);
	pad_loop<<<dim3(nsam/DimBlock.x+1,nsam/DimBlock.y+1),DimBlock>>>(toutd,outd, nsam,nsam4);
	cudaThreadSynchronize();

	//i=1;
	//j=1;
	//rlft3_((CPLX *)outd,outq,&nsam4,&nsam4,&i,&j);
	rlft3_GPU((CPLX *)outd,outq,nsam4,nsam4,1,1);
		
	delete [] tempd;
}


/*
void pad(int nsam, int ipad, float *data, CPLX *speq, float *outd, CPLX *outq)
{
	int i,j,ii,jj,id,id4,nsam4,nsamh;
	float circ,scal,a;
	
	nsam4=ipad*nsam;
	nsamh=nsam/2;
	scal=nsam*nsam/2;
	
	memset(outd,0,sizeof(float)*nsam4*nsam4);

	for(i=0;i<nsam*nsam;++i)
	{
		outd[i]=data[i]/scal;
	}
	for(i=0;i<nsam;++i)
	{
		outq[i].x=speq[i].x/scal;
		outq[i].y=speq[i].y/scal;
	}
	
	i=1;
	j=-1;
	rlft3_((CPLX *)outd,outq,&nsam,&nsam,&i,&j);
	
	circ=0.0f;
	for(i=1;i<=nsam;++i)
		for(j=1;j<nsam;++j)
		{
			if(i==(nsamh+1) || j==(nsamh+1) || i==nsamh || j==nsamh)
			{
				id=i+nsam*(j-1);
				circ+=outd[id-1];
			}
		}
	circ/=(4*nsam-4);
	
	for(i=0;i<nsam4*nsam4;++i)
	{
		outd[i]+=circ;
	}
	
	for(j=nsam;j>=1;--j)
		for(i=nsam;i>=1;--i)
		{
			jj=j;
			if(jj>nsamh) jj+=nsam4-nsam;
			ii=i;
			if(ii>nsamh) ii+=nsam4-nsam;
			id=i+nsam*(j-1);
			id4=ii+nsam4*(jj-1);
			a=outd[id-1]-circ;
			outd[id-1]=circ;
			outd[id4-1]=a;
		}
		
	i=1;
	j=1;

	rlft3_((CPLX *)outd,outq,&nsam4,&nsam4,&i,&j);

}
*/

