#include "pextract_gpu.h"

extern pthread_key_t cuMemKey;

	texture<CPLX, 3, cudaReadModeElementType> texRef_c3df;	
	texture<CPLX, 2, cudaReadModeElementType> texRef_c3ds;
	texture<float, 1, cudaReadModeElementType> texRef_sinclut;
	texture<CPLX, 2, cudaReadModeElementType> texRef_ctff;
	texture<CPLX, 1, cudaReadModeElementType> texRef_ctfs;

static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}

static __device__ float boxft_lut(float *arg,float *sinclut)
{
	int i,j;
	float sinc,prod=1.0f;
	
	for(i=0;i<3;i++)
	{
		arg[i]=abs(arg[i]);
		if(arg[i]<0.02f)
		{
			sinc=1.0f;
		}
		else
		{
			j=int(180.0f*arg[i]+0.5f);
			//sinc=sinclut[j-1];
			sinc=tex1Dfetch(texRef_sinclut,j-1);
		}
		prod=prod*sinc;
	}
	return prod;

}

static __device__ CPLX ainterpo3dbig(int nsama, int ipad, CPLX *a3df, CPLX *a3ds, 
                                      float &x, float &y, float &z)
{
	int nsam,jc;
	int l,m,n,ll,mm,nn,id,nsamh;
	CPLX samp;
	
	nsam=ipad*nsama;
	samp.x=0.0f;
	samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	l=(x>=0.0f)?int(ipad*x+0.5f):int(ipad*x-0.5f);
	m=(y>=0.0f)?int(ipad*y+0.5f):int(ipad*y-0.5f);
	n=(z>=0.0f)?int(ipad*z+0.5f):int(ipad*z-0.5f);
	
	ll=(l>=0)?(l):(-l);
	mm=(l>=0)?(m):(-m);				
	nn=(l>=0)?(n):(-n);
	if(mm<0) mm=mm+nsam;
	if(nn<0) nn=nn+nsam;
	
	if(ll!=jc)
	{
		//id=ll+nsamh*((mm-1)+nsam*(nn-1));
		//samp=cPc(samp,(l>=0)?a3df[id-1]:conjg(a3df[id-1]));
		samp=cPc(samp,(l>=0)?tex3D(texRef_c3df,ll,mm,nn):conjg(tex3D(texRef_c3df,ll,mm,nn)));	
	}
	else
	{
		//id=mm+nsam*(nn-1);
		//samp=cPc(samp,(l>=0)?a3ds[id-1]:conjg(a3ds[id-1]));
		samp=cPc(samp,(l>=0)?tex2D(texRef_c3ds,mm,nn):conjg(tex2D(texRef_c3ds,mm,nn)));
	}

	return samp;
}

/*__device__ CPLX ainterpo3ds(int nsam, int irad, CPLX *a3df, CPLX *a3ds, 
                        float xx, float yy, float zz,float *sinclut, int &ipad)
{
	int ls,lt,ns,nt,ms,mt,jc;
	int l,m,n,ll,mm,nn,is,id,nsamh;
	float x,y,z,arg[3],boxftv;
	CPLX samp;

	samp.x=samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	x=ipad*xx;
	y=ipad*yy;
	z=ipad*zz;
	
	ls=int(x)-irad+1;
	if(x<0.0f) ls=ls-1;
	if(ls<(-jc)) ls=-jc+1;
	lt=int(x)+irad;
	if(x<0.0f) lt=lt-1;
	if(lt>=jc) lt=jc-1;
	
	ms=int(y)-irad+1;
	if(y<0.0f) ms=ms-1;
	if(ms<=(-jc)) ms=-jc+1;
	mt=int(y)+irad;
	if(y<0.0f) mt=mt-1;
	if(mt>=jc) mt=jc-1;
	
	ns=int(z)-irad+1;
	if(z<0.0f) ns=ns-1;
	if(ns<=(-jc)) ns=-jc+1;
	nt=int(z)+irad;
	if(z<0.0f) nt=nt-1;
	if(nt>=jc) nt=jc-1;
	
	for(l=ls;l<=lt;l++)
	{
		arg[0]=x-l;
		for(m=ms;m<=mt;m++)
		{
			arg[1]=y-m;
			for(n=ns;n<=nt;n++)
			{
				arg[2]=z-n;
				boxftv=boxft_lut(arg, sinclut);
				if(l>=0)
				{
					ll=l+1;
					if(ll<1) ll=1;
					mm=m+1;
					if(mm<1) mm=mm+nsam;
					nn=n+1;
					if(nn<1) nn=nn+nsam;
					if(ll!=jc)
					{
						id=ll+nsamh*((mm-1)+nsam*(nn-1));
						samp=cPc(samp,cXr(a3df[id-1],boxftv));
					}
					else
					{
						is=mm+nsam*(nn-1);
						samp=cPc(samp,cXr(a3ds[is-1],boxftv));
					}
				}
				else
				{
					ll=-l+1;
					mm=-m+1;
					if(mm<1) mm=mm+nsam;
					nn=-n+1;
					if(nn<1) nn=nn+nsam;
					if(ll!=jc)
					{
						id=ll+nsamh*((mm-1)+nsam*(nn-1));
						samp=cPc(samp,conjg(cXr(a3df[id-1],boxftv)));
					}
					else
					{
						is=mm+nsam*(nn-1);
						samp=cPc(samp,conjg(cXr(a3ds[is-1],boxftv)));
					}
				}
				
			}
		}
	}
	
	return samp;
} */          

static __device__ CPLX ainterpo3ds(int nsam, int irad, CPLX *a3df, CPLX *a3ds, 
                        float &xx, float &yy, float &zz, float *sinclut, int &ipad)
{
	int ls,lt,ns,nt,ms,mt,jc;
	int l,m,n,ll,mm,nn;
	//int is,id,nsamh;
	float x,y,z,arg[3],boxftv;
	CPLX samp,c3d;
	
	samp.x=samp.y=0.0f;
	jc=nsam/2+1;
	//nsamh=nsam/2;
	x=ipad*xx;
	y=ipad*yy;
	z=ipad*zz;
	
	ls=int(x)-irad+1;
	if(x<0.0f) ls=ls-1;
	if(ls<(-jc)) ls=-jc+1;
	lt=int(x)+irad;
	if(x<0.0f) lt=lt-1;
	if(lt>=jc) lt=jc-1;
	
	ms=int(y)-irad+1;
	if(y<0.0f) ms=ms-1;
	if(ms<=(-jc)) ms=-jc+1;
	mt=int(y)+irad;
	if(y<0.0f) mt=mt-1;
	if(mt>=jc) mt=jc-1;
	
	ns=int(z)-irad+1;
	if(z<0.0f) ns=ns-1;
	if(ns<=(-jc)) ns=-jc+1;
	nt=int(z)+irad;
	if(z<0.0f) nt=nt-1;
	if(nt>=jc) nt=jc-1;
	
	for(l=ls;l<=lt;l++)
	{
		arg[0]=x-l;
		for(m=ms;m<=mt;m++)
		{
			arg[1]=y-m;
			for(n=ns;n<=nt;n++)
			{
				arg[2]=z-n;
				boxftv=boxft_lut(arg,sinclut);
				//if(l>=0)
				//{
					ll=(l>=0)?(l):(-l);
					mm=(l>=0)?(m):(-m);				
					nn=(l>=0)?(n):(-n);
					if(mm<0) mm=mm+nsam;
					if(nn<0) nn=nn+nsam;
					c3d=(ll!=(jc-1)) ? tex3D(texRef_c3df,ll,mm,nn) : tex2D(texRef_c3ds,mm,nn);
					if(l<0) c3d.y=-c3d.y;
					samp=cPc(samp,cXr(c3d,boxftv));
				//}
				/*else
				{
					ll=-l+1;
					mm=-m+1;
					if(mm<1) mm=mm+nsam;
					nn=-n+1;
					if(nn<1) nn=nn+nsam;
					c3d=(ll!=jc) ? tex3D(texRef_c3df,ll-1,mm-1,nn-1) : tex2D(texRef_c3ds,mm-1,nn-1);
					samp=cPc(samp,conjg(cXr(c3d,boxftv)));
				}*/
				
			}
		}
	}
	
	return samp;
} 


static __device__ CPLX ewaldex(int &nsam, int &irad, CPLX *a3df, CPLX *a3ds, 
             float *sinclut, int &ipad,
             int &i, int &j,float *dm, float &thet, CPLX &ctfr, CPLX &ctfl)
{
	float x3,y3,z3;
	float thetah,gpix,x,y,z;
	CPLX ainpo1,ainpo2;
	
	gpix=sqrt(float(i*i+j*j));  //length of resol.vector g in pixel
	thetah=gpix*thet/2;       //THET=(WL/(PSIZE*NSAM))/AMAGP
	x=i*cos(thetah);          //THETAH=scattering angle/2.
	y=j*cos(thetah);
	z=gpix*sin(thetah);
	x3=dm[0]*x+dm[3]*y+dm[6]*z;
	y3=dm[1]*x+dm[4]*y+dm[7]*z;
	z3=dm[2]*x+dm[5]*y+dm[8]*z;
	
	if(irad==0) //AINPO1 is first beam
	{
		ainpo1=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
	}
	else
	{
		ainpo1=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
	}
	
	x3=dm[0]*x+dm[3]*y-dm[6]*z;  //AINPO2 is second beam = conjg(x,y,-z)
	y3=dm[1]*x+dm[4]*y-dm[7]*z;
	z3=dm[2]*x+dm[5]*y-dm[8]*z;
	if(irad==0)
	{
		ainpo2=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
	}
	else
	{
		ainpo2=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
	}
	return cPc(cXc(ainpo1,ctfr),cXc(ainpo2,conjg(ctfl)));  //combine beams
}

__global__ void pextract_pix(int nsam, int jc, int ic, int ipad, int iewald, int irad,
					CPLX *outc, CPLX *outq,	CPLX * a3df, CPLX *a3ds, float *sinclut,
					float *dm, float thet, CPLX *ctff, CPLX *ctfs)
{
	int i=blockIdx.x*blockDim.x+threadIdx.x;
	int j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;
	if(i>(jc-1) || j>(jc-1)) return;
	
	int id;
	float x3,y3,z3;
	CPLX ctfr,ctfl, ainpo;
   int ii=i+1;
   int jj=j+1;
   if(jj<1) jj=jj+nsam;
   int ir2=i*i+j*j;
   if(ir2<=(nsam*nsam/4))
   {	
		if(ii!=jc)
		{
			id=ii+nsam/2*(jj-1);
			ctfr=ctff[id-1];
			ctfl=ctff[id+ic-1];
		}
		else
		{
			ctfr=ctfs[jj-1];
			ctfl=ctfs[jj+nsam-1];
		}
				
		if(iewald!=0)
		{
			if(iewald<0)
			{
				ctfr=conjg(ctfr);
				ctfl=conjg(ctfl);
			}
			//MW call ewald sphere corrected extraction
			ainpo=ewaldex(nsam,irad,a3df,a3ds,sinclut,ipad,i,j,dm,thet,ctfr,ctfl);
		}
		else
		{
			x3=dm[0]*i+dm[3]*j;
			y3=dm[1]*i+dm[4]*j;
			z3=dm[2]*i+dm[5]*j;
			if(irad==0)
			{
				ainpo=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);
			}
			else
			{
				ainpo=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad);
			}
			ainpo=cXc(ainpo,cPc(ctfr,conjg(ctfl)));
		}
		
		if(ii!=jc)
		{
			outc[id-1]=ainpo;
		}
		else
		{
			outq[jj-1]=ainpo;
		}
	}
	else
	{
		if(ii!=jc)
		{
			id=ii+nsam/2*(jj-1);
			outc[id-1].x=0.0f;
			outc[id-1].y=0.0f;
		}
		else
		{
			outq[jj-1].x=0.0f;
			outq[jj-1].y=0.0f;
		}
	}
		
}

void cuda_pextract(int nsam, int irad, CPLX *outc, CPLX *outq,
					CPLX * a3df, CPLX *a3ds, float phi, float theta, float psi,
					float *sinclut, int ipad, float thet, int iewald, 
					CPLX *ctff, CPLX *ctfs)
{	
	int jc,ic;
	float dm[9],cphi,sphi,cthe,sthe,cpsi,spsi;

	
	jc=nsam/2+1;
	ic=nsam*nsam/2;
	cphi = cos(phi);
   sphi = sin(phi);
   cthe = cos(theta);
   sthe = sin(theta);
   cpsi = cos(psi);
   spsi = sin(psi);

   dm[0] = cphi * cthe * cpsi - sphi * spsi;
   dm[1] = sphi * cthe * cpsi + cphi * spsi;
   dm[2] = -sthe * cpsi;
   dm[3] = -cphi * cthe * spsi - sphi * cpsi;
   dm[4] = -sphi * cthe * spsi + cphi * cpsi;
   dm[5] = sthe * spsi;
	// MW next lines uncommented for Ewald sphere correction
   dm[6] = sthe * cphi;
   dm[7] = sthe * sphi;
   dm[8] = cthe;

	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	dim3 dimBlock(pGM->rc_x,pGM->rc_y);
	dim3 dimGrid(jc/dimBlock.x+1,(2*jc-1)/dimBlock.y+1);
	cudaMemcpy(pGM->d_dm,dm,sizeof(float)*9,cudaMemcpyHostToDevice);
	
		//###################texture######################
	//bind to texture reference
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<CPLX>();
	cudaBindTextureToArray(&texRef_c3df, pGM->cu_c3df, &channelDesc);
	cudaBindTextureToArray(&texRef_c3ds, pGM->cu_c3ds, &channelDesc);
	
	size_t offset=0;
	cudaBindTextureToArray(&texRef_ctff, pGM->cu_ctff, &channelDesc);
	cudaBindTexture(&offset, &texRef_ctfs, (void *)pGM->d_ctfs, &channelDesc,size_t(sizeof(CPLX)*nsam*2));

	cudaChannelFormatDesc channelfloat = cudaCreateChannelDesc<float>();
	cudaBindTexture(&offset, &texRef_sinclut, (void *)pGM->d_sinclut, &channelfloat,size_t(sizeof(float)*2000));	
   //###############################################

   pextract_pix<<<dimGrid,dimBlock>>>(nsam, jc, ic, ipad, iewald, irad,
					outc, outq, a3df, a3ds, sinclut, pGM->d_dm, thet, ctff, ctfs);
	cudaThreadSynchronize();

}









