#include "pinsert_s_gpu.h"


//******************
static __device__ inline void atomicFloatAdd(float *address, float val)
{
      int i_val = __float_as_int(val);
      int tmp0 = 0;
      int tmp1;

      while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0)
      {
              tmp0 = tmp1;
              i_val = __float_as_int(val + __int_as_float(tmp1));
      }
}

static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0f;
	return f;
}

static __device__ float cabs(CPLX f)
{
	return sqrt(f.x*f.x+f.y*f.y);
}

static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cDc(CPLX a, CPLX b) // a/b
{
	CPLX c;
	float ss=b.x*b.x+b.y*b.y;
	c.x=(a.x*b.x+a.y*b.y)/ss;
	c.y=(a.y*b.x-a.x*b.y)/ss;
	return c;
}

static __device__ CPLX cXr(CPLX a, float b)
{
	CPLX c;
	c.x=a.x*b;
	c.y=a.y*b;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}

static __device__ CPLX cMc(CPLX a, CPLX b)  //a-b
{
	CPLX c;
	c.x=a.x-b.x;
	c.y=a.y-b.y;
	return c;
}

static __device__ float boxft_lut(float *arg, float *sinclut)
{
	int i,j;
	float sinc,prod=1.0f;
	
	for(i=0;i<3;i++)
	{
		arg[i]=abs(arg[i]);
		if(arg[i]<0.02f)
		{
			sinc=1.0f;
		}
		else
		{
			j=int(180.0f*arg[i]+0.5f);
			sinc=sinclut[j-1];			
		}
		prod=prod*sinc;
	}
	return prod;

}


static __device__ float pdiff(CPLX c1, CPLX c2)
{
	CPLX c;
	
	c=cXc(c1,conjg(c2));
	if(abs(c.x)<0.00000001f && abs(c.y)<0.00000001f) return 0.0f;
	else return abs(atan2(c.y,c.x));

}

static __device__ CPLX ainterpo3dbig(int nsama, int ipad, CPLX *a3df, CPLX *a3ds, 
                                      float &x, float &y, float &z)
{
	int nsam,jc;
	int l,m,n,ll,mm,nn,is,id,nsamh;
	CPLX samp;
	
	nsam=ipad*nsama;
	samp.x=0.0f;
	samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	l=(x>=0.0f)?int(ipad*x+0.5f):int(ipad*x-0.5f);
	m=(y>=0.0f)?int(ipad*y+0.5f):int(ipad*y-0.5f);
	n=(z>=0.0f)?int(ipad*z+0.5f):int(ipad*z-0.5f);
	if(l>=0)
	{
		ll=l+1;
		mm=m+1;
		if(mm<1) mm=mm+nsam;
		nn=n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,a3df[id-1]);
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,a3ds[is-1]);
		}
	}
	else
	{
		ll=-l+1;
		mm=-m+1;
		if(mm<1) mm=mm+nsam;
		nn=-n+1;
		if(nn<1) nn=nn+nsam;
		if(ll!=jc)
		{
			id=ll+nsamh*((mm-1)+nsam*(nn-1));
			samp=cPc(samp,conjg(a3df[id-1]));
		}
		else
		{
			is=mm+nsam*(nn-1);
			samp=cPc(samp,conjg(a3ds[is-1]));
		}
	}
	return samp;
}           

static __device__ CPLX ainterpo3ds(int nsam, int irad, CPLX *a3df, CPLX *a3ds, 
                        float &xx, float &yy, float &zz, float *sinclut, int &ipad)
{
	int ls,lt,ns,nt,ms,mt,jc;
	int l,m,n,ll,mm,nn,is,id,nsamh;
	float x,y,z,arg[3],boxftv;
	CPLX samp;
	
	samp.x=samp.y=0.0f;
	jc=nsam/2+1;
	nsamh=nsam/2;
	x=ipad*xx;
	y=ipad*yy;
	z=ipad*zz;
	
	ls=int(x)-irad+1;
	if(x<0.0f) ls=ls-1;
	if(ls<(-jc)) ls=-jc+1;
	lt=int(x)+irad;
	if(x<0.0f) lt=lt-1;
	if(lt>=jc) lt=jc-1;
	
	ms=int(y)-irad+1;
	if(y<0.0f) ms=ms-1;
	if(ms<=(-jc)) ms=-jc+1;
	mt=int(y)+irad;
	if(y<0.0f) mt=mt-1;
	if(mt>=jc) mt=jc-1;
	
	ns=int(z)-irad+1;
	if(z<0.0f) ns=ns-1;
	if(ns<=(-jc)) ns=-jc+1;
	nt=int(z)+irad;
	if(z<0.0f) nt=nt-1;
	if(nt>=jc) nt=jc-1;
	
	for(l=ls;l<=lt;l++)
	{
		arg[0]=x-l;
		for(m=ms;m<=mt;m++)
		{
			arg[1]=y-m;
			for(n=ns;n<=nt;n++)
			{
				arg[2]=z-n;
				boxftv=boxft_lut(arg,sinclut);
				if(l>=0)
				{
					ll=l+1;
					mm=m+1;
					if(mm<1) mm=mm+nsam;
					nn=n+1;
					if(nn<1) nn=nn+nsam;
					if(ll!=jc)
					{
						id=ll+nsamh*((mm-1)+nsam*(nn-1));
						samp=cPc(samp,cXr(a3df[id-1],boxftv));
					}
					else
					{
						is=mm+nsam*(nn-1);
						samp=cPc(samp,cXr(a3ds[is-1],boxftv));
					}
				}
				else
				{
					ll=-l+1;
					mm=-m+1;
					if(mm<1) mm=mm+nsam;
					nn=-n+1;
					if(nn<1) nn=nn+nsam;
					if(ll!=jc)
					{
						id=ll+nsamh*((mm-1)+nsam*(nn-1));
						samp=cPc(samp,conjg(cXr(a3df[id-1],boxftv)));
					}
					else
					{
						is=mm+nsam*(nn-1);
						samp=cPc(samp,conjg(cXr(a3ds[is-1],boxftv)));
					}
				}
				
			}
		}
	}
	
	return samp;
}           

static __device__ void ewaldin(int &nsam, int &irad, CPLX *a3df, CPLX *a3ds, 
             float *sinclut, int &ipad,
             int &i, int &j, float *dm, float &thet, float *xyz, CPLX &fo_r, CPLX &fo_l, 
             CPLX &fobs, float &prel, int &iewald, CPLX &ctfr, CPLX &ctfl)
{
	float x3,y3,z3;
   float thetah,gpix,x,y,z,aobs,aref;
   CPLX fref,frr,frl,frel;


// First get reference amplitude and phase for left and right beam
// from previous 3D model
	gpix=sqrt(float(i*i+j*j));  //length of resol.vector g in pixel
	thetah=gpix*thet/2;         //THET=(WL/(PSIZE*NSAM))/AMAGP
	x=i*cos(thetah);            //THETAH=scattering angle/2.
	y=j*cos(thetah);
	z=gpix*sin(thetah);

   xyz[0]=dm[0]*x+dm[3]*y+dm[6]*z;
   xyz[1]=dm[1]*x+dm[4]*y+dm[7]*z;
   xyz[2]=dm[2]*x+dm[5]*y+dm[8]*z;
   xyz[3]=dm[0]*x+dm[3]*y-dm[6]*z;
   xyz[4]=dm[1]*x+dm[4]*y-dm[7]*z;
   xyz[5]=dm[2]*x+dm[5]*y-dm[8]*z;
   
   if(abs(iewald)>1)
   {
   	x3=xyz[0];
   	y3=xyz[1];
   	z3=xyz[2];
   	if(irad==0)	frr=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);  
   	else frr=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad); 
   	
   	x3=xyz[3];
   	y3=xyz[4];
   	z3=xyz[5];
   	if(irad==0)	frl=ainterpo3dbig(nsam,ipad,a3df,a3ds,x3,y3,z3);  
   	else frl=ainterpo3ds(ipad*nsam,irad,a3df,a3ds,x3,y3,z3,sinclut,ipad); 
   	
   	// we only use the relative phase between the two extracted beams AINPO1/2 to
		// reconstitute the unknowns FOR and FOL while maintaining amplitude and phase
		// of the observed sample FOBS.
		
		aobs=cabs(fobs);   //amplitude, observed sf
		fref=cPc(cXc(frr,ctfr),cXc(frl,conjg(ctfl))); //beam combination acc. to  DDR
		aref=cabs(fref);    //amplitude, reference sf
		prel=0.0f;
		
		if(aobs<0.00000001f)
		{
			fo_r.x=fo_r.y=0.0f;
			fo_l.x=fo_l.y=0.0f;
		}
		else
		{
			if(aref>=0.00000001)
			{
				frel=cXr(cDc(fobs,fref),pow(cabs(cPc(ctfr,conjg(ctfl))),2));  //difference vector
				prel=abs(atan2(frel.y,frel.x));  //phase of diff. vector
				fo_r=cXc(frr,frel);  //right derived sf
				fo_l=cXc(frl,frel);  //left derived sf
			}
			else
			{
				fo_r=cXc(fobs,conjg(ctfr));  //without reference just use
				fo_l=cXc(fobs,ctfl);         //the observed values
			}	
		}
   }
   else
   {
   	fo_r=cXc(fobs,conjg(ctfr));  //without reference just use
		fo_l=cXc(fobs,ctfl);         //the observed values
   }
   
}
//******************


__global__ void insert_pix(int nsam, int irad, int beam, 
		int irrec2, CPLX *spec, CPLX *speq, CPLX *ctff, CPLX *ctfs, 
		CPLX *b3df, CPLX *b3ds, float *t3df, float *t3ds, 
		double *pttd, float *asum, float *vsum, 
		float *psum, CPLX *c3df, int *ksum, float *sinclut, 
		int irada, int ipad, float thet, int iewald, CPLX *c3ds, 
		float *sumprel, int *nprel, int nnstat, long *pnksum, 
		int jc, float *dm, float tmp, int nsamh, int ic, 
		int irad2, int i1, int i2)
{
	int i,j,ii,jj,ls,lt,ms,mt,ns,nt;
	int l,m,n,ll,mm,nn,id,is;
	int jrad2=0,id4;
	float x3,y3,z3,boxftv;
	float ctfv2,ctfv2s2,arg[3];
//	float phase;
//	CPLX pshft; //maybe there is something wrong! see line 59	
	float wgt,g2;
	float a;
	CPLX samp,bsamp;
	float xyz[6],prel;
	int gpix;
	CPLX ffor,fol,ctfr,ctfl;

	double ttd=0.0;
	long nksum=0;

	i=blockIdx.x*blockDim.x+threadIdx.x+i1;
	j=blockIdx.y*blockDim.y+threadIdx.y-jc+1;
	
	
	if(i<=i2 && j<=(jc-1)) 
	{		
	
			if(iewald==0)     //planar insertion
			{
				x3=dm[0]*i+dm[3]*j;
				y3=dm[1]*i+dm[4]*j;
				z3=dm[2]*i+dm[5]*j;
			}

			g2=i*i+j*j;
			gpix=int(sqrt(g2)); //int(length resol.vector g in pixel)
			wgt=exp(tmp*g2);
			//phase=shx*i+shy*j;
			//pshft.x=cos(phase);   //something wrong in line 59 of ori ffloatign
			//pshft.y=sin(phase);

			// use 'is' as a sign to save the register, when i>=0 is==1, when i<0 is==-1
			is=((i>=0)*2-1);
			ii=is*i+1;
			jj=is*j+1;
			if(jj<1) jj=jj+nsam;
			
			id=ii+nsamh*(jj-1);
			ctfr=((ii!=jc)?cXr(ctff[id-1],wgt):cXr(ctfs[jj-1],wgt));
			ctfl=((ii!=jc)?cXr(ctff[id+ic-1],wgt):cXr(ctfs[jj+nsam-1],wgt));

	
			if(iewald<=0)
			{
				ctfr=conjg(ctfr);
				ctfl=conjg(ctfl);
			}
			//samp=((ii!=jc)?cXr(cXc(spec[id-1],pshft),wgt*wgt):cXr(cXc(speq[jj-1],pshft),wgt*wgt));
			samp=((ii!=jc)?cXr(spec[id-1],wgt*wgt):cXr(speq[jj-1],wgt*wgt));
			samp.y*=((ii!=jc)?is:1); // if is==-1, then equal to conjg(samp)

			if(iewald!=0)
			{
				ewaldin(nsam, irada, c3df, c3ds, sinclut, ipad, i, j, dm, 
							thet, xyz, ffor, fol, samp, prel, iewald, ctfr, ctfl);
				if(beam==1)
				{			
					x3=xyz[0];
					y3=xyz[1];
					z3=xyz[2];
					samp=ffor;  //right beam
					//sumprel[gpix-1]=sumprel[gpix-1]+prel;  //accum.phase of diff.vector
					//nprel[gpix-1]=nprel[gpix-1]+1;  //in resolution zones, N
					sumprel[gpix]+=prel;  //accum.phase of diff.vector
					nprel[gpix]++;  //in resolution zones, N
				}
				else
				{

					x3=xyz[3];
					y3=xyz[4];
					z3=xyz[5];
					samp=fol;
				}
			}
			else
			{
				samp=cXc(samp,cPc(ctfr,conjg(ctfl)));
			}


	
			//ctfv2=0.25f*pow(cabs(cPc(ctfr,conjg(ctfl))),2);
			ctfv2=0.25f*pow(cabs(cPc(ctfr,conjg(ctfl))),2);
					
			a=x3+0.5-float(irad-1)/2.0;
			ls=int(a);
			if(a<0.0) ls=ls-1;
			if(ls<0) ls=0;
			a=x3+float(irad)/2.0;
			lt=int(a);
			if(a<0.0) lt=lt-1;
			if(lt>=jc) lt=jc-1;
			
			a=y3+0.5-float(irad-1)/2.0;
			ms=int(a);
			if(a<0.0) ms=ms-1;
			if(ms<=(-jc)) ms=-jc+1;
			a=y3+float(irad)/2.0;
			mt=int(a);
			if(a<0.0) mt=mt-1;
			if(mt>=jc) mt=jc-1;
			
			a=z3+0.5-float(irad-1)/2.0;
			ns=int(a);
			if(a<0.0) ns=ns-1;
			if(ns<=(-jc)) ns=-jc+1;
			a=z3+float(irad)/2.0;
			nt=int(a);
			if(a<0.0) nt=nt-1;
			if(nt>=jc) nt=jc-1;
			
			//arg(1),arg(2),arg(3) not multiplied by pi after 4.05 
			for(l=ls;l<=lt;l++)		
				for(m=ms;m<=mt;m++)
					for(n=ns;n<=nt;n++)
					{
						arg[0]=x3-l;
						arg[1]=y3-m;
						arg[2]=z3-n;
				
						boxftv=boxft_lut(arg,sinclut); 

						jrad2=l*l+m*m+n*n;
						if(jrad2<=irrec2)
						{
							ll=l+1;
							mm=m+1;
							if(mm<1) mm=mm+nsam;
							nn=n+1;
							if(nn<1) nn=nn+nsam;
							
							ctfv2s2=ctfv2*boxftv;
							bsamp=cXr(samp,boxftv);
	
								if(ll!=jc)
								{
									id=ll+nsamh*((mm-1)+nsam*(nn-1));
									//b3df[id-1]=cPc(b3df[id-1],bsamp);
									//t3df[id-1]=t3df[id-1]+ctfv2s2;
									//********test atomic*********
									atomicFloatAdd(&(b3df[id-1].x),bsamp.x);
									atomicFloatAdd(&(b3df[id-1].y),bsamp.y);
									atomicFloatAdd(&(t3df[id-1]),ctfv2s2);
									//****************************
									nksum++;

									if(nnstat!=0)
									{
										id4=ipad*(ll-1)+1+ipad*nsamh*((mm-1)+ipad*nsam*(nn-1));
										//ksum[id-1]=ksum[id-1]+1;
										//vsum[id-1]=vsum[id-1]+pow(cabs(bsamp),2);
										//psum[id-1]=psum[id-1]+pdiff(bsamp,c3df[id4-1]);
										//asum[id-1]=asum[id-1]+cabs(bsamp);
										atomicAdd(&(ksum[id-1]),1);
										atomicFloatAdd(&(vsum[id-1]),pow(cabs(bsamp),2));
										atomicFloatAdd(&(psum[id-1]),pdiff(bsamp,c3df[id4-1]));
										atomicFloatAdd(&(asum[id-1]),cabs(bsamp));										
									}
								}
								else
								{
									is=mm+nsam*(nn-1);
									//b3ds[is-1]=cPc(b3ds[is-1],bsamp);
									//t3ds[is-1]=t3ds[is-1]+ctfv2s2;
									//********test atomic*********
									atomicFloatAdd(&(b3ds[is-1].x),bsamp.x);
									atomicFloatAdd(&(b3ds[is-1].y),bsamp.y);
									atomicFloatAdd(&(t3ds[is-1]),ctfv2s2);
									//****************************
								}
								
								ttd+=double(ctfv2s2);							
						}
					}	

		
			i=blockIdx.x*blockDim.x+threadIdx.x;
			j=blockIdx.y*blockDim.y+threadIdx.y;
			pttd[i+j*(nsam+1)]+=ttd;
			pnksum[i+j*(nsam+1)]+=nksum;		

	}
}

void pinsert_s(int nsam, int irad, float rii, 
		int irrec2, CPLX *spec, CPLX *speq, CPLX *ctff, CPLX *ctfs, 
		CPLX *b3df, CPLX *b3ds, float *t3df, float *t3ds, 
		double *ttd, float shx, float shy, float *asum, float *vsum, 
		float *psum, CPLX *c3df, int *ksum, int isign, float *sinclut, 
		int irada, int ipad, float thet, int iewald, CPLX *c3ds, 
		float *sumprel, int *nprel, int nnstat, long *nksum, 
		int jc, float *dm, float tmp, int nsamh, int ic, 
		int irad2, float st, int i, cuMem *pGM)
{
	//printf("%d %d\n",ipad, irad);
	int i1,i2;
	int beam;

	/*id=(i-1)*nsam/2+1;
	if(i!=1)
	{
		for(j=0;j<nsam/2;j++)
		{
			sumprel[id+j-1]=0.0;
			nprel[id+j-1]=0;
		}
	}*/
	i1=-jc+1+int((i-1)*st);
	i2=-jc+int(i*st);

	//####################################################
	//copy data to GPU
//	cudaMemcpy((void *)pGM->d_spec,(void *)spec,sizeof(float)*nsam*nsam,cudaMemcpyHostToDevice);
//	cudaMemcpy((void *)pGM->d_speq,(void *)speq,sizeof(float)*nsam*2,cudaMemcpyHostToDevice);
//	cudaMemcpy((void *)pGM->d_ctff,(void *)ctff,sizeof(float)*nsam*nsam*2,cudaMemcpyHostToDevice);
//	cudaMemcpy((void *)pGM->d_ctfs,(void *)ctfs,sizeof(float)*nsam*4,cudaMemcpyHostToDevice);
	cudaMemcpy((void *)pGM->d_dm,(void *)dm,sizeof(float)*9,cudaMemcpyHostToDevice);
	//####################################################

	dim3 dimBlock(pGM->rc_x,pGM->rc_y);
	dim3 dimGrid((i2-i1)/dimBlock.x+1,(2*jc-2)/dimBlock.y+1);

	beam=1; //left beam
	insert_pix<<<dimGrid,dimBlock>>>(nsam,irad,beam, irrec2, 
		(CPLX *)pGM->d_datc, (CPLX *)pGM->d_datq, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, 
		b3df, b3ds, t3df, t3ds,ttd, asum, vsum, 
		psum, c3df, ksum, pGM->d_sinclut, irada, ipad, thet, iewald, c3ds, 
		sumprel, nprel, nnstat, nksum, jc, pGM->d_dm, tmp, nsamh, ic, 
		irad2, i1, i2);

	beam=2; //right beam
	if(iewald!=0)   //skip 2nd beam if planar insertion
	{	
		cudaThreadSynchronize();
		insert_pix<<<dimGrid,dimBlock>>>(nsam,irad,beam, irrec2, 
		(CPLX *)pGM->d_datc, (CPLX *)pGM->d_datq, (CPLX *)pGM->d_ctff, (CPLX *)pGM->d_ctfs, 
		b3df, b3ds, t3df, t3ds,ttd,asum, vsum, 
		psum, c3df, ksum, pGM->d_sinclut, irada, ipad, thet, iewald, c3ds, 
		sumprel, nprel, nnstat, nksum, jc, pGM->d_dm, tmp, nsamh, ic, 
		irad2, i1, i2);
	}	
	cudaThreadSynchronize();
}





























/*	for(i=i1;i<=i2;i++)
		for(j=(-jc+1);j<=(jc-1);j++)
			for(beam=1;beam<=2;beam++)  //insert both left and right beam if IEWALD
			{
				if(iewald==0 && beam==2) continue;  //skip 2nd beam if planar insertion
				if(iewald==0)     //planar insertion
				{
					x3=dm[0]*i+dm[3]*j;
					y3=dm[1]*i+dm[4]*j;
					z3=dm[2]*i+dm[5]*j;
				}

				g2=i*i+j*j;
				gpix=int(sqrt(g2)); //int(length resol.vector g in pixel)
				wgt=exp(tmp*g2);
				phase=shx*i+shy*j;
				pshft.x=cos(phase);   //something wrong in line 59 of ori ffloatign
				pshft.y=sin(phase);

				if(i>=0)
				{
					ii=i+1;
					jj=j+1;
					if(jj<1) jj=jj+nsam;
					if(ii!=jc)
					{
						id=ii+nsamh*(jj-1);
						ctfr=cXr(ctff[id-1],wgt);
						ctfl=cXr(ctff[id+ic-1],wgt);
						if(iewald<=0)
						{
							ctfr=conjg(ctfr);
							ctfl=conjg(ctfl);
						}
						samp=cXr(cXc(spec[id-1],pshft),wgt*wgt);
						if(iewald!=0)
						{
							if(beam==1)
							{
								ewaldin(nsam, irada, c3df, c3ds, sinclut, ipad, i, j, dm, 
											thet, xyz, ffor, fol, samp, prel, iewald, ctfr, ctfl);
								x3=xyz[0];
								y3=xyz[1];
								z3=xyz[2];
								samp=ffor;  //right beam
								sumprel[gpix-1]=sumprel[gpix-1]+prel;  //accum.phase of diff.vector
								nprel[gpix-1]=nprel[gpix-1]+1;  //in resolution zones, N
							}
							else
							{
								x3=xyz[3];
								y3=xyz[4];
								z3=xyz[5];
								samp=fol;
							}
						}
						else
						{
							samp=cXc(samp,cPc(ctfr,conjg(ctfl)));
						}
					}
					else
					{
						ctfr=cXr(ctfs[jj-1],wgt);
						ctfl=cXr(ctfs[jj+nsam-1],wgt);
						if(iewald<0)
						{
							ctfr=conjg(ctfr);
							ctfl=conjg(ctfl);
						}
						samp=cXr(cXc(speq[jj-1],pshft),wgt*wgt);
						if(iewald!=0)
						{
							if(beam==1)
							{
								ewaldin(nsam, irada, c3df, c3ds, sinclut, ipad, i, j, dm, 
											thet, xyz, ffor, fol, samp, prel, iewald, ctfr, ctfl);
								x3=xyz[0];
								y3=xyz[1];
								z3=xyz[2];
								samp=ffor;
								sumprel[gpix-1]=sumprel[gpix-1]+prel;
								nprel[gpix-1]=nprel[gpix-1]+1;
							}
							else
							{
								x3=xyz[3];
								y3=xyz[4];
								z3=xyz[5];
								samp=fol;
							}
						}
						else
						{
							samp=cXc(samp,cPc(ctfr,conjg(ctfl)));
						}
					}
				}
				else
				{
					ii=-i+1;
					jj=-j+1;
					if(jj<1) jj=jj+nsam;
					if(ii!=jc)
					{
						id=ii+nsamh*(jj-1);
						ctfr=cXr(ctff[id-1],wgt);
						ctfl=cXr(ctff[id+ic-1],wgt);
						if(iewald<0)
						{
							ctfr=conjg(ctfr);
							ctfl=conjg(ctfl);
						}
						samp=conjg(cXr(cXc(spec[id-1],pshft),wgt*wgt));
						if(iewald!=0)
						{
							if(beam==1)
							{
								ewaldin(nsam, irada, c3df, c3ds, sinclut, ipad, i, j, dm, 
											thet, xyz, ffor, fol, samp, prel, iewald, ctfr, ctfl);
								x3=xyz[0];
								y3=xyz[1];
								z3=xyz[2];
								samp=ffor;
								sumprel[gpix-1]=sumprel[gpix-1]+prel;
								nprel[gpix-1]=nprel[gpix-1]+1;
							}
							else
							{
								x3=xyz[3];
								y3=xyz[4];
								z3=xyz[5];
								samp=fol;
							}
						}
						else
						{
							samp=cXc(samp,cPc(ctfr,conjg(ctfl)));					
						}
					}
					else
					{
						ctfr=cXr(ctfs[jj-1],wgt);
						ctfl=cXr(ctfs[jj+nsam-1],wgt);
						if(iewald<0)
						{
							ctfr=conjg(ctfr);
							ctfl=conjg(ctfl);
						}
						samp=cXr(cXc(speq[jj-1],pshft),wgt*wgt);
						if(iewald!=0)
						{
							if(beam==1)
							{
								ewaldin(nsam, irada, c3df, c3ds, sinclut, ipad, i, j, dm, 
											thet, xyz, ffor, fol, samp, prel, iewald, ctfr, ctfl);
								x3=xyz[0];
								y3=xyz[1];
								z3=xyz[2];
								samp=ffor;
								sumprel[gpix-1]=sumprel[gpix-1]+prel;
								nprel[gpix-1]=nprel[gpix-1]+1;
							}
							else
							{
								x3=xyz[3];
								y3=xyz[4];
								z3=xyz[5];
								samp=fol;
							}
						}
						else
						{
							samp=cXc(samp,cPc(ctfr,conjg(ctfl)));
						}
					}
				}
				ctfv2=0.25*pow(cabs(cPc(ctfr,conjg(ctfl))),2);
							
				a=x3+0.5-float(irad-1)/2.0;
				ls=int(a);
				if(a<0.0) ls=ls-1;
				if(ls<0) ls=0;
				a=x3+float(irad)/2.0;
				lt=int(a);
				if(a<0.0) lt=lt-1;
				if(lt>=jc) lt=jc-1;
			
				a=y3+0.5-float(irad-1)/2.0;
				ms=int(a);
				if(a<0.0) ms=ms-1;
				if(ms<=(-jc)) ms=-jc+1;
				a=y3+float(irad)/2.0;
				mt=int(a);
				if(a<0.0) mt=mt-1;
				if(mt>=jc) mt=jc-1;
			
				a=z3+0.5-float(irad-1)/2.0;
				ns=int(a);
				if(a<0.0) ns=ns-1;
				if(ns<=(-jc)) ns=-jc+1;
				a=z3+float(irad)/2.0;
				nt=int(a);
				if(a<0.0) nt=nt-1;
				if(nt>=jc) nt=jc-1;
			
				//arg(1),arg(2),arg(3) not multiplied by pi after 4.05 
				for(l=ls;l<=lt;l++)
				{
					l2=l*l;
					arg[0]=x3-l;
					for(m=ms;m<=mt;m++)
					{
						m2=m*m;
						arg[1]=y3-m;
						for(n=ns;n<=nt;n++)
						{
							n2=n*n;
							arg[2]=z3-n;
						
							boxftv=boxft_lut(arg,sinclut);   
							jrad2=l2+m2+n2;
							if(jrad2<=irrec2)
							{
								ll=l+1;
								mm=m+1;
								if(mm<1) mm=mm+nsam;
								nn=n+1;
								if(nn<1) nn=nn+nsam;
							
								ctfv2s2=ctfv2*boxftv;
								bsamp=cXr(samp,boxftv);

								if(isign>0)
								{
									if(ll!=jc)
									{
										id=ll+nsamh*((mm-1)+nsam*(nn-1));
										id4=ipad*(ll-1)+1+ipad*nsamh*((mm-1)+ipad*nsam*(nn-1));
										b3df[id-1]=cPc(b3df[id-1],bsamp);
										t3df[id-1]=t3df[id-1]+ctfv2s2;
										nksum=nksum+1;
										
										if(nnstat!=0)
										{
											ksum[id-1]=ksum[id-1]+1;
											vsum[id-1]=vsum[id-1]+pow(cabs(bsamp),2);
											psum[id-1]=psum[id-1]+pdiff(bsamp,c3df[id4-1]);
											asum[id-1]=asum[id-1]+cabs(bsamp);
											
										}
									}
									else
									{
										is=mm+nsam*(nn-1);
										b3ds[is-1]=cPc(b3ds[is-1],bsamp);
										t3ds[is-1]=t3ds[is-1]+ctfv2s2;
									}
									ttd=ttd+ctfv2s2;
								}
								else
								{
									if(ll!=jc)
									{
										id=ll+nsamh*((mm-1)+nsam*(nn-1));
										b3df[id-1]=cMc(b3df[id-1],bsamp);
									}
									else
									{
										is=mm+nsam*(nn-1);
										b3ds[is-1]=cMc(b3ds[is-1],bsamp);
									}
								}
							}
						}
					}
					
				}
				
			}
	*/	

