#include "rlft3.h"
#include "cuda_runtime.h"
#include "cuMem.h"
extern vector<planFFT> planlist;
extern pthread_key_t cuMemKey;

static __device__ CPLX conjg(CPLX f)
{
	f.y*=-1.0;
	return f;
}


static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

static __device__ CPLX cPc(CPLX a, CPLX b) //a+b
{
	CPLX c;
	c.x=a.x+b.x;
	c.y=a.y+b.y;
	return c;
}

static __device__ CPLX cMc(CPLX a, CPLX b)  //a-b
{
	CPLX c;
	c.x=a.x-b.x;
	c.y=a.y-b.y;
	return c;
}

static __device__ void swap(CPLX *a, CPLX *b)
{
	CPLX w;
	w=*a;
	*a=*b;
	*b=w;
}
//###########################################
/*
	for(i1=0;i1<(nnh/2+1);++i1)
		for(i3=0;i3<nn3;++i3)
			for(i2=0;i2<nn2;++i2)
			{
				j1=0;
				if(i1!=0) j1=nnh-i1;
				id=i1+nnh*(i2+nn2*(i3));
				is=j1+nnh*(i2+nn2*(i3));
				w=data[id];
				data[id]=data[is];
				data[is]=w;
			}
*/
__global__ void fa_1(CPLX *data, int nnh, int nn2, int nn3, int size)
{
	//id=i2+nn2*(i3+nn3*i1)
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int is,j1=0;
	int i2=id%nn2;
	int i3=id/nn2%nn3;
	int i1=id/nn2/nn3;
	//if(i1>=(nnh/2+1) || i3>=nn3 || i2>=nn2) return;
	
	if(i1!=0) j1=nnh-i1;
	id=i1+nnh*(i2+nn2*(i3));
	is=j1+nnh*(i2+nn2*(i3));
	swap(&data[id],&data[is]);
}


/*
	for(i2=0;i2<(nn2/2+1);++i2)
		for(i3=0;i3<nn3;++i3)
		{
			j2=0;
			if(i2!=0) j2=nn2-i2;
			id=i2+nn2*(i3);
			is=j2+nn2*(i3);
			w=speq[id];
			speq[id]=speq[is];
			speq[is]=w;
		}
*/
__global__ void fa_2(CPLX *speq, int nn2, int nn3, int size)
{
	//id=i3+nn3*i2
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int is,j2=0;
	int i3=id%nn3;
	int i2=id/nn3;
	//if(i3>=nn3 || i2>=(nn2/2+1)) return;
	
	if(i2!=0) j2=nn2-i2;
	id=i2+nn2*(i3);
	is=j2+nn2*(i3);
	swap(&speq[id],&speq[is]);
}


/*
	for(i2=0;i2<(nn2/2+1);++i2)
		for(i3=0;i3<nn3;++i3)
			for(i1=0;i1<nnh;++i1)
			{
				j2=0;
				if (i2!=0) j2=nn2-i2;
				id=i1+nnh*(i2+nn2*(i3));
				is=i1+nnh*(j2+nn2*(i3));
				w=data[id];
				data[id]=data[is];
				data[is]=w;
			}
*/
__global__ void fa_3(CPLX *data, int nnh, int nn2, int nn3, int size)
{
	//id=i1+nnh*(i3+nn3*i2)
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int is,j2=0;
	int i1=id%nnh;
	int i3=id/nnh%nn3;
	int i2=id/nnh/nn3;
	//if(i2>=(nn2/2+1) || i3>=nn3 || i1>=nnh) return;
	
	if(i2!=0) j2=nn2-i2;
	id=i1+nnh*(i2+nn2*i3);
	is=i1+nnh*(j2+nn2*i3);
	swap(&data[id],&data[is]);
}


/*
	for(i3=0;i3<(nn3/2+1);++i3)   
		for(i2=0;i2<nn2;++i2)
		{
			j3=0;  
			if(i3!=0) j3=nn3-i3; 
			id=i2+nn2*(i3);
			is=i2+nn2*(j3);
			w=speq[id];
			speq[id]=speq[is];
			speq[is]=w;
		}
*/
__global__ void fa_4(CPLX *speq, int nn2, int nn3, int size)
{
	//id=i2+nn2*i3
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int is,j3=0;
	int i2=id%nn2;
	int i3=id/nn2;
	//if(i2>=nn2 || i3>=(nn3/2+1)) return;
	
	if(i3!=0) j3=nn3-i3;
	id=i2+nn2*(i3);
	is=i2+nn2*(j3);
	swap(&speq[id],&speq[is]);
}


/*
	for(i3=0;i3<(nn3/2+1);++i3)   
		for(i2=0;i2<nn2;++i2)
			for(i1=0;i1<nnh;++i1)
			{  
				j3=0;  
				if(i3!=0) j3=nn3-i3;   
				id=i1+nnh*(i2+nn2*(i3));
				is=i1+nnh*(i2+nn2*(j3));     
				w=data[id];
				data[id]=data[is];
				data[is]=w;
			}
*/
__global__ void fa_5(CPLX *data, int nnh, int nn2, int nn3, int size)
{
	//id=i1+nnh*(i2+nn2*i3)
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int is,j3=0;
	int i1=id%nnh;
	int i2=id/nnh%nn2;
	int i3=id/nnh/nn2;
	//if(i1>=nnh || i2>=nn2 || i3>=(nn3/2+1)) return;
	
	if(i3!=0) j3=nn3-i3;
	id=i1+nnh*(i2+nn2*(i3));
	is=i1+nnh*(i2+nn2*(j3));
	swap(&data[id],&data[is]);
}

void flip_array(CPLX *data, CPLX *speq, int nn1, int nn2, int nn3)
{
	int size;	
	int nnh=nn1/2;
	
	size=(nnh/2+1)*nn3*nn2;
	fa_1<<<size/256+1,256>>>(data, nnh, nn2, nn3, size);
	cudaThreadSynchronize();
			
	size=(nn2/2+1)*nn3;
	fa_2<<<size/64+1,64>>>(speq, nn2, nn3, size);
	cudaThreadSynchronize();
		
	size=(nn2/2+1)*nn3*nnh;
	fa_3<<<size/256+1,256>>>(data, nnh, nn2, nn3, size);
	cudaThreadSynchronize();
		
	size=(nn3/2+1)*nn2;
	fa_4<<<size/64+1,64>>>(speq, nn2, nn3, size);
	cudaThreadSynchronize();
		
	size=(nn3/2+1)*nn2*nnh;
	fa_5<<<size/256+1,256>>>(data, nnh, nn2, nn3, size);
	cudaThreadSynchronize();
}
//###########################################

//###########################################

void do_fft(int n1, int n2, int n3, CPLX *data, int isign)
{
	cufftHandle mplan;
	pthread_t tid=pthread_self();
	
	int i;
	//search for plan in planlist
	for(i=0;i<planlist.size();i++)
	{
		if(planlist[i].N1==n1 && planlist[i].N2==n2 && 
			planlist[i].N3==n3 && planlist[i].tid==tid)
			{
				mplan=planlist[i].plan;
				break;
			}
	}
	//if no plan available, create a new plan and add it to planlist
	if(i==planlist.size() || planlist.size()==0)
	{
		cufftPlan3d(&mplan,n3,n2,n1,CUFFT_C2C);			
		planFFT t_plan;
		t_plan.N1=n1;
		t_plan.N2=n2;
		t_plan.N3=n3;
		t_plan.tid=tid;
		t_plan.plan=mplan;
		planlist.push_back(t_plan);
	}
	
	if(isign==1)
	cufftExecC2C(mplan,(cufftComplex *)data,(cufftComplex *)data,CUFFT_FORWARD);
	
	if(isign==-1)
	cufftExecC2C(mplan,(cufftComplex *)data,(cufftComplex *)data,CUFFT_INVERSE);


}
//#######################################################


//#######################################################
/*
	for(i3=0;i3<nn3;++i3)
			for(i2=0;i2<nn2;++i2)
			{
				id=nnh*(i2+nn2*(i3));
				is=i2+nn2*(i3);
				speq[is]=data[id];
			}
*/
__global__ void rlft_1(CPLX* data, CPLX *speq, int nnh, int nn2, int nn3, int size)
{
	//id=i2+nn2*i3
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int is;
	int i2=id%nn2;
	int i3=id/nn2;
	if(i2>=nn2 || i3>=nn3) return;
	
	id=nnh*(i2+nn2*(i3));
	is=i2+nn2*(i3);
	speq[is]=data[id];
}

/*
	for(i3=0;i3<nn3;++i3)     	
		for(i1=0;i1<(nnq+1);++i1)
			for(i2=0;i2<nn2;++i2)
			{
				j3=0;
				if(i3!=0) j3=nn3-i3;
				j1=nnh-i1;
				w=wt[i1];
				j2=0;
				if (i2!=0) j2=nn2-i2;
				if(i1==0)
				{
					id=nnh*(j2+nn2*(j3));
					is=i2+nn2*(i3);
					h1=cXc(c1,cPc(data[id],conjg(speq[is])));
					h2=cXc(c2,cMc(data[id],conjg(speq[is])));
					data[id]=cPc(h1,h2);
					speq[is]=conjg(cMc(h1,h2));
            }
				else
            {
					id=j1+nnh*(j2+nn2*(j3));
					is=i1+nnh*(i2+nn2*(i3));
              h1=cXc(c1,cPc(data[id],conjg(data[is])));
              h2=cXc(c2,cMc(data[id],conjg(data[is])));
              data[id]=cPc(h1,cXc(w,h2));
              data[is]=conjg(cMc(h1,cXc(w,h2)));
            }         
			}

*/
__global__ void rlft_2(CPLX* data, CPLX *speq, CPLX *wt, CPLX c1, CPLX c2, 
						int nnh,int nn2, int nn3, int size)
{
	//id=i2+nn2*(i1+(nnh/2+1)*i3)
	int id=blockIdx.x*blockDim.x+threadIdx.x;
	if(id>=size) return;
	int j1,j2,j3,is;
	int i2=id%nn2;
	int i1=id/nn2%(nnh/2+1);
	int i3=id/nn2/(nnh/2+1);
	if(i3>=nn3 || i1>=(nnh/1+1) || i2>=nn2) return;
	if(i1==nnh/2 && i2>nn2/2) return;
	CPLX h1,h2,w;
	
	j3=0;
	if(i3!=0) j3=nn3-i3;
	j1=nnh-i1;
	w=wt[i1];
	j2=0;
	if (i2!=0) j2=nn2-i2;
	if(i1==0)
	{
		id=nnh*(j2+nn2*(j3));
		is=i2+nn2*(i3);
		h1=cXc(c1,cPc(data[id],conjg(speq[is])));
		h2=cXc(c2,cMc(data[id],conjg(speq[is])));
		data[id]=cPc(h1,h2);
		speq[is]=conjg(cMc(h1,h2));
	}
	else
	{
		id=j1+nnh*(j2+nn2*(j3));
		is=i1+nnh*(i2+nn2*(i3));
      h1=cXc(c1,cPc(data[id],conjg(data[is])));
      h2=cXc(c2,cMc(data[id],conjg(data[is])));
      data[id]=cPc(h1,cXc(w,h2));
      data[is]=conjg(cMc(h1,cXc(w,h2)));
      if(i1==nnh/2 && i2>0 && i2<nn2/2)
     	{
		  	h1=cXc(c1,cPc(data[is],conjg(data[id])));
		  	h2=cXc(c2,cMc(data[is],conjg(data[id])));
		 	data[is]=cPc(h1,cXc(w,h2));
		 	data[id]=conjg(cMc(h1,cXc(w,h2)));
     	}
    }   
}

//#######################################################

void rlft3_GPU(CPLX *data, CPLX *speq, int nn1, int nn2, int nn3, int isign)
{
	int nnh,nnq;
	int size;
	//get Globla Variable GM
	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
	CPLX c1,c2;
	CPLX *d_wt;
	
	c1.x=0.5;
	c1.y=0.0;
	c2.x=0.0;
	c2.y=-0.5*isign;

	
	nnh=nn1/2;
	nnq=nn1/4;
	
	if(isign==1)
	{
		do_fft(nnh,nn2,nn3,data,isign);
		size=nn2*nn3;
		rlft_1<<<size/512+1,512>>>(data, speq, nnh, nn2, nn3, size);
		cudaThreadSynchronize();
	}
	
	if(isign==-1) flip_array(data,speq,nn1,nn2,nn3);
	
	/*CPLX w;
	w.x=0;
	w.y=0;
	double theta,wi,wpi,wpr,wr,wtemp;
	theta=6.28318530717959/double(isign*nn1);
	wpr=-2.0*pow(sin(0.5*theta),2);
	wpi=sin(theta);
	CPLX wt[512];	
	wr=1.0;
	wi=0.0;
	for(int i=0;i<=nnq;i++)
	{
		wt[i]=w;
		wtemp=wr;
		wr=wr*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
		w.x=wr;
		w.y=wi;
	}
	cudaMalloc((void **)&d_wt,sizeof(CPLX)*(nnq+1));
	cudaMemcpy(d_wt,wt,sizeof(CPLX)*(nnq+1),cudaMemcpyHostToDevice);*/

	if(nn1==pGM->m_nsam)
	{
		if(isign==1) d_wt=(CPLX *)pGM->d_wtf;
		else d_wt=(CPLX *)pGM->d_wtb;
	}
	else if(nn1==(pGM->m_nsam*4))
	{
		if(isign==1) d_wt=(CPLX *)pGM->d_wtf4;
		else d_wt=(CPLX *)pGM->d_wtb4;
	}
	else if(nn1==pGM->m_nsamr)
	{
		if(isign==1) d_wt=(CPLX *)pGM->d_wtfh;
		else d_wt=(CPLX *)pGM->d_wtbh;
	}
	else
	{
		printf("Wrong d_wt size in rlft3");
		exit(0);
	}
	
	size=nn3*(nnq+1)*nn2;
	rlft_2<<<size/512+1,512>>>(data, speq, d_wt, c1, c2, nnh,nn2, nn3, size);
	cudaThreadSynchronize();
	
	if(isign==1) flip_array(data,speq,nn1,nn2,nn3);          
	
	if(isign==-1) do_fft(nnh,nn2,nn3,data,isign);
	
	//cudaFree(d_wt);
}

extern "C" void rlft3pl_(CPLX *data, CPLX *speq, int *nn1, int *nn2, int *nn3, int *isign)
{
	float *d_data=NULL;
	float *d_speq=NULL;
	if(cudaHostGetDevicePointer((void **)&d_data,(void *)data,0)!=cudaSuccess)
	{
		printf("error: Failed to do FFT to page-locked data.\n");
		exit(0);
	}
	if(cudaHostGetDevicePointer((void **)&d_speq,(void *)speq,0)!=cudaSuccess)
	{
		printf("error: Failed to do FFT to page-locked speq.\n");
		exit(0);
	}
	rlft3_GPU((CPLX *)d_data, (CPLX *)d_speq, *nn1, *nn2, *nn3, *isign);
}

extern "C" void rlft3_(CPLX *data, CPLX *speq, int *nn1, int *nn2, int *nn3, int *isign)
{
	float *d_data;
	float *d_speq;
		
	cudaMalloc((void **)&d_data,sizeof(CPLX)*(*nn1)*(*nn2)*(*nn3)/2);
	cudaMalloc((void **)&d_speq,sizeof(CPLX)*(*nn2)*(*nn3));
	cudaMemcpy(d_data,data,sizeof(CPLX)*(*nn1)*(*nn2)*(*nn3)/2,cudaMemcpyHostToDevice);
	cudaMemcpy(d_speq,speq,sizeof(CPLX)*(*nn2)*(*nn3),cudaMemcpyHostToDevice);
	
	rlft3_GPU((CPLX *)d_data, (CPLX *)d_speq, *nn1, *nn2, *nn3, *isign);

	cudaMemcpy(data,d_data,sizeof(CPLX)*(*nn1)*(*nn2)*(*nn3)/2,cudaMemcpyDeviceToHost);
	cudaMemcpy(speq,d_speq,sizeof(CPLX)*(*nn2)*(*nn3),cudaMemcpyDeviceToHost);
	cudaFree(d_data);
	cudaFree(d_speq);

}























