#include "sigma2_gpu.h"
#include "pextract_gpu.h"
extern pthread_key_t cuMemKey;

static __device__ CPLX cXc(CPLX a, CPLX b) // a*b
{
	CPLX c;
	c.x=a.x*b.x-a.y*b.y;
	c.y=a.x*b.y+a.y*b.x;
	return c;
}

__global__ void sigma2_1(int nsam, int jc, float shx, float shy, int ib2, int ib3,
				float *rbuf, CPLX *datc, CPLX *datq)
{
	int l=blockIdx.x*blockDim.x+threadIdx.x+1;
	int m=blockIdx.y*blockDim.y+threadIdx.y+1;
	if(l>jc || m>nsam) return;
	
	int id,id2,m2;
	float phase;
	CPLX pshft,z;
	int ll=l-1;
	int mm=m-1;
	if(mm>=jc) mm=mm-nsam;
	phase=-shx*ll-shy*mm;
	pshft.x=cos(phase);
	pshft.y=sin(phase);
	if(l!=jc)
	{
		id=l+nsam/2*(m-1);
		z=cXc(pshft,datc[id-1]);
		id2=2*id+ib2;
		rbuf[id2-2]=z.x;
		rbuf[id2-1]=z.y;
	}
	else
	{
		z=cXc(pshft,datq[m-1]);
		m2=2*m+ib3;
		rbuf[m2-2]=z.x;
		rbuf[m2-1]=z.y;
	}	
	__syncthreads();
	if(l==1 && m==1) rbuf[ib2]=0.0f;
	if(l==2 && m==1) rbuf[ib2+1]=0.0f;
	if(l==3 && m==1) rbuf[0]=0.0f;
	if(l==4 && m==1) rbuf[1]=0.0f;		
}

__global__ void sigma2_2(int nsam2, int ib2, float *rbuf,
						float *mccpart, float *msum1, float *msum2)
{
	int l=blockIdx.x*blockDim.x+threadIdx.x+1;
	int tid=threadIdx.x;
	int id=ib2+l;
	float rbufl,rbufid;
	__shared__ float ccpart[512];
	__shared__ float sum1[512];
	__shared__ float sum2[512];
	
	ccpart[tid]=0.0f;
	sum1[tid]=0.0f;
	sum2[tid]=0.0f;
	__syncthreads();
		
	//if(l>nsam2) return;  //nsam2=nsam*nsam
	if(l<=nsam2)
	{
		rbufl=rbuf[l-1];
		rbufid=rbuf[id-1];
		ccpart[tid]+=rbufl*rbufid;
		sum1[tid]+=rbufl*rbufl;
		sum2[tid]+=rbufid*rbufid;
	}
		
	__syncthreads();
	
	//tree-type sum
	if(tid<256)
	{
		ccpart[tid]+=ccpart[tid+256];
		sum1[tid]+=sum1[tid+256];
		sum2[tid]+=sum2[tid+256];
	}
	__syncthreads();
	
	if(tid<128)
	{
		ccpart[tid]+=ccpart[tid+128];
		sum1[tid]+=sum1[tid+128];
		sum2[tid]+=sum2[tid+128];
	}
	__syncthreads();
	
	if(tid<64)
	{
		ccpart[tid]+=ccpart[tid+64];
		sum1[tid]+=sum1[tid+64];
		sum2[tid]+=sum2[tid+64];
	}
	__syncthreads();
	
	if(tid<32)
	{
		ccpart[tid]+=ccpart[tid+32];
		sum1[tid]+=sum1[tid+32];
		sum2[tid]+=sum2[tid+32];
	}
	__syncthreads();
	
	if(tid<16)
	{
		ccpart[tid]+=ccpart[tid+16];
		sum1[tid]+=sum1[tid+16];
		sum2[tid]+=sum2[tid+16];
	}
	__syncthreads();
	
	if(tid<8)
	{
		ccpart[tid]+=ccpart[tid+8];
		sum1[tid]+=sum1[tid+8];
		sum2[tid]+=sum2[tid+8];
	}
	__syncthreads();
	
	if(tid<4)
	{
		ccpart[tid]+=ccpart[tid+4];
		sum1[tid]+=sum1[tid+4];
		sum2[tid]+=sum2[tid+4];
	}
	__syncthreads();
	
	if(tid<2)
	{
		ccpart[tid]+=ccpart[tid+2];
		sum1[tid]+=sum1[tid+2];
		sum2[tid]+=sum2[tid+2];
	}
	__syncthreads();
	
	if(tid<1)
	{
		ccpart[tid]+=ccpart[tid+1];
		sum1[tid]+=sum1[tid+1];
		sum2[tid]+=sum2[tid+1];
		
		mccpart[blockIdx.x]=ccpart[0];
		msum1[blockIdx.x]=sum1[0];
		msum2[blockIdx.x]=sum2[0];
	}	
	__syncthreads();
}

//the thread number must be 512
__global__ void calfinalsum(float *mccpart, float *msum1, float *msum2)
{
	int tid=threadIdx.x;
	__shared__ float ccpart[512];
	__shared__ float sum1[512];
	__shared__ float sum2[512];
	
	ccpart[tid]=mccpart[tid];
	sum1[tid]=msum1[tid];
	sum2[tid]=msum2[tid];
	__syncthreads();
	
	//tree-type sum
	if(tid<256)
	{
		ccpart[tid]+=ccpart[tid+256];
		sum1[tid]+=sum1[tid+256];
		sum2[tid]+=sum2[tid+256];
	}
	__syncthreads();
	
	if(tid<128)
	{
		ccpart[tid]+=ccpart[tid+128];
		sum1[tid]+=sum1[tid+128];
		sum2[tid]+=sum2[tid+128];
	}
	__syncthreads();
	
	if(tid<64)
	{
		ccpart[tid]+=ccpart[tid+64];
		sum1[tid]+=sum1[tid+64];
		sum2[tid]+=sum2[tid+64];
	}
	__syncthreads();
	
	if(tid<32)
	{
		ccpart[tid]+=ccpart[tid+32];
		sum1[tid]+=sum1[tid+32];
		sum2[tid]+=sum2[tid+32];
	}
	__syncthreads();
	
	if(tid<16)
	{
		ccpart[tid]+=ccpart[tid+16];
		sum1[tid]+=sum1[tid+16];
		sum2[tid]+=sum2[tid+16];
	}
	__syncthreads();
	
	if(tid<8)
	{
		ccpart[tid]+=ccpart[tid+8];
		sum1[tid]+=sum1[tid+8];
		sum2[tid]+=sum2[tid+8];
	}
	__syncthreads();
	
	if(tid<4)
	{
		ccpart[tid]+=ccpart[tid+4];
		sum1[tid]+=sum1[tid+4];
		sum2[tid]+=sum2[tid+4];
	}
	__syncthreads();
	
	if(tid<2)
	{
		ccpart[tid]+=ccpart[tid+2];
		sum1[tid]+=sum1[tid+2];
		sum2[tid]+=sum2[tid+2];
	}
	__syncthreads();
	
	if(tid<1)
	{
		ccpart[tid]+=ccpart[tid+1];
		sum1[tid]+=sum1[tid+1];
		sum2[tid]+=sum2[tid+1];
		
		mccpart[0]=ccpart[0];
		mccpart[1]=sum1[0];
		mccpart[2]=sum2[0];
	}	
	__syncthreads();
}


void sigma2(float *psig2, int *pnsam, int *pirad, float *pamag, 
			CPLX *a3df, CPLX *a3ds,
			CPLX *datc, CPLX *datq, float *pphi, float *ptheta, float *ppsi, float *pshx,
			float *pshy, float *sinclut, int *pipad, int *piewald, float *pthetatr,
			CPLX *ctff, CPLX *ctfs, float *rbuf)
{
	
//	timeval begin,end;
//	gettimeofday(&begin,NULL);
	
	int nsam=*pnsam;
	int irad=*pirad;
	float amag=*pamag;
	float phi=*pphi;
	float theta=*ptheta;
	float psi=*ppsi;
	float shx=*pshx;
	float shy=*pshy;
	int ipad=*pipad;
	int iewald=*piewald;
	float thetatr=*pthetatr;
	
	int jc,ib1,ib2,ib3,nsamh;
	float sum1,sum2,ccpart,a;

	cuMem *pGM=(cuMem *)pthread_getspecific(cuMemKey);
	
	ib1=nsam*nsam;
	ib2=ib1+2*nsam;
	ib3=ib2+nsam*nsam;
	nsamh=nsam/2;
	jc=nsamh+1;
	
	cuda_pextract(nsam,irad,(CPLX *)pGM->d_bf,(CPLX *)&(pGM->d_bf[ib1]),
				(CPLX *)pGM->d_c3df,(CPLX *)pGM->d_c3ds,
				phi,theta,psi,pGM->d_sinclut,ipad,
				thetatr/amag,iewald,(CPLX *)pGM->d_ctff,(CPLX *)pGM->d_ctfs);

//#########################	
/*	for(l=1;l<=jc;l++)
		for(m=1;m<=nsam;m++)
		{
			ll=l-1;
			mm=m-1;
			if(mm>=jc) mm=mm-nsam;
			phase=-shx*ll-shy*mm;
			pshft.x=cos(phase);
			pshft.y=sin(phase);
			if(l!=jc)
			{
				id=l+nsamh*(m-1);
				z=cXc(pshft,datc[id-1]);
				id2=2*id+ib2;
				rbuf[id2-2]=z.x;
				rbuf[id2-1]=z.y
			}
			else
			{
				z=cXc(pshft,datq[m-1]);
				m2=2*m+ib3;
				rbuf[m2-2]=z.x;
				rbuf[m2-1]=z.y;
			}			
		}
	rbuf[ib2]=0.0f;
	rbuf[ib2+1]=0.0f;
	rbuf[0]=0.0f;
	rbuf[1]=0.0f;*/
	
	dim3 dimBlock(8,16);
	dim3 dimGrid(jc/dimBlock.x+1,nsam/dimBlock.y+1);

	sigma2_1<<<dimGrid,dimBlock>>>(nsam, jc, shx, shy, ib2,ib3,
				pGM->d_bf, (CPLX *)pGM->d_datc, (CPLX *)pGM->d_datq);
	safecall(cudaThreadSynchronize(),"sigma2_1@@@@@@@");
//#############################
	//rlft3_(&rbuf[ib2],&rbuf[ib3],nsam,nsam,1,-1);
			
	rlft3_GPU((CPLX *)&(pGM->d_bf[ib2]),(CPLX *)&(pGM->d_bf[ib3]),nsam,nsam,1,-1);
	
	//rlft3_(rbuf,&rbuf[ib1],nsam,nsam,1,-1);
			
	rlft3_GPU((CPLX *)pGM->d_bf,(CPLX *)&(pGM->d_bf[ib1]),nsam,nsam,1,-1);
//###############################	
	/*sum1=0.0f;
	sum2=0.0f;
	ccpart=0.0f;
	
	for(l=1;l<=nsam*nsam;l++)
	{
		id=ib2+l;
		ccpart=ccpart+rbuf[l-1]*rbuf[id-1];
		sum1=sum1+rbuf[l-1]*rbuf[l-1];
		sum2=sum2+rbuf[id-1]*rbuf[id-1];
	}*/
	
	float sum[3];
	float *mccpart=pGM->d_mccpart;
	float *msum1=pGM->d_msum1;
	float *msum2=pGM->d_msum2;
	//cudaMalloc((void **)&mccpart,sizeof(float)*512);	
	cudaMemset(mccpart,0,sizeof(float)*512);
	//cudaMalloc((void **)&msum1,sizeof(float)*512);	
	cudaMemset(msum1,0,sizeof(float)*512);
	//cudaMalloc((void **)&msum2,sizeof(float)*512);	
	cudaMemset(msum2,0,sizeof(float)*512);	

	sigma2_2<<<nsam*nsam/512+1,512>>>(nsam*nsam, ib2, pGM->d_bf, mccpart, msum1, msum2);
	cudaThreadSynchronize();

	calfinalsum<<<1,512>>>(mccpart, msum1, msum2);	
	cudaThreadSynchronize();

	cudaMemcpy(sum,mccpart,sizeof(float)*3,cudaMemcpyDeviceToHost);
	ccpart=sum[0];
	sum1=sum[1];
	sum2=sum[2];

	a=ccpart/sum1;
	*psig2=((sum2/a+sum1*a)/ccpart-2.0f)/nsam/nsam;
	
	//cudaFree(mccpart);
	//cudaFree(msum1);
	//cudaFree(msum2);
//##################################



//gettimeofday(&end,NULL);
//cout<<"@"<<CalTime(begin,end)<<endl;
}












